;; -*-mode:scheme; tab-width:4; -*-
;;
;; QScheme 0.6 compiler
;;
;; Generates code for the virtual machine
;;
;; $Id: comp.scm,v 1.22 2000/11/17 00:48:52 crad Exp $

;; need the simple formater for error messages
;;(load "format2str.scm")	

;;
;; Simple string formater
;;


;; Reconized sequence in fmt:
;;
;; ~a		- string like display does
;; ~w		- string like write does
;; ~~		- single ~
;; ~%		- newline

(define (format->string fmt . args)
  (let ((str (make-string 0))
		(n   (string-length fmt)))

	(define (format-loop i args)
;;	  (display "format-loop: i=") (display i)
;;	  (display " args=") (display args)
;;	  (display " str=")  (display str)
;;	  (newline)
	  (if (< i n)
		  (let ((c (string-ref fmt i)))
			(if (char=? (string-ref fmt i) #\~)
				(let ((newstr #f) (next-arg (cdr args)))
				  (set! c (string-ref fmt (+ i 1)))
				  (cond
				   ((char=? c #\~) (set! newstr "~"))
				   ((char=? c #\%) (set! newstr "\n"))
				   ((char=? c #\a) (set! newstr (object->string (car args) #t)))
				   ((char=? c #\w) (set! newstr (object->string (car args) #f)))
				   (else
					(set! newstr (string-append-char "~" c))
					(set! next-arg args)))
				  (set! str (string-append str newstr))
				  (format-loop (+ i 2) next-arg))

				(begin
				  (set! str (string-append-char str c))
				  (format-loop (+ i 1) args))))))

	(format-loop 0 args)
	str))


;; Note:
;; TODO = code to write
;; KKK  = incomplete code

;;----------------------------------------------------------------
;; Global variables
;;----------------------------------------------------------------
(define verbose? #t)					; verbose progress printing
(define compile-ref? #t)				; compile direct reference

(define-macro (when cond . x)  `(if ,cond (begin ,@x)))
(define-macro (unless cond . x) `(if (not ,cond) (begin ,@x)))

(define (info msg . more)
  (when verbose?
	(display "INFO:") (display msg) (display " ")
	(until (null? more)
	  (write (car more)) (display " ")
	  (set! more (cdr more)))
	(newline)))


;;----------------------------------------------------------------
;; Environment 
;;----------------------------------------------------------------
(define (env-search env atom)
  (info "env-search:" env atom)
  (let ((depth 0) (a #f))
	(until (or a (null? env))
	  (set! a (assoc atom (car env)))
	  (unless a
		(set! depth (+ 1 depth))
		(set! env (cdr env))))
	(info "env-search: result=" a)
	(if a (cons depth a) #f)))


;;----------------------------------------------------------------
;; Assembler generator
;;----------------------------------------------------------------
(define va! vector-append)

(define (compile-instr code instr . args)
  (vector-append code `(,instr ,@args)))

;; Push values on stack
(define (compile-pushlit c var)		(va! c `(%pushlit  ,var)))
(define (compile-pushsym c var)		(va! c `(%pushsym  ,var)))

(define (compile-pushloc c var lev)
  ;; optimizes pushloc with known level
  (cond
   ((eq? lev 0)		(va! c `(%pushloc0  ,(cdr var))))
   (else 			(va! c `(%pushloc  ,(+ (cdr var) (* lev 65536)))))))

(define (compile-pushext c var)		(va! c `(%pushext ,var)))

;; Various set
(define (compile-setloc c var lev)
  (cond
   ((eq? lev 0)		(va! c `(%setloc0 ,(cdr var))))
   (else			(va! c `(%setloc  ,(+ (cdr var) (* lev 65536)))))))

(define (compile-setsym c var)		(va! c `(%setsym  ,var)))
(define (compile-setext c var)		(va! c `(%setext  ,var)))

;; environment 
(define (compile-pushenv c)			(va! c '(%pushenv)))
(define (compile-prepcall c)		(va! c '(%prepcall)))

;;call vm prim
(define (compile-call-prim c prim)		(va! c `(%callprim ,prim)))
(define (compile-call-primv c prim n)	(va! c `(%callprimv ,prim ,n)))

;;call C prim
(define (compile-call-cprimx c arity nargs sym prim)
  (let* ((name  #(%callcprim0  %callcprim1  %callcprim2  %callcprim3
			      %callcprim4  %callcprim5  %callcprim6  %callcprim7
			      %callcprim8  %callcprim9  %callcprim10)))
	(info "compile-call-cprimx" arity nargs sym prim)
	(va! c
		 (if (< arity 0)
			 (list '%callcprim (if compile-ref? prim sym) nargs)
			 (list (vector-ref name arity)  (if compile-ref? prim sym)))
		 )))

;; call whatever is on the stack
(define (compile-call c)			(va! c `(%call)))
(define (compile-call-closure c)	(va! c `(%callclos)))

;; jump
(define (compile-jump c nargs)		(va! c `(%jump ,nargs)))

;;KKK: Change here: do not compile directly but optimize the code and
;;generate a (%asm #(code)) for later compilation

(define (compile-make-proc c env optarg code)
  (let* ((compiled-body (vm3-asm code))
		 (proc (make-procedure compiled-body env (length (car env)) optarg)))
	(info "compile-make-proc: body code:")
	(vm3-dis compiled-body)
	(compile-pushlit c proc)))

(define (compile-make-proc c env optarg code)
  (info "compile-make-proc:" env optarg code)
  (va! c `(%proc ,code ,env ,(length (car env)) ,optarg)))

(define (compile-make-closure c)	(va! c `(%makeclos)))

;; compile-return
(define (compile-return c)			(va! c '(%return)))
(define (compile-end c)				(va! c '(%end)))

;; label generation
(define *compile-current-label* 0)

(define (compile-next-label)
  (set! *compile-current-label* (+ 1 *compile-current-label*))
  *compile-current-label*)

;; Branches
(define (compile-label c lab)		(va! c `(%lab ,lab)))
(define (compile-brf c lab)			(va! c `(%brf ,lab)))
(define (compile-brt c lab)			(va! c `(%brt ,lab)))
(define (compile-bra c lab)			(va! c `(%bra ,lab)))

(define (compile-br-and c lab)		(va! c `(%br_and ,lab)))
(define (compile-br-or c lab)		(va! c `(%br_or  ,lab)))

;; Stack maint
(define (compile-drop c)			(va! c '(%drop)))

;; lets
(define (compile-letstart c)		(va! c '(%letstart)))
(define (compile-letbody  c n)		(va! c `(%letbody ,n)))
(define (compile-letend   c n)		(va! c `(%letend  ,n)))

(define (compile-instr-let*    c n)		(va! c `(%letstar ,n)))
(define (compile-instr-letrec  c n)		(va! c `(%letrec ,n)))


;;----------------------------------------------------------------
;; Compiler
;;----------------------------------------------------------------

;; Search for a global symbol, make an unbound one if it does not
;; exist.
;;
(define (make-global-symbol name)
  (let ((sym (hash-search symbol-hash name)))
	(unless sym
	  (hash-set! symbol-hash name (unbound))
	  (set! sym (hash-search symbol-hash name)))
	sym))

;; Compiler for object and expression. Invokes compile-pair when pair
;; is given.
(define (compile-object icode obj env tail)
  ;; Compile code for a global object reference.
  ;; If symbol does not exist, create a new one with unbound value. If
  ;; symbol is bound to a macro or to an extern variable, generate code
  ;; accordingly

  (define (compile-global-object icode obj)
	(let ((sym (hash-search symbol-hash obj)))
	  (info "found in global env:" sym)
	  (unless sym
			  (info "unbound symbol:" obj)
			  (set! sym (make-symbol obj))
			  (hash-set! symbol-hash obj sym))
	  (let ((svalue (symbol-value sym)))
		(cond 
		 ((extern-variable? svalue)
		  (compile-pushext icode (if compile-ref? svalue obj)))

		 ((and (macro? svalue) (not (null? (macro-func svalue))))
		  (compile-pushlit  icode (macro-func svalue)))

		 (else
		  (compile-pushsym  icode sym))))))

  (cond
   ((pair? obj)
	(compile-pair icode obj env tail))

   ((pure-symbol? obj)
	(info "got pure-symbol" obj)
	;; is this symbol already bound ?
	(let ((value (hash-search symbol-hash obj)))
	  (info "pure-symbol value" obj value)))

   ((atom? obj)
	(info "got atom" obj)
	(info "environment is" env tail)
	(if (null? env)
		(compile-global-object icode obj)

		;; have an env chain
		(let ((r (env-search env obj)))
		  (if r
			  (begin (info "found in env" r)
					 (compile-pushloc icode (cdr r) (car r)))
			  (compile-global-object icode obj)))))

   ;; assume default is a ref to a literal
   (else (info "literal object" obj)
		 (compile-pushlit icode obj))
   ))


;; compile arguments for funcall
;; Note: args are never tail
(define (compile-args icode args env tail)
  (if (null? args)
	  icode
	  (begin (compile-object icode (car args) env #f)
			 (compile-args icode (cdr args) env #f))))

;; Compiler for quote

(define (compile-quote icode args env tail)
  (compile-pushlit icode (car args)))

;; The quasiquote compiler

(define (backquotify ic l env tail level)
  (cond
   ((not (pair? l))
	(compile-pushlit ic l))
   ((eqv? (car l) 'quasiquote)
	(backquotify ic (cadr l) env tail (+ 1 level))
	(compile-pushlit 	ic 'quasiquote)
	(compile-call-prim  ic list2))
   ((eqv? (car l) 'unquote)
	(info "unquote" l level)
	(if (eq? level 1)
		(compile-object ic (cadr l) env tail)
		(begin
		  (backquotify ic (cadr l) env tail (- 1 level))
		  (compile-pushlit ic 'unquote)
		  (compile-call-prim ic list2))))
   ((and (pair? (car l)) (eqv? (caar l) 'unquote-splicing))
	(compile-object ic (car (cdar l)) env tail)
	(unless (null? (cdr l))
			(backquotify ic (cdr l) env tail level)
			(compile-call-prim ic qq-append2)))
   (else
	(backquotify ic (car l) env tail level)
	(backquotify ic (cdr l) env tail level)
	(compile-call-prim ic cons)))
  ic)

(define (compile-qquote icode arg env tail)
  (info "compile-qquote: " arg)
  (backquotify icode (car arg) env tail 1))

;; Compiler for set

(define (compile-set! icode args env tail)
  (info "compile-syntax: set!" args)
  (compile-object icode (cadr args) env #f)
  (let ((sym (env-search env (car args))))
	(if sym								;found in local env
		(begin							;local symbol
		  (compile-setloc   icode (cdr sym) (car sym)))
		(begin
		  (set! sym (hash-search symbol-hash (car args)))
		  (if (not sym) 
			  (throw 'comp-error	"set! needs an existing symbol"))
		  (cond
		   ((extern-variable? (symbol-value sym))
			(compile-setvar icode
							(if compile-ref? (symbol-value sym) (car args))))
		   (else (compile-setsym icode sym)))))))


;; Compiler for if
;; Note: True and False expr are tails

(define (compile-if icode args env tail)
  (info "compile-syntax: if" args)
  (info "compile-syntax: if tail" tail)
  
  (let ((len (length args)))
	(info "compile-syntax: if nargs" len)
	(cond
	 ((eq? len 2)
	  (let ((labx (compile-next-label)))
		;; (if E T)   => E (br_and x) T (lab x)
		(compile-object icode (car args) env #f)
		(compile-br-and icode labx)
		(compile-object icode (cadr args) env tail)
		(compile-label  icode labx)))

	 ((eq? len 3)
	  (let ((labx (compile-next-label)) (laby (compile-next-label)))
		;; (if E T F) => E (brf x) T (bra y) (lab x) F (lab y)
		(compile-object icode (car args) env #f)
		(compile-brf    icode labx)
		(compile-object icode (cadr args) env tail)
		(compile-bra    icode laby)
		(compile-label  icode labx)
		(compile-object icode (caddr args) env tail)
		(compile-label  icode laby)
		))
	 (else
	  (throw 'comp-err "if syntax: (if <expr> <true> [<false>])")))))

;; Compiler for (begin ...)
;; Note: last expr is tail.

(define (compile-begin icode args env tail)
  (info "compile-syntax: begin" args env tail)
  (if (null? args)
	  (compile-pushlit icode #t)
	  (until (null? args)
		(if (and tail (null? (cdr args)))
			(compile-object icode (car args) env #t)
			(begin 
			  (compile-object icode (car args) env #f)
			  (unless (null? (cdr args)) (compile-drop   icode))))
		(set! args (cdr args)))))

;; Compiler for (lambda (..) ..)
;; Note: last expr is tail recursive
(define (compile-lambda icode args env tail)
  (let ((optarg #f))

	(define (check-arg-unicity obj list)
	  (if (assoc obj list)
		  (throw 'comp-err "argument defined twice")))


	(define (build-arg-list l)
	  (let ((arglist '()) (argnum 0) (last #f) (new #f))
		
		(while (pair? l)
		  (check-arg-unicity (car l) arglist)
		  (set! new  (cons (cons (car l) argnum) '()))
		  (if last   (set-cdr! last new)   (set! arglist new))
		  (set! last new);
		  (set! l (cdr l))
		  (set! argnum (+ 1 argnum)))
		(if (not (null? l))
			(begin
			  (check-arg-unicity l arglist)
			  (set! new (cons (cons l argnum) '()))
			  (if last   (set-cdr! last new)   (set! arglist new))
			  (set! optarg #t)))
		arglist))

	;; Try to detect if we should preserve environment.

	;; Environment should be preserved if body of lambda references
	;; variables defined in the enclosing environment. In other words,
	;; variable that are not global and not argument of the lambda.
	

	(define (has-external-refs? body env)
	  (info "has-external-refs?: body, env" body env)
	  (cond
	   ((null? body)			#f)
	   ((pair? (car body))		(has-external-refs? (car body) env))
	   ((atom? (car body))
		(if (env-search env (car body))
			#t
			(has-external-refs? (cdr body) env)))
	   (else #f)))

	(when verbose? (display "lambda: args=") (print args))
	(if (not (>= (length args) 2))
		(throw 'comp-err "lambda requires at least 2 elements"))
	(let ((formal (car args))
		  (body (cdr args))
		  (newenv '())
		  (body-code (make-vector 0)))

	  (set! newenv (cons (build-arg-list formal) env))
	  (when verbose?
			(display "formal=")(print formal)
			(display "body=") (print body)
			(display "newenv=") (print newenv)
			(display "nargs=")  (print (length (car newenv)))
			(display "optargs=") (print optarg))

	  ;;KKK: closure detection 
	  (if (has-external-refs? body env)
		  (info "compile-lambda: has external refs")
		  (info "compile-lambda: does not have external refs"))
	  
	  (compile-begin body-code body newenv #t)

	  ;; Note: return will possibly be removed by optimizer
	  (compile-return body-code)

	  (compile-make-proc icode newenv optarg (compile-optimize body-code))
	  (compile-make-closure icode)
	  )))

;; compiler for define
;; accepts: (define ATOM [EXPR]) | (define PAIR ...)

(define (compile-define icode args env tail)
  (let ((first (car args)) (rest (cdr args)))
	(cond
	 ((atom? first)						; define for a variable
	  (if (>= (length rest) 2) (throw 'comp-error "too many expr"))
	  (if (< (length rest) 1) (set! rest (list (unbound))))
	  (let ((sym (env-search env first)))
		(if sym
			(begin						;local symbol
			  (info "compile-define: local symbol found" sym)
			  (compile-object icode (car rest) env #f)
			  (compile-setloc icode (cdr sym) (car sym)))
			(begin						; global symbol
			  (info "compile-define: not a local symbol" first)
			  (set! sym (make-global-symbol first))
			  (compile-object icode (car rest) env #f)
			  (compile-setsym icode sym)))))

	 ((pair? first)						; define for a function
	  (let ((func-name (car first))
			(lambda-args (cdr first)))
		(info "compile-define: function name" func-name)
		(info "compile-define: lambda args"  lambda-args)
		(info "compile-define: lambda"
			  `(,func-name (lambda ,lambda-args ,@rest)))
		(compile-define icode
						`(,func-name (lambda ,lambda-args ,@rest))
						env
						tail)
	  
	  ))
	 (else
	  (throw 'comp-error "bad define syntax"))
	 );cond
	);let
  )

;; compiler for let
;;
;; Let values are taken outside of the (let ...) scope
;; Runtime translation:
;;
;;TODO

;; Check unicity of symbol in let binding
;;
(define (compile-let:check-binding-unicity args)
  (if (or (null? args) (null? (cdr args)))
	  #f
	  (let ((search (assoc (caar args) (cdr args))))
		(info "compile-let:check-binding-unicity: testing"
			  (caar args) (cdr args) search)
		(if search
			search
			(compile-let:check-binding-unicity (cdr args))))))
		
;; Check for correct binding format in let declaration 
;;
(define (compile-let:check-bindings-format l)
  ;; first check for valid list
  (if (not (list? l))
	  (throw 'comp-err
		(format->string 
		 "let binding: bad let binding syntax. ~w is not a list" bind)))

  ;; check for unique symbols
  (let ((sym (compile-let:check-binding-unicity l)))
	(if sym
		(throw 'comp-err
		 (format->string "let binding: symbol '~a is defined multiple times"
						 (car sym))))))

;; Compile code for binding creation
;;
(define (compile-let-binding code bind env)
  (unless (null? bind)
	(let ((expr (cdar bind)))
	  (if (null? expr)
		  (compile-pushlit code (unbound))
		  (compile-object code (car expr) env #f)))
	(compile-let-binding code (cdr bind) env #f)))

;; Build a new env for let
;;
(define (compile-let:build-new-env l)
	(let ((blist '()) (num 0) (last #f) (new #f))
	  (while (pair? l)
		(set! new (cons (cons (caar l) num) '()))
		(if last (set-cdr! last new) (set! blist new))
		(set! last new)
		(set! l (cdr l))
		(set! num (+ num 1)))
	  blist))
  
;; Compiler for let
;;
(define (compile-let code args env tail)

  (let ((bind (car args)) (body (cdr args)) (newenv '()))

	(info "compile-let: bind=" bind)
	(info "compile-let: body=" body)
	(compile-let:check-bindings-format bind)

	(set! newenv  (cons (compile-let:build-new-env bind) env))
	(info "compile-let: newenv=" newenv)

	(compile-letstart 		code)
	(compile-let-binding  	code bind env)
	(compile-letbody 		code (length (car newenv)))
	(compile-begin    		code body newenv #f)
	(compile-letend			code (length (car newenv)))
	)
  )

;; compiler for let*
;;
(define (compile-let* code args env tail)
  (let ((bind (car args)) (body (cdr args))	(newenv '()))

	(info "compile-let*: bind=" bind)
	(info "compile-let*: body=" body)
	(compile-let:check-bindings-format bind)

	(set! newenv (cons (compile-let:build-new-env bind) env))
	(info "compile-let: newenv=" newenv)

	(compile-instr-let*		code (length (car newenv)))
	(compile-let-binding	code bind newenv)
	(compile-begin    		code body newenv #f)
	(compile-letend			code (length (car newenv)))
	)
  )

;; Compile code for letrec binding creation
;;
(define (compile-letrec-binding code bind env)
  (let ((varnum 0))
	(until (null? bind)
	  (let ((expr (cdar bind)))
		(unless (null? expr)
		  (compile-object  code (car expr) env #f)
		  (compile-setloc  code (cons '() varnum) 0)))
	  (set! bind (cdr bind))
	  (set! varnum (+ varnum 1)))))
		  
;; compiler for letrec
;;
(define (compile-letrec code args env tail)
  (let ((bind (car args)) (body (cdr args))	(newenv '()))

	(info "compile-letrec: bind=" bind)
	(info "compile-letrec: body=" body)
	(compile-let:check-bindings-format bind)
	(set! newenv (cons (compile-let:build-new-env bind) env))

	(compile-instr-letrec	 code (length (car newenv)))
	(compile-letrec-binding  code bind newenv)
	(compile-begin			 code body newenv #f)
	(compile-letend			 code (length (car newenv)))))


;; the-env
;;
(define (compile-the-env code args env tail)
  (compile-pushlit	code env)
  )

;; compiler for and / or
;;
;;TODO

(define (compile-and/or null-value branch-func icode args env tail)
  (if (zero? (length args))
	  (compile-pushlit	icode #t)
	  (begin
		(let ((lab (compile-next-label)))
		  (until (null? args)
			(compile-object icode (car args) env #f)
			(set! args (cdr args))
			(if (not (null? args))  (branch-func icode lab)))
		  (compile-label icode lab)))))

(define (compile-and icode args env tail)
  (compile-and/or #t compile-br-and icode args env tail))
  
(define (compile-or icode args env tail)
  (compile-and/or #f compile-br-or icode args env tail))

;; compiler for cond
;;
;;TODO
(define (compile-cond icode args env tail)
  (if (null? args)
	  (compile-pushlit icode #f)
	  (let ((exitlab (compile-next-label)) (nextlab))
		(until (or (null? args)
				   (eq? (caar args) 'else))
		  (let ((clause (car args)))
			(set! nextlab (compile-next-label))
			(info "cond: clause" clause)
			(compile-object icode (car clause) env #f)
			(compile-brf    icode nextlab)
			(compile-begin  icode (cdr clause) env tail)
			(compile-bra	icode exitlab)
			(compile-label  icode nextlab)
			)
		  (set! args (cdr args)))
		(if (and (not (null? args)) (not (eq? (caar args) 'else)))
			(throw 'comp-error "bad else clause"))

		(if (null? args)
			(compile-pushlit 	icode #f)
			(compile-begin  	icode (cdar args) env tail))

		(compile-label icode exitlab)
		)))

;; compiler for catch
;;
;;TODO
(define (compile-catch code args env tail)
  (if (< (length args) 3)
	  (throw 'comp-error "catch: bad number of args"))
  
  (let ((tag (car args))
		(func (cadr args))
		(rest (cddr args))
		(lab (compile-next-label)))
	(info "compile-catch: tag=" tag)
	(info "compile-catch: func=" func)
	(info "compile-catch: rest=" rest)
	(compile-pushlit code tag)
	(compile-object  code func env #f)
	(compile-instr   code '%catch lab)
	(compile-begin	 code rest env #f)
	(compile-instr   code '%uncatch)
	(compile-label   code lab)
	)
  ;(throw 'comp-error "compile-catch: unimplemented")
  )

;; compiler for do
;;
;;TODO
(define (compile-do icode args env tail)
  (if (< (length args) 2)  (throw 'comp-err "bad do construct"))
  (let ((init (car args))
		(test (cadr args))
		(body (cddr args))
		(exitlab (compile-next-label))
		(looplab (compile-next-label))
		(count 0))

	(info "do: init" init)
	(info "do: test" test)
	(info "do: body" body)
	(let* ((var-name (map car init))  (var-value (map cadr init))  (newenv))
	  (info "do: var-name" var-name)
	  (info "do: var-value" var-value)
	  (compile-instr-let* icode (length var-name))
	  (set! newenv
			(cons
			 (let ((i -1))
			   (map (lambda (x) (set! i (+ i 1)) (cons x i)) var-name))
			 env))
	  (info "do: newenv" newenv)
	  (until (null? var-value)
		(compile-object icode (car var-value) newenv #f)
		(set! var-value (cdr var-value)))

	  (compile-label  icode looplab)
	  (compile-object icode (car test) newenv #f)
	  (compile-brt    icode exitlab)
	  (compile-begin  icode body newenv #f)
	  (compile-drop   icode)
	  (set! count 0)
	  (until (null? init)
		(unless (null? (cddar init))
		  (compile-object icode (caddar init) newenv #f)
		  (compile-setloc icode (cons '() count) 0)
		  (compile-drop   icode))
		(set! count (+ 1 count))
		(set! init (cdr init))
		)
	  (compile-bra	  icode looplab)
	  
	  (compile-label  icode exitlab)
	  (compile-letend icode (length var-name))
		  )
	))


;; compiler for loops
;;
(define (compile-loop code args env br-op)
  (let* ((testlab (compile-next-label))
		 (looplab (compile-next-label)))
	(compile-bra		code testlab)
	(compile-label		code looplab)
	(compile-begin		code (cdr args) env #f)
	(compile-drop		code)
	(compile-label		code testlab)
	(compile-object		code (car args) env #f)
	(compile-instr		code br-op looplab)
	(compile-pushlit  	code #t)

	;;(compile-object   code (car args) env tail)
	;;(compile-instr    code br-op testlab)
	;;(compile-label	code looplab)
	;;(compile-begin    code (cdr args) env tail)
	;;(compile-drop     code)
	;;(compile-bra		code looplab)
	;;(compile-label    code testlab)

	code))

;; compiler for (while TEST EXPR...)
;;
(define (compile-while code args env tail)
  (info "compile-while:" args)
  (compile-loop code args env '%brt))

;; compiler for (until TEST EXPR...)
;;
(define (compile-until code args env tail)
  (info "compile-until:" args)
  (compile-loop code args env '%brf))


;; Possible implementation of the syntax compiler is to use a hash
;; whose key is the type and whose value is a compiler function

; (define compile-syntax-hash
;   #{
; 	(quote 			. #.compile-quote)
; 	(quasiquote		. #.compile-qquote)
; 	(set!			. #.compile-set!)
; 	(if				. #.compile-if)
; 	(begin			. #.compile-begin)
; 	(lambda			. #.compile-lambda)
; 	(define			. #.compile-define)
; 	(let			. #.compile-let)
; 	(let*			. #.compile-let*)
; 	(letrec			. #.compile-letrec)
; 	(the-env		. #.compile-the-env)
; 	(while			. #.compile-while)
; 	(until			. #.compile-until)
; 	(and			. #.compile-and)
; 	(or				. #.compile-or)
; 	(cond			. #.compile-cond)
; 	(do				. #.compile-do)
; 	})
(define compile-syntax-hash
  (list->hash
   `(
	 (quote 		. ,compile-quote)
	 (set!			. ,compile-set!)
	 (if			. ,compile-if)
	 (begin			. ,compile-begin)
	 (lambda		. ,compile-lambda)
	 (define		. ,compile-define)
	 (let			. ,compile-let)
	 (let*			. ,compile-let*)
	 (letrec		. ,compile-letrec)
	 (the-env		. ,compile-the-env)
	 (while			. ,compile-while)
	 (until			. ,compile-until)
	 (and			. ,compile-and)
	 (or			. ,compile-or)
	 (cond			. ,compile-cond)
	 (do			. ,compile-do)
	 (catch			. ,compile-catch)
	 )))

;; Special definition, because quasiquote fools quasiquote
(hset! compile-syntax-hash 'quasiquote compile-qquote)

(define (syntax? obj)
  (let ((val (hash-search compile-syntax-hash obj)))
	(if (not val) #f val)))

(define (compile-funcall icode func args env tail)

  ;; syntax compiler
  (define (compile-syntax icode syntax args env tail)
	(let ((func (href compile-syntax-hash syntax)))
	  (if (not func) (throw 'comp-err "bad syntax"))
	  (info "compile-syntax:" func)
	  (func icode args env tail)))

  (info "compiling funcall" func args)
  (info "compiling funcall: env is" env)
  (when (atom? func) (info "compile-funcall: func is atom" func))
  (when (pure-symbol? func) (info "compile-funcall: func is pure-symbol" func))

  (let ((sym (env-search env func)) (nargs (length args)))
	(if sym								; local symbol

		(begin							; symbol found in env chain
		  (if (not tail)  (compile-prepcall	icode))
		  (compile-args 	icode args env tail)
		  (compile-pushloc 	icode (cdr sym) (car sym))
		  (if tail
			  (compile-jump icode (length args))
			  (compile-call	icode)))

		(begin							; should be a global symbol
		  (set! sym (hash-ref symbol-hash func))
		  (info "compile-funcall: global symbol" sym)
		  (cond
		   ((primitive? sym)			; is a VM primitive
			(let ((arity (primitive-arity sym)))
			  (compile-args 	icode args env tail)
			  (if (negative? arity)
				  (compile-call-primv 	icode func nargs)
				  (begin				; fixed number of args
					(if (not (eq? nargs arity))
						(throw 'comp-error "bad arity"))
					(compile-call-prim 	icode func))))
			)

		   ((cprimitive? sym)			; is a cprimitive
			(let ((arity (cprimitive-arity sym)))
			  (compile-args icode args env tail)
			  (compile-call-cprimx icode arity nargs func sym)
			  ))

		   ((macro? sym)
			(info "compile-funcall: macro")
			(compile-object icode (execute-macro sym (cons sym args) env) env tail))

		   ((syntax? func)
			(compile-syntax icode func args env tail))

		   ((closure? sym)
			(compile-prepcall 		icode)
			(compile-args 			icode args env tail)
			(compile-pushsym 		icode (hash-search symbol-hash func))
			(compile-call-closure 	icode))

		   (else
			;; unbound global symbol: something like forward
			;; reference. Create it now.
			(info "compile-funcall: symbol should be unbound" sym)
			(set! sym (make-global-symbol func))
			(if (not tail)
				(compile-prepcall icode))
			(compile-args		icode args env tail)
			(compile-pushsym 	icode sym)
			(if tail
				(compile-jump	icode (length args))
				(compile-call 	icode))))
		  )))
  icode
  )

(define (compile-pair icode form env tail)
  (let ((f (car form)))
	(info "compiling pair" form)
	(cond
	 ((pair? f)
	  (info "compile-pair: first is pair" f)
	  (compile-prepcall 	icode)
	  (compile-args 		icode (cdr form) env tail)
	  (compile-pair 		icode f env tail)
	  (compile-call			icode)
	  ;; (throw 'comp-error "not implemented")
	  )
	 ((number? f)
	  (info "compile-pair: first is number" f)
	  (throw 'comp-error "illegal (number ...) form"))
	 ((symbol? f)
	  (compile-funcall icode f (cdr form) env tail))
	 ((macro? f)
	  (info "compile-pair: macro" f))
	 (else
	  (throw 'comp-error "unsupported func"))))
  )


;; resolve label

(define (compile-resolve-labels code)
  (let ((i 0) (len (vector-length code))
		(instr '())
		(h '())
		)
	(info "compile-resolve-label")
	(set! h (make-hash))

	;; first pass: build hash of labels
	(while (< i len)
	  (set! instr (vref code i))
	  (if (eq? (car instr) '%lab)
		  (begin
			(info "compile-resolve-label: found label" (cadr instr))
			(hset! h (cadr instr) i)
			)
		  )
	  (set! i (+ i 1))
	  )
	(info "compile-resolve-label: label hash is" h)

	;; 2nd pass: resolve branches
	(set! i 0)
	(while (< i len)
	  (set! instr (vref code i))
	  (case (car instr)
		((%bra %brf %brt)
		 (info "compile-resolve-label: step" i " to " (href h (cadr instr)))
		 (set-cdr! instr (cons (href h (cadr instr)) '())))
		((%lab)
		 (info "compile-resolve-label: %lab " i " to %nop")
		 (vset! code i '%nop)))
	  (set! i (+ i 1)))

	;; removing nop
	(set! i 0)
	(while (< i (vector-length code))
	  (info "compile-resolve-label: code" code)
	  (if (eq? (vref code i) '%nop)
		  (let ((j 0) (l i) (ins '()))
			(info "compile-resolve-label: %nop at" i)
			(while (< j l)
			  (set! ins (vref code j))
			  (if (and (member (car ins) '(%bra %brf %brt))
					   (> (cadr ins) l))
				  (begin
					(set-car! (cdr ins) (- (cadr ins) 1))
					(info "compile-resolve-label: new instr" ins)))
			  
			  (set! j (+ 1 j)))
			(vector-remove! code i))
		  ;; not a nop
		  (set! i (+ i 1)))
	  )
	))


(define (compile-optimize code)
  (define (extract-labels code)
	(let ((i 0) (len 0) (instr) (v))
	  (set! i 0)
	  (set! v (make-vector 0 0))
	  (while (< i (vector-length code))
		(set! instr (vref code i))
		(if (eq? (car instr) '%lab)
			(begin
			  (info "found label" (cadr instr))
			  (vector-append v (cons i (cadr instr)))
			  (vector-remove! code i))

			(set! i (+ i 1))))
	  
	  (info "label crossref" v)
	  v))

  (define (make-labref code)
	(let ((i 0)
		  (len (vector-length code))
		  (instr)
		  (labref))
	  (set! labref '())
	  (while (< i len)
		(set! instr (vref code i))
		(if (eq? (car instr) '%lab)
			(begin
			  (set! labref (cons (cons (cadr instr) i) labref))))
		(set! i (+ i 1)))
	  
	  (info "labref is" labref)
	  labref))
  
  (let ((i 0) (instr) (limit) (labref))
	(info "compile-optimize" code)
	;; (compile-resolve-labels code)
	;; (set! labref (extract-labels code))

	(set! limit (vector-length code))
	(set! labref (make-labref code))

	(while (< i limit)
	  (set! instr (vref code i))
	  ;; transform (%callprim prim) => (prim)
	  (if (or (eq? (car instr) '%callprim) (eq? (car instr) '%callprimv))
		  (begin
			(info "optimize: removing %callprim" instr)
			(vset! code i (cdr instr))))
	  (set! i (+ i 1)))
	
	(set! i 0)
	(while (< i limit)
	  (set! instr (vref code i))
	  
	  ;; try to propagate return and end
	  (if (eq? (car instr) '%bra)
		  (let ((where (assoc (cadr instr) labref)))
			(info "branch at" i)
			(info "target of branch is" (cdr where))
			(let ((j (cdr where)) (propagate #f) (flag #t))
			  (while (and flag (not propagate) (< j limit))
				(if (not (eq? (car (vref code j)) '%lab))
					(if (member (car (vref code j)) '(%return %end))
						(set! propagate j)
						(set! flag #f)))
				(set! j (+ j 1)))
			  (info "propagate from" propagate)
			  (when propagate
				(vset! code i (vref code propagate))
				(info "propagate:" code)
				(set! i -1)))		;start again
			))
	  
	  (set! i (+ i 1))
	  )
	)
  code)

(define (compile-expr form env)
  (let ((icode (make-vector 0)))
	(compile-object icode form env #f)
	(compile-end icode)
	(compile-optimize icode)
	icode))

(define pp-code-depth 1)

(define (tabs n)
  (if (>= n 0) (begin (display #\tab) (tabs (- n 1)))))

(define (pp-code1 code)
  (let ((i 0) (l (vector-length code)) (expr))
	(while (< i l)
		   (set! expr (vector-ref code i))
;		   (print (format->string "pp-code1: expr=~w" (car expr)))
		   (cond
			((eq? (car expr) '%proc)
			 (tabs pp-code-depth) (print "(%proc #(")
			 (set! pp-code-depth (+ pp-code-depth 1))
			 (pp-code1 (cadr expr))
			 (tabs pp-code-depth)
			 (print (format->string ") ~w ~w ~w)"
									(nth 2 expr) (nth 3 expr) (nth 4 expr)))
			 (set! pp-code-depth (- pp-code-depth 1)))
			((eq? (car expr) '%pushsym)
			 (tabs pp-code-depth)
			 (print (format->string "(%pushsym ~a)"
									(symbol->string (cadr expr)))))
			((eq? (car expr) '%setsym)
			 (tabs pp-code-depth)
			 (print (format->string "(%setsym ~a)"
									(symbol->string (cadr expr)))))
			((primitive? (car expr))
			 (tabs pp-code-depth)
			 (display (format->string "(~a" (primitive-name (car expr))))
			 (unless (null? (cdr expr))
			   (display (cdr expr)))
			 (print ")"))
			 
			(else
			 (tabs pp-code-depth) (write expr) (newline)
			 ))
		   (set! i (+ i 1)))))

(define (pp-code code)
  (display "#(")  (pp-code1 code)  (print ")"))

(define (compile-file file)
  (let ((was-verbose? verbose?)
		(did-compile-ref? compile-ref?))
	(set! verbose? #f)
	(set! compile-ref? #f)
	(with-input-from-file file
	  (lambda ()
		(let ((expr) (code))
		  (until (eof-object? (set! expr (read)))
			(gc)
			(pp-code (compile-expr expr '()))))))
	(set! verbose? was-verbose?)
	(set! compile-ref? did-compile-ref?)
	))

(define (compile-to-file file)
  (let* ((i (string-index file ".scm"))
		 (new (string-append (if i (substring file 0 i) file) ".sbf")))
	(display "output to ") (print new)
	(with-output-to-file new
	  (lambda ()
		(compile-file file)))))

(define (vm3-repl)
  (let ((expr) (code))
	(display "vm3> ")
	(until (eof-object? (set! expr (read)))
	  (set! code (compile-expr expr '()))
	  (display "code: ") (print code)
	  (print (vm3-exec (vm3-asm code)))
	  (display "vm3> "))))

(define (vm3-eval expr)
  (vm3-exec (vm3-asm (compile-expr expr '()))))

(define (vm3-compile expr)
  (vm3-nasm (vm3-asm (compile-expr expr '()))))

(define (vm3-load file)
  (let ((expr) (code) (was-verbose? verbose?))
	(set! verbose? #f)
	(with-input-from-file file
	  (lambda ()
		(until (eof-object? (set! expr (read)))
		  (display "; ") (print expr)
		  (print (compile-expr expr '())))))
	(set! verbose? was-verbose?)))

;; Bootstrap the system
;;(if bootstraping
;;	(vm3-exec 
;;	 (vm3-asm #((%prepcall) (%pushsym vm3-repl) (%callclos) (%end)))))
