/* -*- tab-width:8; -*-
 *
 * Jit compiler for QScheme. 
 *
 * $Id: jit-comp.c,v 1.5 2000/10/20 20:46:38 crad Exp $
 */

#include "s.h"

/* VM configuration */

#include "vm3.h"
#include "jit/lightning.h"

/* FWR:
 * index: index relative to start of icode where something branches to.
 * addr:  the location where instruction was compiled before.
 */
#define SCM_JC_FWR	struct _SCM_JC_FWR
SCM_JC_FWR {
  int		index;		/* index, relative to start of icode */
  jit_insn	*addr;		/* instruction to patch */
};

#include "jit-comp.h"

/* jit label resolution */

static void scm_jc_add_forward(SCM_JC *jc, int index, jit_insn *fix)
{
  int i;

  fprintf(stderr, "scm_jc_add_forward: index=%d fix=%p\n", index, fix);
  
  for (i = 0; i < jc->nfwr; i++) {
    if (jc->fwr[i].index == -1) 	goto found;
  }
  i = jc->nfwr++;
found:
  jc->fwr[i].index = index;
  jc->fwr[i].addr  = fix;
}

static void scm_jc_try_forward_resolv(SCM_JC *jc, int index)
{
  int i;
  
  for (i = 0; i < jc->nfwr; i++) {
    if (jc->fwr[i].index == index) { /* have something to resolve */
      fprintf(stderr, "scm_jc_try_forward_resolv: index=%d addr=%p -> %p\n",
	      index, jc->fwr[i].addr, jit_get_ip().ptr);

      jit_patch(jc->fwr[i].addr);
      
      /* try to compact pending array or at least mark as free */
      if (i == (jc->nfwr - 1)) { /* last of array */
	jc->nfwr--;
      } else {
	jc->fwr[i].index = -1;
      }
    }
  }
}

/********************************
 * jit special compiler
 ********************************/
static void scm_jc_inline_op(SCM_JC *jc, int op)
{
  int len = jc->prim_table[op].end - jc->prim_table[op].start;

  memcpy(jc->ncp, jc->prim_table[op].start, len);
  
  jc->ncp += len;
  (void)jit_set_ip(jc->ncp);
}


/* compile code to call a vm primitive and return to native code back */
void scm_jc_callprim(SCM_JC *jc, int op)
{
  void *oldpc, *restart_addr, *prim = jc->prim_table[op].start;
  int i;

  /* 2 pass compiler: first pass generates dummy code to obtain branch
   * address (restart_addr). During 1st pass, restart_addr is 0x12345678 */

  oldpc = jit_get_ip().ptr;
  restart_addr = (void*)0x12345678;
  for (i = 0; i < 2; i++) {
    jit_set_ip	(oldpc);
    jit_movi_l	(VM_TMP,(long)restart_addr);
    jit_pushr_i	(VM_TMP);
    jit_pushr_i	(VM_IP);
    jit_addi_l	(VM_IP, JIT_SP, sizeof(long));
    jit_jmpi	(prim);
    restart_addr = jit_get_ip().ptr;
  }

  /* this is where the primitive will return after the NEXT macro */
  jit_popr_i(VM_IP);
  jit_addi_l(JIT_SP, JIT_SP, sizeof(long));
}

/* call a scheme function. */
void scm_jc_call(SCM_JC *jc, int op)
{
  extern void *scm_call_table[];
  void *oldpc, *restart_addr;
  int i;

  /* get address of the gen_callxxxx func to call */
  jit_movr_l(VM_TMP, VM_TOS);
  jit_ldr_l (VM_TMP, VM_TMP);
  jit_lshi_l(VM_TMP2, VM_TMP, 2);
  jit_ldxi_l(VM_TMP2, VM_TMP2,  (long)scm_call_table);
  /*KKK: copy code of call prim */

  /* see comment on jc_callprim */
  oldpc = jit_get_ip().ptr;
  restart_addr = (void*)0x12345678;
  for (i = 0; i < 2; i++) {
    jit_set_ip	(oldpc);
    jit_movi_l	(VM_TMP,(long)restart_addr);
    jit_pushr_i	(VM_TMP);
    jit_pushr_i	(VM_IP);
    jit_addi_l	(VM_IP, JIT_SP, sizeof(long));
    jit_jmpr	(VM_TMP2);
    restart_addr = jit_get_ip().ptr;
  }
  /* this is where the primitive will return after the NEXT macro */
  jit_popr_i(VM_IP);
  jit_addi_l(JIT_SP, JIT_SP, sizeof(long));
}


/****************************************************************
 * JIT implementation of primitives
 ****************************************************************/

static void scm_jc_pushlit(SCM_JC *jc, int op)
{
  SOBJ lit = jc->icode[jc->icp++];
  
  if (!SCM_INUMP(lit)) 	jc->mcode->lit[jc->mcode->nlits++] = lit;

  jit_str_l	(VM_SP,  VM_TOS);
  jit_addi_l	(VM_SP,  VM_SP, sizeof(SOBJ));
  jit_movi_l	(VM_TOS, (long)lit);
}

static void scm_jc_pushsym(SCM_JC *jc, int op)
{
  SOBJ lit = jc->icode[jc->icp++];

  if (!SCM_INUMP(lit)) 	jc->mcode->lit[jc->mcode->nlits++] = lit;

  jit_str_l	(VM_SP,  VM_TOS);
  jit_addi_l	(VM_SP,  VM_SP, sizeof(SOBJ));
  jit_movi_l	(VM_TMP, (long)lit);
  jit_ldxi_l	(VM_TOS, VM_TMP, offsetof(Sobject, data.symbol.value));
}

static void scm_jc_pushloc0(SCM_JC *jc, int op)
{
  int loc = SCM_INUM(jc->icode[jc->icp++]);
  jit_str_i	(VM_SP,  VM_TOS);
  jit_addi_i	(VM_SP,  VM_SP, sizeof(SOBJ));
#ifdef HAS_ENV_REG
  jit_ldxi_i	(VM_TOS, VM_ENV, 8 + (loc * sizeof(SOBJ)));
#else
  jc->ncp = jit_get_ip().ptr;  scm_jc_inline_op(jc, SCM_OP_LOAD_ENV);
  jit_ldxi_i	(VM_TOS, VM_TMP, 8 + (loc * sizeof(SOBJ)));
#endif	
}

static void scm_jc_setsym(SCM_JC *jc, int op)
{
  SOBJ lit = jc->icode[jc->icp++];

  if (!SCM_INUMP(lit)) 	jc->mcode->lit[jc->mcode->nlits++] = lit;

  jit_movi_p(VM_TMP, (long)lit);
  jit_stxi_p(offsetof(Sobject, data.symbol.value), VM_TMP, VM_TOS);
}
			 
static void scm_jc_setloc0(SCM_JC *jc, int op)
{
  int loc = SCM_INUM(jc->icode[jc->icp++]);

#ifdef HAS_ENV_REG
  jit_stxi_l(8 + (loc * sizeof(SOBJ)), VM_ENV, VM_TOS);
#else
  jc->ncp = jit_get_ip().ptr;  scm_jc_inline_op(jc, SCM_OP_LOAD_ENV);
  jit_stxi_l(8 + (loc * sizeof(SOBJ)), VM_TMP, VM_TOS);
#endif
}

static void scm_jc_branch(SCM_JC *jc, int op)
{
  int ofs, target;
  jit_insn *lab = NULL;
  void *addr;

  ofs = SCM_INUM(jc->icode[jc->icp]);
  target = jc->icp + ofs;
  jc->icp++;

  addr = (ofs < 0) ? jc->ref[target] : jit_forward();

  switch(op) {
  case SCM_OP_BRA:	
    lab = jit_jmpi(addr);
    break;
  case SCM_OP_BRF:
    jit_movr_i(VM_TMP, VM_TOS); 	/* save TOS to eax */
    jit_subi_i(VM_SP, VM_SP, 4); 	/* pop stack */
    jit_ldr_l(VM_TOS, VM_SP);
    lab = jit_beqi_i(addr, VM_TMP, (long)scm_false);	
    break;
  case SCM_OP_BRT:	
    jit_movr_i(VM_TMP, VM_TOS); 	/* save TOS to eax */
    jit_subi_i(VM_SP, VM_SP, 4); 	/* pop stack */
    jit_ldr_l(VM_TOS, VM_SP);
    lab = jit_bnei_i(addr, VM_TMP, (long)scm_false);	
    break;
  case SCM_OP_BR_AND:
  case SCM_OP_BR_OR:
  default:
  }
  if (ofs >= 0)  scm_jc_add_forward(jc, target, lab);
}

static void scm_jc_makeclos(SCM_JC *jc, int op)
{
  /*TOS = scm_vm3_make_closure(vm, env, sp, TOS); */
  jit_prepare(4);
  jit_pushr_i(VM_TOS);
  jit_pushr_i(VM_SP);
  jit_pushr_i(VM_ENV);
  jit_ldxi_l(VM_TMP, JIT_FP, 8);
  jit_pushr_i(VM_TMP);
  jit_finish(scm_vm3_make_closure);
  jit_movr_l(VM_TOS, JIT_RET);
  /*env = SCM_CLOSURE_ENV(TOS);*/
  jit_ldxi_l(VM_ENV, VM_TOS, 4);
}

static void scm_jc_letbody(SCM_JC *jc, int op)
{
  long lit = SCM_INUM(jc->icode[jc->icp++]);
  
  jit_subi_l	(VM_TMP,  VM_SP, (lit + 1) * sizeof(SOBJ));
  jit_str_p	(VM_TMP,  VM_ENV);
  jit_movi_p	(VM_TMP2, SCM_MKINUM(lit));
  jit_stxi_l	(4, 	  VM_TMP,   VM_TMP2);
  jit_movr_p	(VM_ENV,  VM_TMP);
}

static void scm_jc_letend(SCM_JC *jc, int op)
{
  long n = SCM_INUM(jc->icode[jc->icp++]);

  jit_subi_l	(VM_SP,  VM_SP,  (n +  ENV_HEADER_SLOTS) * sizeof(SOBJ));
  jit_ldr_p	(VM_ENV, VM_ENV);
}

static void scm_jc_letstar(SCM_JC *jc, int op)
{
  SOBJ ns = jc->icode[jc->icp++];
  int i;

  jit_str_l	(VM_SP,	VM_TOS);
  jit_stxi_l	(1 * sizeof(SOBJ), VM_SP, VM_ENV);
  jit_addi_l	(VM_ENV, VM_SP, sizeof(SOBJ));
  jit_movi_l	(VM_TOS, (long)ns);
  // jit_stxi_l	(2 * sizeof(SOBJ), VM_SP, VM_TOS);
  jit_addi_l	(VM_SP,	VM_SP, 2 * sizeof(SOBJ));
  jit_movi_l	(VM_TMP, scm_unbound);
  for (i = 1; i <= SCM_INUM(ns); i++) {
    jit_stxi_l	(i * sizeof(SOBJ), VM_SP, VM_TMP);
  }
}

static void scm_jc_letrec(SCM_JC *jc, int op)
{
  SOBJ ns = jc->icode[jc->icp++];
  int i;
  
  jit_str_l	(VM_SP, VM_TOS);
  jit_stxi_l	(1 * sizeof(SOBJ), VM_SP, VM_ENV);
  jit_addi_l	(VM_ENV, VM_SP, sizeof(SOBJ));
  jit_movi_l	(VM_TOS, (long)ns);
  jit_stxi_l	(2 * sizeof(SOBJ), VM_SP, VM_TOS);

  if (SCM_INUM(ns) > 0)	{	/* more than one slot */
    jit_movi_l	(VM_TOS, scm_unbound);
    
    for (i = 0; i < SCM_INUM(ns); i++) {
      jit_stxi_l    ((i + 3) * sizeof(SOBJ), VM_SP, VM_TOS);
    }
    jit_addi_l	(VM_SP,	VM_SP, (i + 2) * sizeof(SOBJ));
  } else {
    jit_addi_l 	(VM_SP,	VM_SP,  2 * sizeof(SOBJ));
  }
}


static void scm_jc_callcprim(SCM_JC *jc, int op)
{
  void *func = SCM_CPRIM_FUNC(jc->icode[jc->icp++]);
  int  nargs = SCM_INUM(jc->icode[jc->icp++]);
  
	jit_str_i	(VM_SP, VM_TOS);
	jit_subi_l	(VM_SP, VM_SP, (nargs - 1) * sizeof(SOBJ));
	jit_prepare	(2);
	jit_pusharg_l	(VM_SP);
	jit_movi_l	(VM_TMP, nargs);
	jit_pusharg_l	(VM_TMP);
	jit_finish	(func);
	jit_movr_l	(VM_TOS, JIT_RET);
}	

static void scm_jc_callcprim0(SCM_JC *jc, int op)
{
  void *func = SCM_CPRIM_FUNC(jc->icode[jc->icp++]);

	jit_str_i	(VM_SP, VM_TOS);		/* supdate() */
	jit_prepare	(0);
	jit_finish	(func);
	jit_addi_l	(VM_SP, VM_SP, sizeof(SOBJ));
	jit_movr_l	(VM_TOS, JIT_RET);
}

static void scm_jc_callcprim1(SCM_JC *jc, int op)
{
  void *func = SCM_CPRIM_FUNC(jc->icode[jc->icp++]);

	jit_str_i	(VM_SP, VM_TOS);		/* supdate() */
	jit_prepare	(1);
	jit_pusharg_l	(VM_TOS);
	jit_finish	(func);
	jit_movr_l	(VM_TOS, JIT_RET);
}

static void scm_jc_callcprim2(SCM_JC *jc, int op)
{
  void *func = SCM_CPRIM_FUNC(jc->icode[jc->icp++]);

       	jit_str_i	(VM_SP, VM_TOS);		/* supdate() */		

	jit_subi_l	(VM_SP, VM_SP, 1 * sizeof(SOBJ));
	jit_prepare	(2);
	jit_pusharg_l	(VM_TOS);
	jit_ldr_l	(VM_TMP,VM_SP);
	jit_pusharg_l	(VM_TMP);
	jit_finish	(func);
	jit_movr_l	(VM_TOS, JIT_RET);
}  

static void scm_jc_callcprim3(SCM_JC *jc, int op)
{
  void *func = SCM_CPRIM_FUNC(jc->icode[jc->icp++]);

       	jit_str_i	(VM_SP, VM_TOS);		/* supdate() */

	jit_subi_l	(VM_SP, VM_SP, 2 * sizeof(SOBJ));
	jit_prepare	(2);
	jit_pusharg_l	(VM_TOS);
	jit_ldxi_l	(VM_TMP,VM_SP, 1 * sizeof(SOBJ));
	jit_pusharg_l	(VM_TMP);
	jit_ldr_l		(VM_TMP,VM_SP);
	jit_pusharg_l	(VM_TMP);
	jit_finish	(func);
	jit_movr_l	(VM_TOS, JIT_RET);
}

static void scm_jc_callcprim4(SCM_JC *jc, int op)
{
  void *func = SCM_CPRIM_FUNC(jc->icode[jc->icp++]);

       	jit_str_i	(VM_SP, VM_TOS);		/* supdate() */

	jit_subi_l	(VM_SP, VM_SP, 3 * sizeof(SOBJ));
	jit_pusharg_l	(VM_TOS);
	jit_ldxi_l	(VM_TMP,VM_SP, 2 * sizeof(SOBJ));
	jit_pusharg_l	(VM_TMP);
	jit_ldxi_l	(VM_TMP,VM_SP, 1 * sizeof(SOBJ));
	jit_pusharg_l	(VM_TMP);
	jit_ldr_l	(VM_TMP,VM_SP);
	jit_pusharg_l	(VM_TMP);
	jit_finish	(func);
	jit_movr_l	(VM_TOS, JIT_RET);
}

static void scm_jc_end(SCM_JC *jc, int op)
{
  void *func = jc->prim_table[SCM_OP_END].start;

       jit_jmpi(func);
}

/********************************
 * Compiler for primitives
 ********************************/
static void scm_jc_add2(SCM_JC *jc, int op)
{
  jit_insn *r1, *r2, *r3;

	/* pop 2nd value */
	jit_subi_l	(VM_SP,		VM_SP,		sizeof(SOBJ));
	jit_ldr_l	(VM_TMP,	VM_SP);

	/* test for 2 inums */
	jit_andr_i	(VM_TMP2,	VM_TMP,		VM_TOS);
  r1 =	jit_bmsi_i	(jit_forward(),	VM_TMP2,	1);

  r2 =  jit_get_label	();
 	jit_prepare	(2);
	jit_pusharg_l	(VM_TOS);
	jit_ldr_l	(VM_TMP,	VM_SP);
	jit_pusharg_l	(VM_TMP);
	jit_finish	(scm_add2);
	jit_movr_l	(VM_TOS,	JIT_RET);
  r3 =	jit_jmpi	(jit_forward());

  jit_patch(r1);
	jit_andi_l	(VM_TMP,	VM_TMP,		SOBJ_INUM_MASK);
	jit_boaddr_l	(r2,		VM_TOS,		VM_TMP);

  jit_patch(r3);
}


static void scm_jc_plusn(SCM_JC *jc, int op, int n)
{
  jit_insn *ref1, *ref2, *ref3;
  
		 jit_movr_i		(VM_TMP, VM_TOS);
  ref1 = jit_bmsi_i		(jit_forward(), VM_TOS, 1); /* branch for inum */

  ref2 = jit_get_label	();
	   	 jit_prepare	(2);
		 jit_pusharg_l	(VM_TMP);
		 jit_movi_l		(VM_TMP, (long)SCM_MKINUM(n));
		 jit_pusharg_l	(VM_TMP);
		 jit_finish		(scm_add2);
	 	 jit_movr_i		(VM_TOS, JIT_RET);
  ref3 = jit_jmpi		(jit_forward());

  jit_patch(ref1);		/* add 1 */
	 jit_boaddi_l		(ref2, VM_TOS, n << SOBJ_INUM_SHIFT);

  jit_patch(ref3);
}

static void scm_jc_plus1(SCM_JC *jc, int op)
{
  scm_jc_plusn(jc, op, 1);
}

static void scm_jc_plus2(SCM_JC *jc, int op)
{
  scm_jc_plusn(jc, op, 2);
}

static void scm_jc_cmp_op(SCM_JC *jc, int op)
{
  jit_insn *ref1, *ref2, *ref3=NULL;
  void *func=NULL;

	jit_subi_i	(VM_SP,		VM_SP, 4);
	jit_ldr_l	(VM_TMP,	VM_SP);
	/* compare tmp with tos */
	jit_andr_i	(VM_TMP2,	VM_TMP,		VM_TOS);
  ref1= jit_bmsi_i	(jit_forward(), VM_TMP2, 1);

	jit_prepare		(2);
	jit_pusharg_l	(VM_TOS);
	jit_pusharg_l	(VM_TMP);
	switch(op) {
	case SCM_OP_LT2:	func = scm_lt2;		break;
	case SCM_OP_LE2:	func = scm_le2;		break;
	case SCM_OP_GT2:	func = scm_gt2;		break;
	case SCM_OP_GE2:	func = scm_ge2;		break;
	case SCM_OP_EQ2:	func = scm_eq2;		break;
	}
	jit_finish	(func);
	jit_movr_l	(VM_TOS, JIT_RET);

  ref2=	jit_jmpi	(jit_forward());

		/* test INUM */
  jit_patch(ref1);
  switch(op) {
  case SCM_OP_LT2: ref3 = jit_bltr_l(jit_forward(), VM_TMP, VM_TOS);  break;
  case SCM_OP_LE2: ref3 = jit_bler_l(jit_forward(), VM_TMP, VM_TOS);  break;
  case SCM_OP_GT2: ref3 = jit_bgtr_l(jit_forward(), VM_TMP, VM_TOS);  break;
  case SCM_OP_GE2: ref3 = jit_bger_l(jit_forward(), VM_TMP, VM_TOS);  break;
  case SCM_OP_EQ2: ref3 = jit_beqr_l(jit_forward(), VM_TMP, VM_TOS);  break;
  }
	jit_movi_l	(VM_TOS,	scm_false);
  jit_patch(ref2);
  jit_patch(ref3);		
}


/********************************
 * Native code assembler
 ********************************/

/* This assembler takes a VM code array as input and delivers a mcode
 * object (native code).
 */
SOBJ scm_vm3_engine_nasm(SOBJ code, SCM_VM3_PRIM *t)
{
  SOBJ *p, *l;
  SCM_DICT_ENTRY *de;
  SCM_VM3_PRIM *te;
  int nlits = 0, code_size = 0;
  int i, op;
  SCM_JC jc;
  void *regstart;		/* start of region for disassembler */
  
  if (!SCM_ARRAYP(code))    SCM_ERR("bad vector", code);

  /* init jc struct */
  jc.icode = SCM_ARRAY(code);
  jc.isize = SCM_ASIZE(code);
  jc.prim_table = t;

  /* First pass: determine size of code and number of literals */
  p = jc.icode;
  l = jc.icode + jc.isize;
  while(p < l) {
    SOBJ instr;
    instr = *p++;
    if ((de = scm_dict_search(scm_prim_addr_dict, instr)) == NULL) {
      fprintf(stderr, "NASM: opcode %p not found in prim_addr_dict\n", instr);
      SCM_ERR("instr not found", instr);
    }
    te = de->value;
    fprintf(stderr, "NASM:counting: %s, start=%p size=%d follow=%d\n",
	    te->name, te->start, te->end - te->start, te->follow);
    code_size += te->end - te->start;
    nlits += te->follow;
    p += te->follow;
  }

  /* Print some stats */
  fprintf(stderr, "NASM: estimated code size=%d\n", code_size);
  fprintf(stderr, "NASM: number of literals=%d\n",  nlits);
    
  /* Next pass: generate code */
  
  /* init mcode structure */
  jc.mcode 	    = scm_must_alloc(sizeof(SCM_MCode));
  jc.mcode->code    = scm_must_alloc(code_size);
  jc.mcode->csize   = code_size;
  jc.mcode->lit	    = scm_must_alloc(nlits * sizeof(SOBJ));
  jc.mcode->nlits   = 0;

  jc.ncp   	= jc.mcode->code;
  jc.fwr 	= scm_must_alloc(nlits * sizeof(SCM_JC_FWR));
  jc.nfwr 	= 0;
  jc.ref 	= scm_must_alloc(SCM_ASIZE(code) * sizeof(void*));

  p = SCM_ARRAY(code);
  i = 0;
  while(i < SCM_ASIZE(code)) {

    jc.ref[i] = jc.ncp;
    regstart  = jc.ncp;
    (void)jit_set_ip(jc.ncp);

    scm_jc_try_forward_resolv(&jc, i);

    te = scm_dict_search(scm_prim_addr_dict,p[i++])->value;
    op = te - t;

    if (te->jcomp) {
      regstart = jc.ncp;
      /* got compiler for this op */
      fprintf(stderr, "NASM:gen:%p %2d %s\tjcomp\n",
	      jc.ncp, te->end - te->start, te->name);
      jc.icp = i;
      (*te->jcomp)(&jc, op);
      jc.ncp = jit_get_ip().ptr;
      i = jc.icp;
      continue;
    }
    fprintf(stderr, "NASM:gen:%p %2d %s\tcopy\n",
	    jc.ncp, te->end - te->start, te->name);

    /* regular operation: just copy code */
    if (te->follow != 0) {	/* have following args */
      if (te->follow != 1) {
	fprintf(stderr, "NASM:%s:oops more than one follow\n", te->name);
      	SCM_ERR("too much follower", NULL);
      }
      /* Save literal and compile code to set VM_IP at runtime to this
       * address */

      jc.mcode->lit[jc.mcode->nlits] = p[i];
      jit_movi_i(VM_IP,	 jc.mcode->lit + jc.mcode->nlits);
      i++;
      jc.mcode->nlits++;
      jc.ncp = jit_get_ip().ptr;
    }
    scm_jc_inline_op(&jc, op);
  }
  jc.mcode->csize = jc.ncp - jc.mcode->code;
  fprintf(stderr, "NASM:code size: bytecode=%ld estimated=%d real=%d\n",
	  SCM_ASIZE(code) * sizeof(SOBJ), code_size, jc.mcode->csize);
  scm_disassemble_region(jc.mcode->code, jc.mcode->csize);
  fprintf(stderr, "LITERALS: nlits=%d\n", jc.mcode->nlits);
  for (i = 0; i < jc.mcode->nlits; i++) {
    fprintf(stderr, "\tlit[%d] = %p\n", i, jc.mcode->lit[i]);
  }
  scm_free(jc.ref);
  return(scm_mkmcode(jc.mcode));
}

void scm_jc_install_prim_compiler(SCM_VM3_PRIM *t)
{
  /* JIT: declare primitive that should not be copied but should be
   * (tricky) called */

  t[SCM_OP_END].jcomp		= scm_jc_end;
  t[SCM_OP_PUSHLIT].jcomp	= scm_jc_pushlit;
  t[SCM_OP_PUSHSYM].jcomp	= scm_jc_pushsym;
  t[SCM_OP_PUSHLOC0].jcomp	= scm_jc_pushloc0;
  t[SCM_OP_SETSYM].jcomp	= scm_jc_setsym;
  t[SCM_OP_SETLOC0].jcomp	= scm_jc_setloc0;

  t[SCM_OP_BRA].jcomp		= scm_jc_branch;
  t[SCM_OP_BRF].jcomp		= scm_jc_branch;
  t[SCM_OP_BRT].jcomp		= scm_jc_branch;
  t[SCM_OP_BR_AND].jcomp	= scm_jc_branch;
  t[SCM_OP_BR_OR].jcomp		= scm_jc_branch;

  t[SCM_OP_LETBODY].jcomp	= scm_jc_letbody;
  t[SCM_OP_LETEND].jcomp	= scm_jc_letend;
  t[SCM_OP_LETSTAR].jcomp	= scm_jc_letstar;
  t[SCM_OP_LETREC].jcomp	= scm_jc_letrec;

  t[SCM_OP_MAKECLOS].jcomp	= scm_jc_makeclos;
  t[SCM_OP_CALLCPRIM].jcomp	= scm_jc_callcprim;
  t[SCM_OP_CALLCPRIM0].jcomp	= scm_jc_callcprim0;
  t[SCM_OP_CALLCPRIM1].jcomp	= scm_jc_callcprim1;
  t[SCM_OP_CALLCPRIM2].jcomp	= scm_jc_callcprim2;
  t[SCM_OP_CALLCPRIM3].jcomp	= scm_jc_callcprim3;
  t[SCM_OP_CALLCPRIM4].jcomp	= scm_jc_callcprim4;

  t[SCM_OP_CALL].jcomp		= scm_jc_call;

  t[SCM_OP_CONS].jcomp 		= scm_jc_callprim;
  t[SCM_OP_LIST2].jcomp 	= scm_jc_callprim;

  t[SCM_OP_ADD2].jcomp		= scm_jc_add2;

  t[SCM_OP_LT2].jcomp		= scm_jc_cmp_op;
  t[SCM_OP_LE2].jcomp		= scm_jc_cmp_op;
  t[SCM_OP_GT2].jcomp		= scm_jc_cmp_op;
  t[SCM_OP_GE2].jcomp		= scm_jc_cmp_op;
  t[SCM_OP_EQ2].jcomp		= scm_jc_cmp_op;

  t[SCM_OP_PLUS1].jcomp		= scm_jc_plus1;
  t[SCM_OP_PLUS2].jcomp		= scm_jc_plus2;
}
