/* -*- c -*- */
#define Prim(name,str,nargs)	l_##name##: /* str */
#define PrimVarargs(name,str)	l_##name##: /* str */

#define requires(x,t,msg) if(SCM_INUMP(x)||((SOBJ)x)->type!=t){SCM_ERR(msg,x);}

Prim(pairp, "pair?", 1)   	/* obj -- bool */
{
  TOS = (!SCM_INUMP(TOS) && SCM_PAIRP((SOBJ)TOS)) ?	scm_true: scm_false;
  NEXT;
}

Prim(cons, "cons", 2) 		/* n2 n1 -- obj */
{
  SOBJ new = scm_newcell(SOBJ_T_PAIR);
  spop(SCM_CAR(new));
  SCM_CDR(new) = TOS;
  TOS = new;
  NEXT;
}

Prim(car, "car", 1)		/* obj -- obj */
{
  requires(TOS, SOBJ_T_PAIR, "car: wrong arg type");
  TOS = SCM_CAR((SOBJ)TOS);
  NEXT;
}

Prim(cdr, "cdr", 1)		/* obj -- obj */
{
  requires(TOS, SOBJ_T_PAIR, "cdr: wrong arg type");
  TOS = SCM_CDR((SOBJ)TOS);
  NEXT;
}

Prim(setcar, "set-car!", 2) 	/* value obj -- undefined */
{
  SOBJ sym;
  requires(TOS, SOBJ_T_PAIR, "set-car!: wrong arg type");
  spop(sym);
  SCM_CAR(sym) = TOS;
  TOS = scm_undefined;
  NEXT;
}

Prim(setcdr, "set-cdr!", 2) 	/* value obj -- undefined */
{
  SOBJ sym;
  requires(TOS, SOBJ_T_PAIR, "set-cdr!: wrong arg type");
  spop(sym);
  SCM_CDR(sym) = TOS;
  TOS = scm_undefined;
  NEXT;
}

Prim(nullp, "null?", 1)		/* obj -- #t | #f */
{
  TOS = SCM_NULLP(TOS) ? scm_true : scm_false;
  NEXT;
}

Prim(listp, "list?", 1)		/* obj -- len */
{
  TOS = (scm_list_length(TOS) >= 0) ? scm_true : scm_false;
  NEXT;
}

PrimVarargs(list, "list")	/* objn .. obj2 obj1 -- (obj1 obj2 .. objn) */
{
  SOBJ list = NULL;
  Label *l;
  rpop(rsp);
  sp[0] = TOS;
  for (l = rsp; l >= sp; l--)  list = scm_cons(*l, list);
  sp = rsp;  TOS = list;
  NEXT;
}

Prim(length, "length", 1) 		/* obj -- n */
{
  int len = scm_list_length(TOS);
  if (len >= 0) {
	TOS = SCM_MKINUM(len);
	NEXT;
  }
  SCM_ERR("length: cannot calculate", TOS);
}

Prim(append, "append", 2)		/* l2 l1 -- (l1 l2) */
{
  SOBJ l1, new, last, list;
  spop(l1);

  if (l1 == NULL) {			/* l2 is TOS*/
	   NEXT;
  }
  if (!SCM_PAIRP(l1)) {  SCM_ERR("append: argument is not a list", l1); }

  list = scm_newcell(SOBJ_T_PAIR);
  new = list;
  last = NULL;
  while(1) {
	SCM_CAR(new) = SCM_CAR(l1);
	if (last) { SCM_CDR(last) = new; }
	last = new;
	l1 = SCM_CDR(l1);
	if (l1 == NULL || ! SCM_PAIRP(l1)) break;
	new = scm_newcell(SOBJ_T_PAIR);
  }
  SCM_CDR(new) = TOS;
  TOS = list;
  NEXT;
}

Prim(reverse, "reverse", 1)		/* l -- l */
{
  SOBJ l = TOS;
  SOBJ n = NULL;
  while(l) {
	if (!SCM_PAIRP(l))	SCM_ERR("reverse: bad list", l);
	n = scm_cons(SCM_CAR(l), n);
	l = SCM_CDR(l);
  }
  TOS = n;
  NEXT;
}

Prim(list_tail, "list-tail", 2) 	/* n list -- list */
{
  SOBJ l;
  int n;

  spop(l);	if (!SCM_PAIRP(l)) SCM_ERR("list-tail: bad list", l);
  if (!SCM_INUMP(TOS)) SCM_ERR("list-tail: bad index", TOS);
  n = SCM_INUM(TOS);
  while(n > 0) {
	l = SCM_CDR(l);
	if (l == NULL) SCM_ERR("list-tail: list too short", l);
	n--;
  }
  TOS = l;
  NEXT;
}

Prim(plus, "+", 2)				/* n2 n1 -- n1+n2 */
{
  SOBJ n1;
  spop(n1);

#ifdef INUM_OPTIMIZATION
  if (SCM_INUMP((long)n1 & (long)TOS)) {
	long r = SCM_INUM(n1) + SCM_INUM(TOS);
	if (SCM_INUM_RANGE(r)) {  TOS = SCM_MKINUM(r);  NEXT; }
	TOS = scm_int2bnum(r);  NEXT;

#ifdef COMMENT	
	/* this seems complex, but it's quite the fastest way to add and
     * detect overflow */
	long a, b, r;

	a = (long)n1;
	b = (long)TOS & SOBJ_INUM_MASK;
	r = a + b;
	if (SCM_INUM(b) == SCM_INUM(r) - SCM_INUM(a)) { TOS = (void*)r;  NEXT; }
	TOS = scm_int2bnum(SCM_INUM(a) + SCM_INUM(b));
	NEXT;
#endif
  }
#endif
  TOS = scm_add2(n1, TOS);
  NEXT;
}

Prim(minus, "-", 2)		/* n2 n1 -- n1-n2 */
{
  SOBJ n1;

  spop(n1);
#ifdef INUM_OPTIMIZATION
  if (SCM_INUMP(n1) && SCM_INUMP(TOS)) {
	long r;
	r = SCM_INUM(n1) - SCM_INUM(TOS);
	if (SCM_INUM_RANGE(r)) {  TOS = SCM_MKINUM(r);  NEXT; }
	TOS = scm_int2bnum(r); NEXT;
  }
#endif
  TOS = scm_sub2(n1, TOS);
  NEXT;
}

Prim(mul, "*", 2)			/* n2 n1 -- n1*n2 */
{
  SOBJ n1;

  spop(n1);
  if (SCM_INUMP(n1) && SCM_INUMP(TOS)) {
	long r = SCM_INUM(n1) * SCM_INUM(TOS);
	if (SCM_INUM(n1) == r / SCM_INUM(TOS)) {
	  TOS = SCM_MKINUM(r);
	  NEXT;
	}
  }
  TOS = scm_mul2(n1, TOS);
  NEXT;
}

Prim(div, "/", 2)			/* n2 n1 -- n1/n2 */
{
  SOBJ n1;
  spop(n1);
  TOS = scm_div2(n1, TOS);
  NEXT;
}

#ifdef INUM_OPTIMIZATION
#define OPTIM_LOGOP(op) \
if (SCM_INUMP((long)n1 & (long)TOS)) { \
  if ((long)n1 op (long)TOS) { TOS=scm_true; NEXT; } TOS=scm_false; NEXT; }

#else
#define OPTIM_LOGOP(op)
#endif

Prim(lt, "<", 2)			/* n2 n1 -- flag */
{
  SOBJ n1;  spop(n1);  OPTIM_LOGOP(<);   TOS = scm_lt(n1, TOS);  NEXT;
}

Prim(le, "<=", 2)			/* n2 n1 -- flag */
{
  SOBJ n1;  spop(n1);  OPTIM_LOGOP(<=);  TOS = scm_le(n1, TOS);  NEXT;
}

Prim(ge, ">=", 2)			/* n2 n1 -- flag */
{
  SOBJ n1;  spop(n1);  OPTIM_LOGOP(>=);  TOS = scm_ge(n1, TOS);  NEXT;
}

Prim(gt, ">", 2)			/* n2 n1 -- flag */
{
  SOBJ n1;  spop(n1);  OPTIM_LOGOP(>);  TOS = scm_gt(n1, TOS);  NEXT;
}


Prim(plus1, "add1", 1)
{
  if (SCM_INUMP(TOS) && (SCM_INUM(TOS) < SOBJ_INUM_MAX)) {
	(long)TOS += (1 << SOBJ_INUM_SHIFT);
	NEXT;
  }
  TOS = scm_add2(SCM_INUM1, TOS);
  NEXT;
}

Prim(plus2, "add2", 1)
{
  if (SCM_INUMP(TOS) && (SCM_INUM(TOS) < SOBJ_INUM_MAX)) {
	(long)TOS += (2 << SOBJ_INUM_SHIFT);
	NEXT;
  }
  TOS = scm_add2(SCM_MKINUM(2), TOS);
  NEXT;
}

Prim(minus1, "sub1", 1)
{
  if (SCM_INUMP(TOS) && (SCM_INUM(TOS) > SOBJ_INUM_MIN)) {
	(long)TOS += (-1 << SOBJ_INUM_SHIFT);
	NEXT;
  }
  TOS = scm_sub2(TOS, SCM_INUM1);
  NEXT;
}

Prim(minus2, "sub2", 1)
{
  if (SCM_INUMP(TOS) && (SCM_INUM(TOS) > SOBJ_INUM_MIN)) {
	(long)TOS += (-2 << SOBJ_INUM_SHIFT);
	NEXT;
  }
  TOS = scm_sub2(TOS, SCM_MKINUM(2));
  NEXT;
}

/*-- IO */
PrimVarargs(display, "display") 	/* [ port ] string -- undef */
{
  rpop(rsp);
  TOS = (rsp < sp) ? scm_undefined :
	scm_display(TOS, (rsp > sp) ? sp[1]:NULL);
  sp = rsp;
  NEXT;
}

PrimVarargs(print, "print") 	/* [ port ] string -- undef */
{
  rpop(rsp);
  TOS = (rsp < sp) ? scm_undefined :
	scm_print(TOS, (rsp > sp) ? sp[1]:NULL);
  sp = rsp;
  NEXT;
}

PrimVarargs(write, "write") 	/* [ port ] string -- undef */
{
  rpop(rsp);
  TOS = (rsp < sp) ? scm_undefined :
	scm_write(TOS, (rsp > sp) ? sp[1]:NULL);
  sp = rsp;
  NEXT;
}

PrimVarargs(newline, "newline") 	/* [ port ] -- undef */
{
  rpop(rsp);
  TOS = scm_newline( (rsp < sp) ? NULL : TOS);
  sp = rsp;
  NEXT;
}

/****************************************************************
 * boolean
 ****************************************************************/
Prim(not, "not", 1)
{
  TOS = (TOS == scm_false) ? scm_true : scm_false;  NEXT;
}

Prim(booleanp, "boolean?", 1)
{
  TOS = SCM_BOOLEANP( (SOBJ) TOS) ? scm_true : scm_false;  NEXT;
}

Prim(eqp, "eq?", 2)
{
  SOBJ n;  spop(n);
  TOS = (TOS == n) ? scm_true : scm_false;  NEXT;
}

/****************************************************************
 * number
 ****************************************************************/

/****************************************************************
 * Control
 ****************************************************************/
Prim(compile, "compile", 1)
{
  sresync();
  TOS=scm_compile(TOS);
  NEXT;
}

Prim(execute, "execute", 1)
{
  SOBJ code;
  spop(code);
  if (!SCM_CODEP(code)) SCM_ERR("execute: bad code", code);
  
  rpush(sp-1);					/* mark */
  rpush(ip);
  rpush(rfp); rfp = rp;
  ip = (Label*)SCM_CODE_CODE(code);
  NEXT;
}

Prim(eval, "eval", 1)
{
  sresync();
  TOS = scm_compile(TOS);
  goto l_execute;
}

Prim(apply, "apply", 2)			/* ds: list func -- result */
{
  SOBJ func, list, obj;
  SOBJ *ptr;

  spop(func);
  spop(list);
  rpush(sp-1);					/* push the mark */
  
  /*-- push list on the rstack then pop rstack to dstack*/
  ptr = rp;
  while(list) {
	if (!SCM_PAIRP(list)) SCM_ERR("apply: bad list", list);
	rpush(SCM_CAR(list));
	list = SCM_CDR(list);
  }
  while(rp < ptr) {
	rpop(obj);  spush(obj);
  }
  spush(func);
  goto l_callv;
}



Prim(call_cc, "call/cc", 1) 	/* ds: proc -- result */
{
  printf("call/cc: ");
  scm_display(TOS, NULL);
  NEXT;
}
