/* -*- mode:c; tab-width:8; -*- */

/****************************************************************
 * cell && list
 ****************************************************************/
PRIM(pairp, "pair?", 1)   	/* obj -- bool */
{
  RETURN(SCM_MKBOOL(TOS != NULL && SCM_PAIRP(TOS)));
}
END_PRIM(pairp);

PRIM(cons, "cons", 2) 		/* n2 n1 -- obj */
{
  SOBJ new;
  supdate();
  new = scm_newcell(SOBJ_T_PAIR);
  TOS = *sp;
  spop(SCM_CDR(new));
  SCM_CAR(new) = TOS;
  RETURN(new);
}
END_PRIM(cons);

/*E* (list2 OBJ1 OBJ2) => LIST */
/*D* Returns a newly allocated list of 2 elements. */
PRIM(list2, "list2", 2)
{
  SOBJ obj1, obj2;
  spop(obj2);
  obj1 = TOS;
  RETURN(SCM_LIST2(obj1, obj2));
}
END_PRIM(list2);

/*S* (car OBJ) => OBJ */
/*D* Return the car of a list */
PRIM(car, "car", 1)
{
  REQUIRES(TOS, SOBJ_T_PAIR, "car: not a pair");
  RETURN(SCM_CAR(TOS));
}
END_PRIM(car);

/*E* (cdr OBJ) => OBJ */
PRIM(cdr, "cdr", 1)
{
  REQUIRES(TOS, SOBJ_T_PAIR, "cdr: not a pair");
  RETURN(SCM_CDR(TOS));
}
END_PRIM(cdr);

/*S* (set-car! PAIR OBJ) => #undefined */
/*D* Change the car field of the PAIR to OBJ. */
PRIM(setcar, "set-car!", 2) 	/* value obj -- undefined */
{
  SOBJ value;
  spop(value);
  REQUIRES(TOS, SOBJ_T_PAIR, "set-car! wrong arg type");
  SCM_CAR(TOS) = value;
  RETURN(scm_undefined);
}
END_PRIM(setcar);

/*S* (set-cdr! PAIR OBJ) => #undefined */
/*D* Change the cdr field of the PAIR to OBJ. */
PRIM(setcdr, "set-cdr!", 2) 	/* value obj -- undefined */
{
  SOBJ value;
  spop(value);
  REQUIRES(TOS, SOBJ_T_PAIR, "set-cdr! wrong arg type");
  SCM_CDR(TOS) = value;
  RETURN(scm_undefined);
}
END_PRIM(setcdr);

PRIM(nullp, "null?", 1) /* obj -- #t | #f */
{
  RETURN(SCM_MKBOOL(SCM_NULLP(TOS)));
}
END_PRIM(nullp);

PRIM(listp, "list?", 1)		/* obj -- len */
{
  RETURN(SCM_MKBOOL(scm_list_length(TOS) >= 0));
}
END_PRIM(listp);

PRIMV(list, "list")	/* objn .. obj2 obj1 -- (obj1 obj2 .. objn) */
{
  SOBJ list;
  SOBJ *base = sp - SCM_INUM(*ip++);

  supdate();
  for (list = NULL; sp > base; sp--) list = scm_cons(*sp, list);
  sp++;
  TOS = list;
}
END_PRIM(list);


PRIM(length, "length", 1) 		/* obj -- n */
{
  int len;
  if ((len = scm_list_length(TOS)) < 0)
	SCM_ERR("length: cannot calculate", TOS);
  RETURN(SCM_MKINUM(len));
}
END_PRIM(length);

/*E* (nth ELT LIST) => OBJ */
PRIM(nth, "nth", 2)				/* l n -- obj */
{
  SOBJ list;
  int n;
  spop(list);
  if (!SCM_INUMP(TOS))	SCM_ERR("nth: bad number", TOS);
  n = SCM_INUM(TOS);
  while(--n >= 0 && list && SCM_PAIRP(list))
	list = SCM_CDR(list);
  if (!SCM_PAIRP(list)) SCM_ERR("nth: bad list", list);
  RETURN(SCM_CAR(list));
}
END_PRIM(nth);

/*-- need by macro */
/*E* (qq-append2 L1 L2) => (L1 L2) */
PRIM(qq_append2, "qq-append2", 2)	/* l2 l1 -- (l1 l2) */
{
  SOBJ l1, l2, list, *lp, new;
  spop(l2);
  l1 = TOS;
  list = NULL; 	lp = &list;
  if (l1 != NULL && l2 != NULL) {
	while(l1 && SCM_PAIRP(l1)) {
	  new = scm_newcell(SOBJ_T_PAIR);
	  SCM_CAR(new) = SCM_CAR(l1);
	  *lp = new;  lp = &SCM_CDR(new);
	  l1 = SCM_CDR(l1);
	}
  }
  *lp = l2;
  RETURN(list);
}
END_PRIM(qq_append2);

/*-- IO */
PRIMV(display, "display") 	/* string [ port ]-- undef */
{
  int nargs = SCM_INUM(*ip++);
  SOBJ port = NULL;
  if (nargs > 1)  spop(port);
  TOS = scm_display2(TOS, port);
}
END_PRIM(display);

PRIMV(print, "print") 	/* [ port ] string -- undef */
{
  int nargs = SCM_INUM(*ip++);
  SOBJ port = NULL;
  if (nargs > 1)  spop(port);
  TOS = (*scm_vm3_print2)(TOS,port);
  /* TOS = scm_print2(TOS, port); */
}
END_PRIM(print);

PRIMV(write, "write") 	/* [ port ] string -- undef */
{
  int nargs = SCM_INUM(*ip++);
  SOBJ port = NULL;
  if (nargs > 1)  spop(port);
  TOS = scm_write2(TOS, port);
}
END_PRIM(write);

/*S* (newline [PORT]) => #undefined */
/*D* Output a newline char on port PORT. If no PORT argument is given,
 * newline is sended to current-output-port */
PRIMV(newline, "newline") 	/* [ port ] -- undef */
{
  int nargs = SCM_INUM(*ip++);
  SOBJ port = NULL;
  if (nargs > 1)  spop(port);
  spush(scm_newline1(port));
}
END_PRIM(newline);

/****************************************************************
 * boolean
 ****************************************************************/
/*S* (not OBJ) => BOOL */
/*D* Returns #t if OBJ is #f, #f otherwise */
PRIM(not, "not", 1)
{
  RETURN(SCM_MKBOOL(TOS == scm_false));
}
END_PRIM(not);

/*S* (boolean? OBJ) => BOOL */
/*D* Returns #t if OBJ is either #t or #f. Otherwise #f is returned. */
PRIM(booleanp, "boolean?", 1)
{
  RETURN(SCM_MKBOOL(SCM_BOOLEANP(TOS)));
}
END_PRIM(booleanp);

/*S* (eq? OBJ1 OBJ2) => BOOL */
/*D* Returns #t if OBJ1 and OBJ2 refer to same scheme object. */
PRIM(eqp, "eq?", 2)
{
  SOBJ n;  spop(n);
  RETURN(SCM_MKBOOL(n == TOS));
}
END_PRIM(eqp);

/****************************************************************
 * control
 ****************************************************************/
/*PRIMV(compile, "compile")
{
  sresync();
  VRETURN(scm_compile(TOS, (NARGS > 1) ? sp[1] : NULL));
}
*/

/*PRIMV(eval, "eval")
{
  sresync();
  TOS = scm_compile(TOS, (NARGS > 1) ? sp[1] : NULL);
  cont->ip = ip; cont->env = env;
  ip = SCM_CODE_CODE(TOS);
  NEXT;
}
*/

/*PRIMV(neval, "neval")
{
  sresync();
#ifdef OLD_COMPILER
  TOS = scm_compile2(TOS, (NARGS > 1) ? sp[1] : NULL);
#else
  TOS = scm_compile(TOS, (NARGS > 1) ? sp[1] : NULL);
#endif
  cont->ip = ip; cont->env = env;
  ip = SCM_CODE_CODE(TOS);
  NEXT;
}
*/

/*PRIM(execute, "execute", 1)
{
  SOBJ code;
  spop(code);
  if (!SCM_CODEP(code)) SCM_ERR("execute: bad code", code);
  
  PUSH_CONT();
  cont->ip = ip;
  ip = (SOBJ*)SCM_CODE_CODE(code);
  NEXT;
}
*/
/*E* (load FILE) => OBJ */
/*D* Interpret the content of the file which name is given in
 * STR. Returns the value of the last evaluated expression */
PRIM(load, "load", 1)
{
  sresync();
  RETURN(scm_load(TOS));
}
END_PRIM(load);

#ifdef OLD
/*S* (apply PROC ARG1 ARG2 ... ARGS) => VALUE*/
/*D* Build an argument list such as (append (list arg1 arg2 ...) args)
 * and call proc with this list as argument */
/*PRIMV(apply, "apply") */
{
  SOBJ list;
  int len, nargs;
  
  spop(proc);

  if (SCM_MACROP(proc)) {
	if (SCM_MACRO_FUNC(proc) == NULL)
	  SCM_ERR("apply: cannot apply to macro", proc);
	proc = SCM_MACRO_FUNC(proc);
  }

  if (NARGS < 1) SCM_ERR("apply: needs at least 2 arguments", proc);

  nargs = NARGS-1;
  
  list = sp[nargs]; /* list should be last */
  if (list) { 
	len = scm_list_length(list);
	if (len < 0)  SCM_ERR("apply: bad argument list", list);

	if (len == 1) {				/* only one element : replace list by elt */
	  sp[nargs] = SCM_CAR(list);
	} else {					/* more than one element */
	  scm_vm_move_stack(sp - (len - 1), sp, nargs);
	  sp -= (len - 1);
	  while(list) {
		sp[nargs++] = SCM_CAR(list);
		list = SCM_CDR(list);
	  }
	}
  } else {						/* pop list */
	spop(list);
  }
  TOS = *sp;
  spush(proc);
  goto l_call;
}

/*PRIM(engine, "engine", 1) */
{
  scm_cprint(TOS);
  RETURN(TOS);
}
#endif
/****************************************************************
 * Symbol
 ****************************************************************/
/*S* (symbol? OBJ) => BOOL */
/*D* Returns #t if OBJ is a symbol, #f otherwise */
PRIM(symbolp, "symbol?", 1)
{
  RETURN(SCM_MKBOOL((SCM_SYMBOLP(TOS) || SCM_ATOMP(TOS))));
}
END_PRIM(symbolp);

/*E* (pure-symbol? OBJ) => BOOL */
/*D* Returns #t if OBJ is a pure symbol, #f otherwise. Pure symbols
 * are binding a name with a value. Quoted symbols are not pure
 * symbols, they are atoms. */
PRIM(pure_symbolp, "pure-symbol?", 1)
{
  RETURN(SCM_MKBOOL(SCM_SYMBOLP(TOS)));
}
END_PRIM(pure_symbolp);

/****************************************************************
 * Keyword
 ****************************************************************/
/*E* (keyword? OBJ) => BOOL */
/*D* Returns #t if OBJ is a keyword, #f otherwise */
PRIM(keywordp, "keyword?", 1)
{
  RETURN(SCM_MKBOOL(SCM_KEYWORDP(TOS)));
}
END_PRIM(keywordp);

/*E* (keyword->string KEYW) => STR */
/*D* Convert a keyword to a string representation */
PRIM(keyw2str, "keyword->string", 1)
{
  RETURN(scm_keyword_to_string(TOS));
}
END_PRIM(keyw2str);

/*E* (string->keyword STR) => KEYWORD */
/*D* Returns the keyword corresponding to the string STR. */

PRIM(str2keyw, "string->keyword", 1)
{
  RETURN(scm_string_to_keyword(TOS));
}
END_PRIM(str2keyw);

/*E* (get-keyword KEYW LIST DEFAULT) => VALUE */
/*D* Search KEYW in the LIST. Returns the value following the keyword
 * or DEFAULT if not found. */
PRIMV(getkeyw, "get-keyword")
{
  SOBJ keyw, list, deflt = scm_false;
  int nargs = SCM_INUM(*ip++);
  if (nargs < 2) SCM_ERR("get-keyword: bad number of args", NULL);
  if (nargs >= 3) {
    spop(deflt);
  }
  spop(list);
  spop(keyw);
  spush(scm_get_keyword(keyw,list,deflt));
}
END_PRIM(getkeyw);

/****************************************************************
 * Misc predicates
 ****************************************************************/

/*E* (atom? OBJ) => BOOL */
/*D* Returns #t if OBJ is an atom, #f otherwise */
PRIM(atomp, "atom?", 1)
{
  RETURN(SCM_MKBOOL(SCM_ATOMP(TOS)));
}
END_PRIM(atomp);

/*E* (undefined-object? OBJ) => BOOL */
/*D* Returns #t if OBJ is undefined, #f otherwise */
PRIM(undefinedp, "undefined-object?", 1)
{
  RETURN(SCM_MKBOOL(SCM_OBJTYPE(TOS) == SOBJ_T_UNDEFINED));
}
END_PRIM(undefinedp);

/*E* (unbound-object? OBJ) => BOOL */
/*D* Returns #t if OBJ is unbound, #f otherwise */
PRIM(unboundp, "unbound-object?", 1)
{
  RETURN(SCM_MKBOOL(SCM_OBJTYPE(TOS) == SOBJ_T_UNBOUND));
}
END_PRIM(unboundp);

/*E* (macro? OBJ) => BOOL */
/*D* Returns #t if OBJ is a macro, #f otherwise */
PRIM(macrop, "macro?", 1)
{
  RETURN(SCM_MKBOOL(SCM_MACROP(TOS)));
}
END_PRIM(macrop);

/****************************************************************
 * number
 ****************************************************************/
/*S* (number? OBJ) => BOOLEAN */
/*D* Returns #t if OBJ is a number, #f otherwise */
PRIM(numberp, "number?", 1)
{
  RETURN( SCM_MKBOOL(SCM_NUMBERP(TOS)) );
}
END_PRIM(numberp);

/*S* (integer? OBJ) => BOOLEAN */
/*D* Returns #t if OBJ is an integer number, #f otherwise */
PRIM(integerp, "integer?", 1)
{
  if (SCM_INUMP(TOS) || SCM_BNUMP(TOS)) {
    TOS = scm_true;
  } else if (SCM_FNUMP(TOS)) {
    double x = SCM_FNUM(TOS);
    TOS = SCM_MKBOOL( floor(x) == x );
  } else {
    TOS = scm_false;
  }
}
END_PRIM(integerp);

/*S* (real? OBJ) => BOOLEAN */
/*D* Returns #t if OBJ is an real number, #f otherwise.*/
PRIM(realp, "real?", 1)
{
  RETURN(SCM_MKBOOL(SCM_NUMBERP(TOS)));
}
END_PRIM(realp);

/*S* (complex? OBJ) => BOOLEAN */
/*D* Returns #t if OBJ is an complex number, #f otherwise.*/
PRIM(complexp, "complex?", 1)
{
  RETURN(SCM_MKBOOL(SCM_NUMBERP(TOS)));
}
END_PRIM(complexp);

/*S* (rational? OBJ) => BOOLEAN */
/*D* Returns #t if OBJ is an rational number, #f otherwise.*/
PRIM(rationalp, "rational?", 1)
{
  RETURN(SCM_MKBOOL(SCM_NUMBERP(TOS)));
}
END_PRIM(rationalp);

/*S* (exact? OBJ) => BOOLEAN */
/*D* Returns #t if OBJ is an exact number, #f otherwise.*/
PRIM(exactp, "exact?", 1)
{
  RETURN(SCM_MKBOOL( SCM_INUMP(TOS) || SCM_BNUMP(TOS) ));
}
END_PRIM(exactp);

/*S* (inexact? OBJ) => BOOLEAN */
/*D* Returns #t if OBJ is an inexact number, #f otherwise.*/
PRIM(inexactp, "inexact?", 1)
{
  RETURN(SCM_MKBOOL( SCM_FNUMP(TOS) ));
}
END_PRIM(inexactp);

/* Inline for logical operation */
#define LOGOP(op, scm_func) 					\
  SOBJ n1; 							\
  spop(n1); 							\
  if (SCM_INUMP((long)n1 & (long)TOS)) {			\
	TOS = ((long)TOS op (long)n1) ? scm_true : scm_false; 	\
  } else { 							\
	TOS = scm_func(TOS, n1); 				\
  }

PRIM(lt2, "*i-n2<*", 2) {  LOGOP(<, scm_lt2); }
END_PRIM(lt2);

PRIM(le2, "*i-n2<=*", 2) {  LOGOP(<=, scm_le2); }
END_PRIM(le2);

PRIM(gt2, "*i-n2>*", 2) {  LOGOP(>, scm_gt2); }
END_PRIM(gt2);

PRIM(ge2, "*i-n2>=*", 2) {  LOGOP(>=, scm_ge2); }
END_PRIM(ge2);

PRIM(eq2, "*i-n2=*", 2) {  LOGOP(==, scm_eq2); }
END_PRIM(eq2);

#undef LOGOP

/* generic logical operation */
#define LOGOP(op)						\
  int nargs = SCM_INUM(*ip++);					\
  SOBJ *base, *p;						\
  base = sp - nargs; base++;  p = base;				\
  supdate();							\
  for (TOS = *p++; p <= sp && scm_cmpnum(TOS, *p) op 0; TOS = *p++) ;	\
  TOS = (p > sp) ? scm_true : scm_false;			\
  sp = base;

PRIMV(ltv, "*i-nv<*")  { LOGOP(<); }
END_PRIM(ltv);

PRIMV(lev, "*i-nv<=*") { LOGOP(<=); }
END_PRIM(lev);

PRIMV(gev, "*i-nv>=*") { LOGOP(>=); }
END_PRIM(gev);

PRIMV(gtv, "*i-nv>*")  { LOGOP(>); }
END_PRIM(gtv);

PRIMV(eqv, "*i-nv=*")  { LOGOP(==); }
END_PRIM(eqv);

#undef LOGOP

/*S* (zero? OBJ) => BOOLEAN */
/*D* Return #t if OBJ is zero, #f otherwise */
PRIM(zerop, "zero?", 1)
{
  RETURN(SCM_INUMP(TOS) ? SCM_MKBOOL(SCM_INUM(TOS) == 0) : scm_zerop(TOS));
}
END_PRIM(zerop);

/*S* (positive? OBJ) => BOOLEAN */
/*D* Return #t if OBJ is positive, #f otherwise */
PRIM(positivep, "positive?", 1)
{
  RETURN(SCM_INUMP(TOS) ? SCM_MKBOOL(SCM_INUM(TOS) > 0) : scm_positivep(TOS));
}
END_PRIM(positivep);

/*S* (negative? OBJ) => BOOLEAN */
/*D* Return #t if OBJ is negative, #f otherwise */
PRIM(negativep, "negative?", 1)
{
  RETURN(SCM_INUMP(TOS) ? SCM_MKBOOL(SCM_INUM(TOS) < 0) : scm_negativep(TOS));
}
END_PRIM(negativep);

/*S* (odd? OBJ) => BOOLEAN */
/*D* Return #t if OBJ is odd, #f otherwise */
PRIM(oddp, "odd?", 1)
{
  RETURN(SCM_INUMP(TOS) ? SCM_MKBOOL((SCM_INUM(TOS) & 1)==1) : scm_oddp(TOS));
}
END_PRIM(oddp);

/*S* (even? OBJ) => BOOLEAN */
/*D* Return #t if OBJ is even, #f otherwise */
PRIM(evenp, "even?", 1)
{
  RETURN(SCM_INUMP(TOS) ? SCM_MKBOOL((SCM_INUM(TOS) & 1)==0) : scm_evenp(TOS));
}
END_PRIM(evenp);

/*S* (min X1 X2 ...) => NUMBER */
/*D* Return the minimum of its arguments */
PRIMV(min, "min")
{
  int nargs = SCM_INUM(*ip++);
  SOBJ n;
  SOBJ *slot = (sp - nargs) + 1;
  
  if (nargs < 1) SCM_ERR("min: wrong number of args", NULL);

  sp--;
  if (SCM_INUMP(TOS)) {
    while(sp >= slot && SCM_INUMP( n = *sp )) {
      if ((long)TOS > (long)n)	TOS = n;
      sp--;
    }
  }
  while(sp >= slot) {
    n = *sp;
    if (scm_cmpnum(TOS, n) > 0)  TOS = n;
    sp--;
  }
  sp = slot;  
}
END_PRIM(min);

/*S* (max X1 X2 ...) => NUMBER */
/*D* Return the maximum of its arguments */
PRIMV(max, "max")
{
  int nargs = SCM_INUM(*ip++);
  SOBJ n;
  SOBJ *slot = (sp - nargs) + 1;
  
  if (nargs < 1) SCM_ERR("max: wrong number of args", NULL);

  sp--;
  if (SCM_INUMP(TOS)) {
    while(sp >= slot && SCM_INUMP( n = *sp )) {
      if ((long)TOS < (long)n)	TOS = n;
      sp--;
    }
  }
  while(sp >= slot) {
    n = *sp;
    if (scm_cmpnum(TOS, n) < 0)  TOS = n;
    sp--;
  }
  sp = slot;  
}
END_PRIM(max);

#define NATIVE_OPTIM

/*E* (add2 n1 n2) => n */
PRIM(add2, "add2", 2)
{
#ifdef NATIVE_OPTIM
  SOBJ n2;

  spop(n2); 
  if (SCM_INUMP( (long)TOS & (long)n2 )) {
	/* add n2 and TOS. if there is an overflow, call scm_add2 */
    ASMV ("movl	%1, %%eax\n"
	  "andb $-2,%%al\n"
	  "addl %0,%%eax\n"
	  "jo	add2_more\n"
	  "movl	%%eax, %0"
       : "+g"(TOS)				/* output */
       : "g" (n2)				/* input */
       : "ax","cc"				/* clobbered */
       );
  } else {
    ASMV ("add2_more:");
    TOS = scm_add2(TOS, n2);
  }
#else
  long r;
  SOBJ n2;
  spop(n2); 
  if (SCM_INUMP( (long)TOS & (long)n2 )) {
	r = SCM_INUM(TOS) + SCM_INUM(n2);
	TOS = SCM_INUM_RANGE(r) ? SCM_MKINUM(r) : scm_int2bnum(r);
  } else {
	TOS = scm_add2(TOS, n2);
  }
#endif /* NATIVE_OPTIM */  
}
END_PRIM(add2);

/*S* (+ N1 ...) => NUMBER */
/*D* return the sum of its arguments */
PRIMV(addv, "addv")
{
  SOBJ *base = (sp - SCM_INUM(*ip++)) + 1;
  SOBJ *p = base;
  SOBJ n;
  
  supdate();
  if (p <= sp) {
    TOS = *p++;
    while( p <= sp ) {
      n = *p++;
      if (SCM_INUMP((long)TOS & (long)n)) {
#ifdef NATIVE_OPTIM
	ASMV("movl %1, %%eax\n"
	     "andb $-2,%%al\n"
	     "addl %0, %%eax\n"
	     "jo   addv_ovf\n"
	     "movl %%eax, %0\n"
	     :"+g"(TOS)
	     :"g"(n)
	     :"ax","cc");
#else
	long r = SCM_INUM(TOS) + SCM_INUM(n);
	if (SCM_INUM_RANGE(r)) {
	  TOS = SCM_MKINUM(r);
	} else {
	  TOS = scm_add2(TOS, n);
	}
#endif /*NATIVE_OTIM*/
      } else {
#ifdef NATIVE_OPTIM
	ASMV("addv_ovf:");
#endif /*NATIVE_OTIM*/
	TOS = scm_add2(TOS, n);
      }
    }
  } else {
    TOS = SCM_INUM0;
  }
  sp = base;
}
END_PRIM(addv);

/*D* (mul2 N1 N2) => NUMBER */
/*E* Return the result of multiplying N1 with N2 */
PRIM(mul2, "mul2", 2)			/* n2 n1 -- n1*n2 */
{
  SOBJ n;

  spop(n);
  if (SCM_INUMP( (long)TOS & (long)n )) {
	long r, x, y;
	if ( (x = SCM_INUM(n)) != 0 && (y = SCM_INUM(TOS)) != 0) {
	  r = x * y;
	  if (y == (r / x)) {
		TOS = SCM_MKINUM(r);
	  } else {
		TOS = scm_mul2(TOS, n);
	  }
	} else {
	  TOS = SCM_INUM0;
	}
  } else {
	TOS = scm_mul2(TOS, n);
  }
  RETURN(TOS);
}
END_PRIM(mul2);

/*S* (* N1 ...) => NUMBER */
/*D* Return the product of its arguments */
PRIMV(mulv, "mulv")
{
  SOBJ *base = (sp - SCM_INUM(*ip++)) + 1;
  SOBJ *p = base;
  SOBJ n;
  supdate();
  if (p <= sp) {
    TOS = *p++;
    while (p <= sp && TOS != SCM_INUM0) {
      if ((n = *p++) == SCM_INUM0) {  TOS = SCM_INUM0;	break;  }
      if (SCM_INUMP((long)TOS & (long)n)) { 	/* inum multiplication */
	int r, x, y;
	x = SCM_INUM(TOS); y = SCM_INUM(n);
	r = x * y;
	if (x == (r / y)) {	/* no overflow */
	  TOS = SCM_MKINUM(r);
	  continue;
	}
      }
      TOS = scm_mul2(TOS, n);
    }
  } else {			/* no arguments */
    TOS = SCM_INUM1;		/* return 1 */
  }
  sp = base;
}
END_PRIM(mulv);

PRIM(sub2, "sub2", 2)
{
  SOBJ n;
  spop(n);
  if (SCM_INUMP( (long)TOS & (long)n )) {
    long r = SCM_INUM(TOS) - SCM_INUM(n);
    if (SCM_INUM_RANGE(r)) {
      TOS = SCM_MKINUM(r);
    } else {
      TOS = scm_sub2(TOS, n);
    }
  } else {
    TOS = scm_sub2(TOS, n);
  }
  RETURN(TOS);
}
END_PRIM(sub2);

/*S* (- N1 ...) => NUMBER */
/*D* Returns the difference of it's arguments. With one argument,
  return the additive inverse of the argument */
PRIMV(subv, "subv")
{
  SOBJ *base = (sp - SCM_INUM(*ip++)) + 1;
  SOBJ *p = base;
  SOBJ n;
  
  supdate();
  if (p <= sp) {
    TOS = *p++;
    while(p <= sp) {
      n = *p++;
      if (SCM_INUMP((long)TOS & (long)n)) {
	long r = SCM_INUM(TOS) - SCM_INUM(n);
	if (SCM_INUM_RANGE(r)) {
	  TOS = SCM_MKINUM(r);  continue;
	}
      }
      TOS = scm_sub2(TOS, n);
    }
  } else {
    TOS = SCM_INUM0;
  }
  sp = base;
}
END_PRIM(subv);

PRIM(div2, "div2", 2)
{
  SOBJ n;
  spop(n);
  TOS = scm_div2(TOS, n);
  RETURN(TOS);
}
END_PRIM(div2);

/*S* (/ N1 ...) => NUMBER */
/*D* Return the quotient of it's argument. With one argument, return
  the inverse of it's argument */
PRIMV(divv, "divv")
{
  SOBJ *base = (sp - SCM_INUM(*ip++)) + 1;
  SOBJ *p = base;
  if (NARGS == 0) SCM_ERR("/: bad number of args", NULL);
  if (NARGS == 1) {
    TOS = scm_div2(SCM_INUM1, TOS);
  } else {
    supdate();
    TOS = *p++;
    while(p <= sp) {
      TOS = scm_div2(TOS, *p++);
    }
  }
  sp = base;
}
END_PRIM(divv);

/*S* (abs X) => NUMBER */
/*D* Returns the absolute value of its argument. */
PRIM(abs, "abs", 1)
{
  if (SCM_INUMP(TOS)) {
    if (SCM_INUM(TOS) < 0) { 
      TOS = SCM_MKINUM( -(SCM_INUM(TOS)));
    }
  } else {
    TOS = scm_abs(TOS);
  }
  RETURN(TOS);
}
END_PRIM(abs);

/*S* (quotient N1 N2) => INTEGER */
/*D* Returns the quotient of N1/N2 rounded toward zero. */
PRIM(quotient, "quotient", 2)
{
  SOBJ x;
  spop(x);
  RETURN(scm_quotient(TOS, x));
}
END_PRIM(quotient);

/*S* (remainder N1 N2) => INTEGER */
/*D* Returns the quotient of N1/N2. */
PRIM(remainder, "remainder", 2)
{
  SOBJ x;
  spop(x);
  RETURN(scm_remainder(TOS, x));
}
END_PRIM(remainder);

/*S* (modulo N1 N2) => NUMBER */
/*D* Returns the modulo of N1/N2. */
PRIM(modulo, "modulo", 2)
{
  SOBJ x;
  spop(x);
  RETURN(scm_modulo(TOS, x));
}
END_PRIM(modulo);

/*S* (gcd N1 ...) => NUMBER */
/*D* Return the greatest common divisor of it's arguments. */
PRIMV(gcd, "gcd")
{
  int nargs = SCM_INUM(*ip++);
  SOBJ *base = (sp - nargs) + 1;
  supdate();  TOS = scm_gcd(nargs, base); sp = base;
}
END_PRIM(gcd);

/*S* (lcm N1 ...) => NUMBER */
/*D* Return the least common multiple of it's arguments */
PRIMV(lcm, "lcm")
{
  int nargs = SCM_INUM(*ip++);
  SOBJ *base = (sp - nargs) + 1;
  supdate();  TOS = scm_lcm(nargs, base); sp = base;
}
END_PRIM(lcm);

/*S* (floor N) => INTEGER */
/*D* Returns the largest integer not larger than N. */
PRIM(floor, 	"floor", 1)
{
  RETURN(scm_floor(TOS));
}
END_PRIM(floor);

/*S* (ceil N) => INTEGER */
/*D* Returns the smallest integer not smaller than N. */
PRIM(ceil, "ceiling", 1) 	
{
  RETURN(scm_ceil(TOS)); 
}
END_PRIM(ceil);

/*S* (truncate N) => INTEGER */
/*D* Returns the integer closest to N whose absolute value is not
  larger than the absolute value of N. */
PRIM(truncate, "truncate", 1) 	
{
  RETURN(scm_truncate(TOS)); 
}
END_PRIM(truncate);


/*S* (round N) => INTEGER */
/*D* Returns the closest integer to N, rounding to even when N is
  halfway between two integers. */
PRIM(round, "round", 1)
{
  RETURN(scm_round(TOS));
}
END_PRIM(round);

/*S* (exp X) => NUMBER */
/*D* Returns the value of e (the base of natural logarithms) raised to
  the power of X. */
PRIM(exp, "exp", 1) 		
{
  RETURN(scm_exp(TOS)); 
}
END_PRIM(exp);
/*S* (log X) => NUMBER */
/*D* Returns the natural logarithm of X. */
PRIM(log, "log", 1) 		
{
  RETURN(scm_log(TOS)); 
}
END_PRIM(log);

/*E* (log10 X) => NUMBER */
/*D* Returns the base-10 logarithm of X. */
PRIM(log10, "log10", 1) 		
{
  RETURN(scm_log10(TOS)); 
}
END_PRIM(log10);

/*S* (sin X) => NUMBER */
/*D* Returns the sinus of X, where X is given in radians. */
PRIM(sin, "sin", 1) 		
{
  RETURN(scm_sin(TOS)); 
}
END_PRIM(sin);

/*S* (cos X) => NUMBER */
/*D* Returns the cosine of X, where X is given in radians. */
PRIM(cos, "cos", 1) 		
{
  RETURN(scm_cos(TOS)); 
}
END_PRIM(cos);

/*S* (tan X) => NUMBER */
/*D* Returns the tangent of X, where X is given in radians. */
PRIM(tan, "tan", 1) 		
{
  RETURN(scm_tan(TOS)); 
}
END_PRIM(tan);

/*S* (asin X) => NUMBER */
/*D* Returns the arc sine of X; that is the value whose sine is X. */
PRIM(asin, "asin", 1) 		
{
  RETURN(scm_asin(TOS)); 
}
END_PRIM(asin);

/*S* (acos X) => NUMBER */
/*D* Returns the arc cosine of X; that is the value whose sine is X. */
PRIM(acos, "acos", 1) 		
{
  RETURN(scm_acos(TOS)); 
}
END_PRIM(acos);

/*S* (atan X) => NUMBER */
/*D* Returns the arc tangent of X in radians. */
/*S* (atan Y X)  => NUMBER */
/*D* calculates the arc tangent of the two variables X and Y.  It is
  similar to calculating the arc tangent of Y / X, except that the
  signs of both arguments are used to determine the quadrant of the
  result. */
PRIM(atan, "atan", 2)
{
  SOBJ x;  spop(x);  RETURN(scm_atan(TOS, x));
}
END_PRIM(atan);

/*S* (sqrt X) => NUMBER */
/*D* Returns the principal square root of X. */
PRIM(sqrt, "sqrt", 1)
{
  RETURN(scm_sqrt(TOS));
}
END_PRIM(sqrt);

/*S* (expt X Y) => NUMBER */
/*D* Returns X raised to the power Y. */
PRIM(expt, "expt", 2)
{
  SOBJ x;  spop(x);  RETURN(scm_expt(x, TOS));
}
END_PRIM(expt);

/*E* (random) => FLOAT */
/*D* Returns a random number in range 0-1.0. */
PRIM(random, "random", 0)
{
  spush(scm_mkfnum(drand48()));
}
END_PRIM(random);

/*S* (exact->inexact Z) => NUMBER */
/*D* Returns an inexact representation of Z. The value returned is the
  inexact number that is numerically closest to the argument. */
PRIM(exact2inexact, "exact->inexact", 1)
{
  RETURN(scm_exact_to_inexact(TOS));
}
END_PRIM(exact2inexact);

/*S* (inexact->exact z) => NUMBER */
/*D* Returns an exact representation of Z.  The value returned is the
  exact number that is numerically closest to the argument. */
PRIM(inexact2exact, "inexact->exact", 1)
{
  RETURN(scm_inexact_to_exact(TOS));
}
END_PRIM(inexact2exact);

/*S* (number->string Z [RADIX]) => STRING */
/*D* Returns as a string an external representation of the given number
  in the given radix */
PRIMV(number2string, "number->string")
{
  int nargs = SCM_INUM(*ip++);
  SOBJ num, base;
  switch(nargs) {
  case 1:    base = NULL;  num = TOS;	break;
  case 2:    spop(base);   num = TOS;	break;
  default:   SCM_ERR("number->string: bad number of args", NULL);
  }
  TOS = scm_number_to_string(num, base);
}
END_PRIM(number2string);

/*S* (string->number STRING RADIX) => NUMBER */
/*D* Returns a number of the maximally precise representation expressed by
  the given string. */
PRIMV(string2number, "string->number")
{
  int nargs = SCM_INUM(*ip++);
  SOBJ n;
  if (nargs < 1) SCM_ERR("string->number: bad number of args", NULL);
  if (nargs > 1) {
    spop(n);
  } else {
    n = NULL;
  }
  TOS = scm_string_to_number(TOS, n);
}
END_PRIM(string2number);

/*E* (1+ X) => NUMBER*/
/*D* Returns X + 1. */
PRIM(plus1, "1+", 1)
{
  if (!SCM_INUMP(TOS) || ( (long)TOS >= (long)SCM_MKINUM(SOBJ_INUM_MAX))) {
    TOS = scm_add2(SCM_INUM1, TOS);
  } else {
    (long)TOS += (1 << SOBJ_INUM_SHIFT);
  }
  RETURN(TOS);
}
END_PRIM(plus1);

/*E* (2+ X) => NUMBER*/
/*D* Returns X + 2. */
PRIM(plus2, "2+", 1)
{
  if (SCM_INUMP(TOS) && ( (long)TOS <= (long)SCM_MKINUM(SOBJ_INUM_MAX-2))) {
    (long)TOS += (2 << SOBJ_INUM_SHIFT);
  } else {
    TOS = scm_add2(SCM_MKINUM(2), TOS);
  }
  RETURN(TOS);
}
END_PRIM(plus2);

/*E* (1- X) => NUMBER*/
/*D* Returns X - 1. */
PRIM(minus1, "1-", 1)
{
  if (SCM_INUMP(TOS) && ( (long)TOS > (long)SCM_MKINUM(SOBJ_INUM_MIN) )) {
    (long)TOS -= (1 << SOBJ_INUM_SHIFT);
  } else {
    TOS = scm_sub2(TOS, SCM_INUM1);
  }
  RETURN(TOS);
}
END_PRIM(minus1);

/*E* (2- X) => NUMBER*/
/*D* Returns X - 2. */
PRIM(minus2, "2-", 2)
{
  if (SCM_INUMP(TOS) && ( (long)TOS >= (long)SCM_MKINUM(SOBJ_INUM_MIN+2) )) {
    (long)TOS -= (2 << SOBJ_INUM_SHIFT);
  } else {
    TOS = scm_sub2(TOS, SCM_MKINUM(2));
  }
  RETURN(TOS);
}
END_PRIM(minus2);

