;
; computing primes using the sieve of Eratosthenes
;
(define size 8190) ;; flags array size

(define flags (make-vector size 1)) ;; makes flags array, set to 1
;; as in C, a vector of size 10 has indices 0..9

(define primes   ;; the driver of the program.  Causes primes to be
  (lambda (x)    ;; calculated 100 times
    (do ((loop 100 (- loop 1)))  ;; do loops are weird -- the terminate
        ((zero? loop) '())       ;; on a TRUE rather than false
      (find-primes x))))

(define find-primes  ;; calculates all the primes until the limit is reached
  (lambda (x)
    (vector-fill! flags 1) ;; set flags array to 1
    (let ((count 0))
      (do ((j 0 (+ j 1)))
          ((not (< count x)) '())
        (if (= 1 (vector-ref flags j)) ;; if the element is 1, we've
            (let ((prime (+ j j 3)))   ;; found a prime
              (set! count (+ count 1))
              (change-flags (+ j prime) prime))))))) ;; set the appropriate
                                                     ;; flags

(define change-flags  ;; iterator to set flags of prime multiples
  (lambda (init primes)
    (do ((k init (+ k primes)))
        ((not (< k  size)) '())
      (vector-set! flags k 0))))	;; set element k to 0 (false)

(define (sieve filter n)
  (cond ((> n 100)   '())
		((filter n) (sieve filter (+ n 1)))
		(else
		 (cons n (sieve (lambda (x)
						  (if (zero? (remainder x n)) #t
							  (filter x)))
						(+ n 1))))))

(sieve #f 2)

(define (prime? n primes)
  (or (null? primes)
	  (and (not (zero? (remainder n (car primes))))
		   (prime? n (cdr primes)))))

(define (prime n)
  (let loop ((i 2) (candidate 2) (primes `()))
	(cond ((not (prime? candidate primes)) (loop i (+ 1 candidate) primes))
		  ((= i n) candidate)
		  (else
		   (loop (+ 1 i) (+ 1 candidate) (cons candidate primes))))))

(define (test)
  (let loop ((count 1))
	(prime 100)
	(format standard-output "%d" count)
	(loop (+ 1 count))))
