/* -*- tab-width:4; -*- */
/*
 * Procedure related functions
 *
 * $Id: proc.c,v 1.5 2000/11/01 22:30:31 crad Exp $
 */
#include "s.h"
#include "proc.h"
#include "heap.h"

#ifndef NEWENV
void scm_env_mark(SOBJ env)
{
  SOBJ val;
  int i;

  if (env == NULL) 	return;

  for (i = 0; i < SCM_INUM(SCM_ENV_FRAME(env)->nslots); i++) {
	val = SCM_ENV_FRAME(env)->slot[i];
	if (scm_is_pointer_to_heap(val)) {
	  scm_gc_mark(val);
	}
  }
  if (SCM_ENV_NEXT(env) != NULL) {
	scm_gc_mark(SCM_ENV_NEXT(env));
  }
}
#endif /* NEWENV */

void scm_proc_mark(SOBJ proc)
{
  SCM_Code *c; 
  int i;
  
  if (SCM_PROC_ENV(proc)) 	scm_gc_mark(SCM_PROC_ENV(proc));

  if ((c = SCM_PROC_CODE(proc)) != NULL) {
	scm_gc_mark(c->envlist);
	
	for (i = 0; i < c->size; i++) {
	  if (scm_is_pointer_to_heap(c->code[i]))
		scm_gc_mark(c->code[i]);
	}
  }
}

void scm_clos_mark(SOBJ obj)
{
#ifdef NEWENV
  SCM_ENVFRAME *ef;
  int i;
  if ((ef = SCM_CLOSURE_ENV(obj)) != NULL) {

	/* tricky part: nslots normaly hold an INUM, but GC preparation
	 * clears the the INUM_TAG bit. After an env is marked, I set the
	 * INUM_TAG again. That's why nslots is tested for INUMP: if it's
	 * an INUM, the envframe has already been visited. */

	while(ef && !SCM_INUMP(ef->nslots)) {
	  /* fprintf(stderr, "scm_clos_mark: before env %p nslots=%02x\n",
			  ef, ef->nslots);*/
	  for (i = 0; i < SCM_INUM(ef->nslots); i++) {
		if (!SCM_INUMP(ef->slot[i]) && ef->slot[i] != NULL)
		  scm_gc_mark(ef->slot[i]);
	  }
	  (long)ef->nslots |= SOBJ_INUM_TAG;
	  /*fprintf(stderr, "scm_clos_mark: after  env %p nslots=%02x\n",
		ef, ef->nslots);*/

	  ef = ef->next;
	}
  }
#else
  if (SCM_CLOSURE_ENV(obj)) 	scm_gc_mark(SCM_CLOSURE_ENV(obj));
#endif /* NEWENV */
  if (SCM_CLOSURE_CODE(obj))	scm_gc_mark(SCM_CLOSURE_CODE(obj));
}

void scm_clos_sweep(SOBJ clos)
{
  SCM_ENVFRAME *ef;

  /*  fprintf(stderr, "scm_clos_sweep: sweeping closure %p\n", clos); */
  ef = SCM_CLOSURE_ENV(clos);

  /* loop while environment and environment has NOT been visited
   * during the closure mark phase. Lost environment have
   * !SCM_INUMP(ef->nslots); */
  
  while(ef != NULL && !SCM_INUMP(ef->nslots)) {
	/*	fprintf(stderr, "scm_clos_sweep: defer sweep env %p nslots=%02x\n",
		ef, ef->nslots); */
	scm_gc_defer_sweep_add(ef);
	ef = ef->next;
  }
}


#ifndef NEWENV
void scm_env_sweep(SOBJ env)
{
  if (SCM_ENV_FRAME(env)) {
	scm_free(SCM_ENV_FRAME(env));
	SCM_ENV_FRAME(env) = NULL;
  }
}
#endif /* NEWENV */

void scm_proc_sweep(SOBJ proc)
{
  if (SCM_PROC_CODE(proc)) {
	scm_free(SCM_PROC_CODE(proc));
	SCM_PROC_CODE(proc) = NULL;
  }
}



#ifndef NEWENV
void scm_env_print(SOBJ obj, PORT *p)
{
  int i;
  if (obj == NULL) return;
	
  if (SCM_ENV_FRAME(obj) == NULL) {
	port_puts(p, "#<env null>");
  } else {
	port_puts(p, "#<env next=");
	port_putx(p, SCM_ENV_NEXT(obj));
	port_puts(p, " nslots=");
	port_putn(p, SCM_INUM(SCM_ENV_FRAME(obj)->nslots));
	port_puts(p, " slot[]=");
	for (i = 0; i < SCM_INUM(SCM_ENV_FRAME(obj)->nslots); i++) {
	  if (!SCM_INUMP(SCM_ENV_FRAME(obj)->slot[i]) &&
		  SCM_CLOSUREP(SCM_ENV_FRAME(obj)->slot[i])) {
		port_puts(p, "#<closure>");
	  } else {
		scm_write_obj(SCM_ENV_FRAME(obj)->slot[i], p, 1);
	  }
	  if (i != SCM_INUM(SCM_ENV_FRAME(obj)->nslots)-1) port_putc(p, ' ');
	}
	port_putc(p, '>');
  }
}

SCM_STRBUF *scm_closure_env2str(SCM_STRBUF *sb, SOBJ obj, int raw)
{
  SOBJ *slot;
  int i;
  char *spacer;

  if (obj == NULL) 	return(sb);
  if (SCM_ENV_FRAME(obj) == NULL)
	return(scm_strbuf_concat_str(sb, "#<env NULL>"));

  sb = scm_strbuf_concat_sprintf(sb, "#<env %p: next=%p nslots=%d slot=[",
								 obj,
								 SCM_ENV_FRAME(obj)->next,
								 SCM_INUM(SCM_ENV_FRAME(obj)->nslots));

  spacer = "[";
  slot = SCM_ENV_FRAME(obj)->slot;
  for (i = 0; i < SCM_INUM(SCM_ENV_FRAME(obj)->nslots); i++) {
	sb = scm_strbuf_concat_str(sb, spacer);
	spacer = " ";
	if (!SCM_INUMP(slot[i]) && SCM_CLOSUREP(slot[i])) {
	  sb = scm_strbuf_concat_str(sb, "#<closure>");
	} else {
	  sb = scm_iobj2str(sb, slot[i], raw);
	}
  }
  sb = scm_strbuf_concat_str(sb, "]");
  return(sb);
}
#endif /* NEWENV */

#ifdef NEWENV
static SCM_STRBUF *scm_envchain2str(SCM_STRBUF *sb, SCM_ENVFRAME *env, int raw)
#else
static SCM_STRBUF *scm_envchain2str(SCM_STRBUF *sb, SOBJ env, int raw)
#endif /* NEWENV */
{
  int i;
  char *spacer;
  SCM_ENVFRAME *ef;

#ifndef NEWENV
  if (env == NULL || (ef = SCM_ENV_FRAME(env)) == NULL)
	return(scm_strbuf_concat_str(sb, "#<env global>"));
#else
  if ((ef = env) == NULL)
	return(scm_strbuf_concat_str(sb, "#<env global>"));
#endif /* NEWENV */

  sb = scm_strbuf_concat_str(sb, "#<env ");
  while(ef) {
	spacer = "[";
	for (i = 0; i < SCM_INUM(ef->nslots); i++) {
	  sb = scm_strbuf_concat_str(sb, spacer); spacer = " ";
	  if (!SCM_INUMP(ef->slot[i]) && SCM_CLOSUREP(ef->slot[i])) {
		sb = scm_strbuf_concat_str(sb, "#<closure>");
	  } else {
		sb = scm_iobj2str(sb, ef->slot[i], raw);
	  }
	}
	sb = scm_strbuf_concat_str(sb, "]");
	ef = ef->next;
  }
  return(scm_strbuf_concat_str(sb, ">"));
}

void scm_envchain_print(SCM_ENVFRAME *ef, PORT *p)
{
  char *spacer;
  int i;
  port_puts(p, "#<env ");
  while(ef) {
	spacer = "[";
	for (i = 0; i < SCM_INUM(ef->nslots); i++) {
	  port_puts(p, spacer); spacer = " ";
	  if (!SCM_INUMP(ef->slot[i]) && SCM_CLOSUREP(ef->slot[i])) {
		port_puts(p, "#<closure>");
	  } else {
		scm_write_obj(ef->slot[i], p, 1);
	  }
	}
	port_puts(p, "]");
	ef = ef->next;
  }
}

void scm_proc_print(SOBJ obj, PORT *p)
{
  SOBJ env, e;
  int i;

  if (SCM_PROC_CODE(obj) == NULL) {
	port_puts(p, "#<proc null>");
  } else {
	port_puts(p, "#<proc (");
	env = SCM_PROC_CODE(obj)->envlist;
	if (env && SCM_PAIRP(env))
	  e = scm_reverse(SCM_CAR(env));
	else
	  e = NULL;

	if (SCM_PROC_CODE(obj)->nlocals) {
	  port_puts(p, "local: ");
	  for (i = 0; i < SCM_PROC_CODE(obj)->nlocals && e; i++) {
		scm_write_obj(SCM_CAAR(e), p, 1);
		e = SCM_CDR(e);  if (e) port_putc(p, ' ');
	  }
	}
	if (SCM_PROC_CODE(obj)->nlocals) {
	  port_puts(p, " args: ");
	}
	while(e) {
	  if (SCM_CAR(e)) {
		scm_write_obj(SCM_CAAR(e), p, 1);
	  } else {
		port_puts(p, "*BAD_ENV*");
		break;
	  }
	  e = SCM_CDR(e);  if (e) port_putc(p, ' ');
	}
	port_puts(p, ") ");
#ifdef NEWENV
	port_puts(p, "#<env>");		/* KKK */
#else
	scm_env_print(SCM_PROC_ENV(obj), p);
#endif /*NEWENV*/
	port_puts(p, " code=");
	port_putx(p, SCM_PROC_CODE(obj));
	port_puts(p, ">");
  }
}

SCM_STRBUF *scm_proc2str(SCM_STRBUF *sb, SOBJ obj, int raw)
{
  SOBJ env, e;
  int i;
  char *spacer;
  
  if (SCM_PROC_CODE(obj) == NULL)
	return(scm_strbuf_concat_str(sb, "#<proc NULL>"));
  
  env = SCM_PROC_CODE(obj)->envlist;
  e = (env && SCM_PAIRP(env)) ? scm_reverse(SCM_CAR(env)) : NULL;
  sb = scm_strbuf_concat_sprintf(sb, "#<proc %p", obj);
  if (SCM_PROC_CODE(obj)->nlocals) {
	spacer = " local:";
	for (i = 0; i < SCM_PROC_CODE(obj)->nlocals && e; i++) {
	  sb = scm_strbuf_concat_str(sb, spacer);  spacer = " ";
	  sb = scm_iobj2str(sb, SCM_CAAR(e), 1);
	}
  }
  spacer = " args:";
  while(e) {
	sb = scm_strbuf_concat_str(sb, spacer);  spacer = " ";
	if (SCM_CAR(e)) {
	  sb = scm_iobj2str(sb, SCM_CAAR(e), 1);
	} else {
	  sb = scm_strbuf_concat_str(sb, "*BADENV*");
	}
	e = SCM_CDR(e);
  }
#ifndef NEWENV
  sb = scm_closure_env2str(sb, SCM_PROC_ENV(obj), 0);
#endif /* NEWENV */

  sb = scm_strbuf_concat_sprintf(sb, " code=%p nargs=%d optargs=%d>", 
								 SCM_PROC_CODE(obj),
								 SCM_PROC_CODE(obj)->nargs,
								 SCM_PROC_CODE(obj)->optargs
								 );
  return(sb);
}

void scm_clos_print(SOBJ obj, PORT *p)
{
  port_puts(p, "#<closure ");
  if (SCM_CLOSURE_ENV(obj)) {
#ifdef NEWENV
	scm_envchain_print(SCM_CLOSURE_ENV(obj), p);
#else
	scm_env_print(SCM_CLOSURE_ENV(obj), p);
#endif /* NEWENV */

  } else {
	port_puts(p, "global");
  }
  port_puts(p, " ");
  scm_proc_print(SCM_CLOSURE_CODE(obj), p);
  port_puts(p, ">");
}

SCM_STRBUF *scm_clos2str(SCM_STRBUF *sb, SOBJ obj, int raw)
{
  sb = scm_strbuf_concat_str(sb, "#<closure ");
  sb = scm_envchain2str(sb, SCM_CLOSURE_ENV(obj), raw);
  sb = scm_strbuf_concat_str(sb, " ");
  sb = scm_proc2str(sb, SCM_CLOSURE_CODE(obj), raw);
  sb = scm_strbuf_concat_str(sb, ">");
  return(sb);
}

/*-- public functions */

/* return #t if x is something we can call */

/*S* (procedure? OBJ) => BOOLEAN */
/*D* Returns #t if OBJ is a procedure, #f otherwise */

SOBJ scm_procedurep(SOBJ x)
{
  return(SCM_MKBOOL(SCM_PROCP(x) || SCM_CLOSUREP(x) ||
					SCM_PRIMP(x) || SCM_CPRIMP(x)));
}

/*E* (environment? OBJ) => BOOLEAN */
/*D* Returns #t if OBJ is an environment, #f otherwise */

/*E* (closure? OBJ) => BOOLEAN */
/*D* Returns #t if OBJ is a closure, #f otherwise */

/*E* (primitive? OBJ) => BOOLEAN */
/*D* Returns #t if OBJ is a primitive, #f otherwise */

/*E* (cprimitive? OBJ) => BOOLEAN */
/*D* Returns #t if OBJ is a C primitive, #f otherwise */

/*E* (syntax? OBJ) => BOOLEAN */
/*D* Returns #t if OBJ is a syntax, #f otherwise */
#ifndef NEWENV
SOBJ scm_environmentp(SOBJ x){  return(SCM_MKBOOL(SCM_ENVP(x))); 		}
#endif /*NEWENV*/
SOBJ scm_closurep(SOBJ x) 	 {	return(SCM_MKBOOL(SCM_CLOSUREP(x))); 	}
SOBJ scm_primitivep(SOBJ x)  {  return(SCM_MKBOOL(SCM_PRIMP(x))); 		}
SOBJ scm_cprimitivep(SOBJ x) {  return(SCM_MKBOOL(SCM_CPRIMP(x))); 		}
SOBJ scm_syntaxp(SOBJ x) 	 {  return(SCM_MKBOOL(SCM_SYNTAXP(x))); 	}

/*E* (cprimitive-arity PRIM) => NUMBER */
SOBJ scm_cprimitive_arity(SOBJ x)
{
  if (!SCM_CPRIMP(x))	SCM_ERR("bad cprimitive", x);
  return(SCM_MKINUM(SCM_CPRIM_NARGS(x)));
}

/*E (primitive-name PRIM) => STRING */
SOBJ scm_primitive_name(SOBJ x)
{
  if (!SCM_PRIMP(x))	SCM_ERR("bad cprimitive", x);
  return(scm_mkstring(SCM_PRIM(x)->name));
}

/*E* (primitive-address PRIM) => POINTER */
SOBJ scm_primitive_address(SOBJ x)
{
  if (!SCM_PRIMP(x))	SCM_ERR("bad cprimitive", x);
  return(scm_mkpointer(SCM_PRIM(x)->address));
}

/*E* (primitive-arity PRIM) => NUMBER */
SOBJ scm_primitive_arity(SOBJ x)
{
  if (!SCM_PRIMP(x))	SCM_ERR("bad cprimitive", x);
  return(SCM_MKINUM(SCM_PRIM(x)->nargs));
}

void scm_init_proc()
{
  scm_add_cprim("procedure?", 	scm_procedurep, 	1);
#ifndef NEWENV
  scm_add_cprim("environment?", scm_environmentp, 	1);
#endif /*NEWENV*/
  scm_add_cprim("closure?",     scm_closurep, 		1);
  scm_add_cprim("primitive?", 	scm_primitivep, 	1);
  scm_add_cprim("cprimitive?", 	scm_cprimitivep, 	1);
  scm_add_cprim("syntax?",		scm_syntaxp,		1);

  scm_add_cprim("cprimitive-arity",		scm_cprimitive_arity,	1);

  scm_add_cprim("primitive-name",		scm_primitive_name,		1);
  scm_add_cprim("primitive-address",	scm_primitive_address,	1);
  scm_add_cprim("primitive-arity",		scm_primitive_arity,	1);
}
