; -*- scheme -*-
;
; My compiler
;

(define nil '())


(define *sc-code* nil)					;the code generated by compile
(define *sc-global-env* nil)			;the global environment
(define *sc-symbols* nil)				;the symbol table
(define *sc-label* 1)					;the label number

(define (sc-pp code)
  (if (null? code)
	  #t
	  (begin (display "\t")
			 (display (car code))
			 (newline)
			 (sc-pp (cdr code)))))
	  

(define (sc-print-code)
  (display "Code:\n")   (sc-pp *sc-code*) (newline)
  (display "Global:\n") (display *sc-global-env*) (newline))

(define (compile expr)
  (set! *sc-code* nil)
  (set! *sc-label* 1)
  (sc-gen expr nil #t)
  (sc-print-code)
  )

;
; the generator
;
(define (sc-gen expr env top)
  (cond
   ((symbol? expr)				(sc-gen-ref expr env top))
   ((not (pair? expr)) 			(sc-gen-const expr env top))
   ((eq? (car expr) 'quote) 	(sc-gen-const (cadr expr) env top))
   ((eq? (car expr) 'set!)		(sc-gen-set! expr env top))
   ((eq? (car expr) 'if)		(sc-gen-if   expr env top))
   ((eq? (car expr) 'lambda)	(sc-gen-lambda expr env top))
   (#t	 
	(let ((args (map (lambda (x) (sc-gen x env #f)) (cdr expr))))
	  (let ((var (and (symbol? (car expr))
					  (sc-env-lookup (car expr) env))))
		(display "args") (display args) (newline)
		(display "var") (display var) (newline)
		(if (null? var)
			(sc-app (cons var args) #t top)
			(sc-app (cons (sc-gen (car expr) env #f) args) #f top)))))))

(define (sc-app vals glo term)
  (display "sc-app vals=") (display vals) (newline)
  (display " global=") (display glo) (newline)
  (display " term=") (display term) (newline))

			
; Add code to *sc-code*

(define (sc-gencode . x)
  (set! *sc-code* (append *sc-code* (list x))))

; Generate a new label

(define (sc-genlab)
  (let ((lab (string->symbol
			  (string-append "label"
							 (number->string *sc-label*)))))
	(set! *sc-label* (+ 1 *sc-label*))
	lab))

; Generate code for a constant value

(define (sc-gen-const expr env top)
  (display "const:") (display expr) (newline)
  (sc-gencode 'push expr))


; Generate code for a variable ref

(define (sc-gen-ref expr env top)
  (display "ref:") (display expr) (newline)
  (let ((ofs (sc-env-lookup expr env)))
	(if (null? ofs)						; not found in env
		(sc-gencode 'ref-glo (sc-global-lookup expr))
		(sc-gencode 'ref-loc expr))))

; Generate code for setting a variable

(define (sc-gen-set! expr env top)
  (display "set:") (display expr) (newline)
  (display "var=") (display (cadr expr)) (newline)
  (display "value=") (display (caddr expr)) (newline)
  (sc-gen (cadr expr) env top)
  (sc-gen (caddr expr) env top)
  (sc-gencode 'store))

; Generate code for an if
(define (sc-gen-if expr env top)
  (display "if:") (display expr) (newline)
  (display "expression length = ") (display (length expr)) (newline)
  (sc-gen (cadr expr) env top)			; generate test
  (if (> (length expr) 3)
	  (let ((l1 (sc-genlab)) (l2 (sc-genlab)))
		(sc-gencode 'brif l1)
		(sc-gen (caddr expr) env top)
		(sc-gencode 'jump l2)
		(sc-gencode 'label l1)
		(sc-gen (cadddr expr) env top)
		(sc-gencode 'label l2))
	  (let ((l (sc-genlab)))
		(sc-gencode 'brif l)
		(sc-gen (caddr expr) env top)
		(sc-gencode 'label l))))

;
; generate code for a lambda expr
;
(define (sc-gen-lambda expr env top)
  (display "Lambda:") (display expr) (newline)
  (let ((current-code *sc-code*)
		(closure-lab (sc-genlab))
		(nargs 0))
	(set! *sc-code* nil)				; reset current code
	;
	; Compile the closure under label xxx
	;
	(sc-gencode 'label closure-lab)

	(let ((arglist (cadr expr))
		  (body (cddr expr))
		  (new-env nil))

	  (set! new-env (append arglist env))
	  (set! nargs (length arglist))
	  (sc-gencode 'args nargs)
	  (map (lambda (x) (sc-gen x new-env #f)) body)
	  )
	(sc-gencode 'return)
	;
	;
	;
	(set! *sc-code* (append *sc-code* current-code))
	(sc-gencode 'closure closure-lab)
	))

; ----------------------------------------------------------------

; Global environment handling
;

; Search assoc for symbol in global env: return the cell. If assoc
; does not exist, create it.

(define (sc-global-lookup sym)
  (let ((cell (assoc sym *sc-global-env*)))
	(if cell
		cell
		(let ((newcell (cons sym '**unbound**)))
		  (set! *sc-global-env* (cons newcell *sc-global-env*))
		  newcell))))

; define symbol in global env

(define (sc-define-global sym value)
  (set-cdr! (sc-global-lookup sym) value))

; lookup an enviroment for a symbol. If symbol if not found, nil is
; returned

(define (sc-env-lookup sym env)
  (let ((x (memq sym env)))
	(if x
		(- (length env) (length x))		; offset into env
		nil)))

; Symbol lookup. If symbol is found in env, return the pos in env
; otherwise return the cell in global env.

(define (sc-symbol-lookup sym env)
  (let ((ofs (sc-env-lookup sym env)))
	(if (null? ofs)
		(sc-global-lookup sym)
		ofs)))

