/* -*- c -*- */

VMPRIM(nop, "%nop", 0)
{
  NEXT;
}

VMPRIM(end, "%nop", 0)
{
  supdate();
  scm_sp=sp;
  return(NULL);
}

#ifdef STACK_CHECKS
VMPRIM(overflow, "%overflow", 0)
{
  SCM_ERR("Stack overflow", NULL);
}

#ifdef NOT_USED
VMPRIM(underflow, "%underflow", 0)
{
  SCM_ERR("Stack underflow", NULL);
}
#endif
#endif
 
VMPRIM(doletstar, "%dolet*", 1)
{
  int n = SCM_INUM(*ip++);
  /*-- mark and push n #unbound object */
  PUSH_CONT();
  while(--n >= 0) spush(scm_unbound);
} 	/* FALL THROUGH */

VMPRIM(dolet , "%dolet ", 0)
{
  /* stack: cont binding_n ... binding_0 */
  int nslots = (SOBJ*)cont - sp;

  cont->env = env;				/* save env */
  *sp = TOS;
  PUSH_ENV_FRAME(env, nslots);
  NEXT;
}


/* simple stack manipulations */

VMPRIM(drop, "%drop", 0)
{
  sdrop();  NEXT;
}
VMPRIM(push, "%push", 1)
{
  spush( *ip++ );  NEXT;
}
VMPRIM(pushv, "%pushv", 1)
{
  spush( SCM_SYM_VALUE((SOBJ)(*ip++))); NEXT;
}

VMPRIM(pushl, "%pushl", 1)
{		/* lev, ofs */
  SOBJ p = env;
  short n = SCM_INUM(*ip) >> 16;
  while(--n >= 0) 	p = SCM_ENV_NEXT(p);
  spush(SCM_ENV_FRAME(p)->binding[(short)SCM_INUM(*ip++)]);
  NEXT;
}

VMPRIM(pushl0 , "%pushl0 ", 1)
{
  spush(cbinding[SCM_INUM(*ip++)]);
  NEXT;
}

VMPRIM(pushl1, "%pushl1", 1)
{
  spush(SCM_ENV_FRAME(SCM_ENV_NEXT(env))
		->binding[SCM_INUM(*ip++)]);
  NEXT;
}

VMPRIM(pushl2, "%pushl2", 1)
{
  spush(SCM_ENV_FRAME(SCM_ENV_NEXT(SCM_ENV_NEXT(env)))
		->binding[SCM_INUM(*ip++)]);
  NEXT;
}

VMPRIM(pushl3, "%pushl3", 1)
{
  spush(SCM_ENV_FRAME(SCM_ENV_NEXT(SCM_ENV_NEXT(SCM_ENV_NEXT(env))))
		->binding[SCM_INUM(*ip++)]);
  NEXT;
}

VMPRIM(alloc, "%alloc", 1)
{					 	/* S: -- [ nil ]n  */
  int n = SCM_INUM(*ip++);
  while(--n >= 0) 
	spush(NULL);
  NEXT;
}

VMPRIM(pushlist, "%pushlist", 1)
{								/* mark -- mark item#n-1 ... item#0 */
  SOBJ p, list = *ip++;
  int i;

  *sp = TOS;
  for (i = 0, p = list; p; p = SCM_CDR(p), i++)
	;
  sp -= i;
  for (i = 0, p = list; p; sp[i++] = SCM_CAR(p), p = SCM_CDR(p))
	;
  TOS = *sp;
  NEXT;
}

#ifdef NOT_USED
VMPRIM(nonimp, "%nonimp", 0)
{
  SCM_ERR("not implemented opcode", NULL);
}
#endif

/*-- create a contframe struct on the stack:
 * we push because of possible use of TOS cache:
 * after PUSH_CONT, cont will point to a contframe struct...
 */
VMPRIM(mark, "%mark", 0)
{						/* push a continuation frame on the stack */
  PUSH_CONT();
  NEXT;
}

VMPRIM(mkclosure, "%mkclosure", 0)
{					/* proc -- closure */
  
  SOBJ closure = scm_newcell(SOBJ_T_CLOSURE);
  SCM_CLOSURE_CODE(closure) = TOS;
  SCM_CLOSURE_ENV(closure) = env;
#ifdef DEBUG
  scm_puts("; make closure\n");
#endif
  TOS = closure;
  NEXT;
}

VMPRIM(callc, "%callc", 0)
{
  SCM_ERR("%callc: not implemented", NULL);
}

VMPRIM(callc0, "%callc0", 0)
{								/* proc -- n */
  sresync();
  TOS = (*SCM_CPRIM_FUNC((SOBJ)TOS)) ();
  check_overflow();
  NEXT;
}
VMPRIM(callc1, "%callc1", 0)
{								/* a1 proc -- n */
  SOBJ r; 
  sresync();
  r = (*SCM_CPRIM_FUNC((SOBJ)TOS))(sp[1]);
  sp += 1;  TOS=r;
  check_overflow();
  NEXT;
}
VMPRIM(callc2, "%callc2", 0)
{								/* a2 a1 proc -- n */
  SOBJ r;
  sresync();
  r = (*SCM_CPRIM_FUNC((SOBJ)TOS))(sp[1], sp[2]);
  sp += 2; TOS=r;
  check_overflow();
  NEXT;
}
VMPRIM(callc3, "%callc3", 0)
{
  SOBJ r;
  sresync();
  r = (*SCM_CPRIM_FUNC((SOBJ)TOS))(sp[1], sp[2], sp[3]);
  sp += 3; TOS = r;
  check_overflow();
  NEXT;
}
VMPRIM(callc4, "%callc4", 0)
{
  SOBJ r;
  sresync();
  r = (*SCM_CPRIM_FUNC((SOBJ)TOS))(sp[1], sp[2], sp[3], sp[4]);
  sp += 4; TOS = r;
  check_overflow();
  NEXT;
}
VMPRIM(callc5, "%callc5", 0)
{
  SOBJ r;
  sresync();
  r = (*SCM_CPRIM_FUNC((SOBJ)TOS))(sp[1], sp[2], sp[3], sp[4], sp[5]);
  sp += 5; TOS = r;
  check_overflow();
  NEXT;
}

VMPRIM(calls, "%calls", 0)
{								/* cont an..a0 proc -- r */
  SOBJ ret, func;

  spop(func);
  cont->env = env;
  cont->ip  = ip;
  ret = (*SCM_CPRIM_FUNC(func))( (SOBJ*)cont - (SOBJ*)sp, sp);
  spush(ret);
  goto l_return;
}

l_no_call1:
  proc = TOS;

l_no_call: 
  SCM_ERR("cannot apply to", proc);

VMPRIM(endlet_jump, "%endlet_jump", 0)
{								/* args proc */
  src = (SOBJ*)cont;
  cont = cont->next;
  goto l_jump1;
}

VMPRIM(jump, "%jump", 0)
{
  src = (SOBJ*)cont;
 
 l_jump1:
  cont = cont->next;
  env = cont->env;
  dst = (SOBJ*)cont;
  spop(proc);
  while(--src >= sp) { *(--dst) = *src; }
  sp = dst; TOS = *sp;

  goto *(scm_type_hook[proc->type].execute); 
}


VMPRIM(call, "%call", 0)
{
  if (TOS == NULL || SCM_INUMP(TOS))	goto l_no_call1;

  cont->ip = ip;
  cont->env = env;
  spop(proc);
  goto *(scm_type_hook[proc->type].execute);
}

VMPRIM(call_prim, "%call_prim", 0)
{
  SCM_PRIM_TABLE *p = SCM_PRIM(proc);
  if (p->nargs >= 0) {			/* pop continuation frame (not used) */
	cont = cont->next;
  }
  goto *(p->address);
}

VMPRIM(call_cprim, "%call_cprim", 0)
{
  spush(proc);
  switch(SCM_CPRIM_NARGS(proc)) {
  case -1: 	goto l_calls;
  case 0:	cont=cont->next;  goto l_callc0;
  case 1:   cont=cont->next;  goto l_callc1;
  case 2:	cont=cont->next;  goto l_callc2;
  case 3: 	cont=cont->next;  goto l_callc3;
  case 4: 	cont=cont->next;  goto l_callc4;
  case 5: 	cont=cont->next;  goto l_callc5;
  }
  goto l_no_call;
}

VMPRIM(call_code, "%call_code", 0)
{
  ip = SCM_CODE_CODE(proc);
  NEXT;
}

VMPRIM(call_proc, "%call_proc", 0)
{
  ip  = SCM_PROC_CODE(proc)->code;
  if (SCM_PROC_CODE(proc)->optargs == 0) {
	COMPLETE_ENV(proc);
	NEXT;
  }
  SCM_ERR("callv: opt args not supported yet", proc);
  /* have optionnal */
  NEXT;
}

VMPRIM(call_closure, "%call_closure", 0)
{
  SCM_Code *code;
  
  code = SCM_PROC_CODE(SCM_CLOSURE_CODE(proc));
  ip  = code->code;
  if (code->optargs != 0) {	/* optionnal arguments */
	SOBJ list, *p, *q;
	if ((q = sp+code->nargs-1) > (SOBJ*)cont)
	  SCM_ERR("callv: bad number of args", proc);
	
	*sp = TOS;  p = (SOBJ*)cont; list = NULL;
	while(--p >= q) list = scm_cons(*p, list);
	if (q == (SOBJ*)cont) {			/* insert space on stack */
	  for (p = sp; p < q; p++) {  p[-1] = p[0]; }
	  q[-1] = list;	sp--;
	} else {
	  p = ((SOBJ*)cont)-1; while(--q >= sp) { *(--p) = *q; }
	  ((SOBJ*)cont)[-1] = list;
	  sp = p;
	}
	TOS=*sp;
  }
  /* sp[0] = TOS; */
  PUSH_ENV_FRAME(SCM_CLOSURE_ENV(proc), code->nargs);
  NEXT;
}


VMPRIM(call_macro, "%call_macro", 0)
{
  if (SCM_MACRO_FUNC(proc)) {
	proc = SCM_MACRO_FUNC(proc);
	goto l_call_closure;
  }
  SCM_ERR("apply macro not implemented", proc);
}

VMPRIM(call_extfunc, "%call_extfunc", 0)
{
  /* don't need supdate(), because last action was poping proc */
  TOS = scm_extfunc_call(proc,
						 (SOBJ*)cont - sp,
						 sp);
  cont = cont->next;
  NEXT;
}

VMPRIM(call_vmfunc, "%call_vmfunc", 0)
{
  SCM_vmRegisters vm;
  void (*func)(SCM_vmRegisters *);

  func = SCM_VMFUNC(proc);
  vm.sp = sp;  vm.ip = ip;  vm.cont = cont;  vm.env  = env;
  (*func)(&vm);
  sp = vm.sp;  ip = vm.ip;  cont = vm.cont;  env = vm.env;
  TOS = *sp;					/* reload tos cache */
  cbinding = NULL; if (env) cbinding = SCM_ENV_FRAME(env)->binding;
  NEXT;
}

/*-- low level error handling */

VMPRIM(catch, "%catch", 1)
{								/* handler tag -- catch-cntxt catch-link cont*/
								/* code: %catch ofs  */
  SOBJ handler, tag, catch;
  SOBJ *next_ip;
  SCM_CatchContext *c;

  spop(tag);  spop(handler);
  next_ip = ip + SCM_INUM(*ip);
  ip++;
  
  catch = scm_mkcatch(tag);
  c = SCM_CATCH_CONTEXT(catch);

  /*-- save everything we want to restore in case of error
   * Note: the saved ip is the ip where we want to resume...
   */
  c->handler = handler;
  c->vm.sp = sp;  c->vm.ip = next_ip;  c->vm.cont = cont;  c->vm.env = env;
  
  spush(catch);
  spush((void*)scm_catch_list);
  scm_catch_list = (void*)sp;

  /*-- complete environement */
  PUSH_CONT();
  cont->env = env;
  cont->ip  = next_ip;
  NEXT;

  SCM_ERR("%catch not implemented yet", NULL);
}

VMPRIM(uncatch, "%uncatch", 0)
{								/* catch-cntxt catch-link res -- res */
  SOBJ res, link, context;
  spop(res);
  spop(link);		scm_catch_list = (void*)link;
  spop(context);
  spush(res);
  NEXT;

  SCM_ERR("%uncatch not implemented yet", NULL);
}

VMPRIM(endlet , "%endlet ", 0)
{
  cont->ip = ip;
} /* FALL THROUGH */

VMPRIM(return, "%return", 0)
{
  if (TOS && SCM_CLOSUREP(TOS)) {
	/*printf("return: OUPS, returning a closure\n"); */
	copy_closure_env(TOS, env);
  }
  return_pop_cont();
  NEXT;
}

VMPRIM(store, "%store", 0)
{								/* value symbol -- value */
  SOBJ sym;
  spop(sym);
  SCM_SYM_VALUE(sym) = TOS;
  TOS = scm_undefined;
  NEXT;
}

VMPRIM(setl, "%setl", 1)
{
  SOBJ p = env;
  short n = SCM_INUM(*ip) >> 16;
  while(--n >= 0) p = SCM_ENV_NEXT(p);
  SCM_ENV_FRAME(p)->binding[(short)SCM_INUM(*ip++)] = TOS;
  NEXT;
}

VMPRIM(setl0, "%setl0", 1)
{
  cbinding[SCM_INUM(*ip++)] = TOS;
  NEXT;
}

VMPRIM(br_and, "%br_and", 1)
{
  if (TOS == scm_false) {
	ip += SCM_INUM(*ip);
	NEXT;
  }
  sdrop();
  ip++;
  NEXT;
}

VMPRIM(br_or, "%br_or", 1)
{
  if (TOS != scm_false) {
	ip += SCM_INUM(*ip);
	NEXT;
  }
  sdrop();
  ip++;
  NEXT;
}

VMPRIM(brif, "%brif", 1)
{
  SOBJ flag;
  spop(flag);
  if (flag != scm_false) {
	ip++;							/* skip jump offset */
	NEXT;
  }
  ip += SCM_INUM(*ip);
  NEXT;
}

VMPRIM(bra, "%bra", 1)
{
  ip += SCM_INUM(*ip);
  NEXT;
}

