/* -*- tab-width: 8; -*-
 *
 * VM3 engine:
 *
 * $Id: vm3.c,v 1.30 2000/11/23 01:51:57 crad Exp $
 *
 */
#ifndef lint
static char vcid[] = "$Id: vm3.c,v 1.30 2000/11/23 01:51:57 crad Exp $";
#endif /* lint */

#include "s.h"

/* VM configuration */

/* define if opcode counter is wanted */

/*#define VM3_PROFILE */
/*#define VM3_PROFILE_HIRES*/
/*#define VM3_DEBUG */
/*#define DEBUG_DICT */

#include "vm3.h"
#include "jit-comp.h"

static SCM_VMD 		scm_vm3_vmd;
void 			*scm_call_table[SOBJ_T_MAX];
static SCM_VM3_PRIM 	*scm_prim_table;

SCM_DICT		*scm_prim_name_dict;
SCM_DICT		*scm_prim_addr_dict;

/****************************************************************
 * Dict tools
 ****************************************************************/

/*** Driver for hash where keys are strings */
Uint scm_dict_string_hash(unsigned char *s)
{
  Uint c, r=0;
  while((c = *s++) != 0)
    r = (r * 31) + c;

  return(r);
}

SCM_DICT_ENTRY *scm_dict_string_newkey(char *key, void *value)
{
  SCM_DICT_ENTRY *de = scm_must_alloc(sizeof(SCM_DICT_ENTRY));

  de->next = NULL;
  de->key = scm_must_strdup(key);
  de->value = value;
  return(de);
}

void scm_dict_string_setkey(SCM_DICT_ENTRY *de, void *value)
{
  de->value = value;
}

void scm_dict_set_string_type(SCM_DICT *dict)
{
  dict->newkey 	= scm_dict_string_newkey;
  dict->setkey 	= scm_dict_string_setkey;
  dict->hash 	= scm_dict_string_hash;
  dict->cmp  	= strcmp;
}

/*** Driver for hash where keys are address */
Uint scm_dict_addr_hash(void *p)
{
  return(((Uint)p) >> 2);
}

SCM_DICT_ENTRY *scm_dict_addr_newkey(void *addr, void *value)
{
  SCM_DICT_ENTRY * de = scm_must_alloc(sizeof(SCM_DICT_ENTRY));
  de->next = NULL;
  de->key = addr;
  de->value = value;
  return(de);
}

void scm_dict_addr_setkey(SCM_DICT_ENTRY *de, void *value)
{
  de->value = value;
}

int scm_dict_addr_cmp(void *x, void *y)
{
  return( (unsigned long)x - (unsigned long)y );
}

void scm_dict_set_addr_type(SCM_DICT *dict)
{
  dict->newkey 	= scm_dict_addr_newkey;
  dict->setkey 	= scm_dict_addr_setkey;
  dict->hash 	= scm_dict_addr_hash;
  dict->cmp  	= scm_dict_addr_cmp;
}

/*** dictionnary */
SCM_DICT *scm_dict_new()
{
  int i;
  SCM_DICT *dict = scm_must_alloc(sizeof(SCM_DICT));
  dict->size = 4;
  dict->nkeys = 0;

  scm_dict_set_string_type(dict);

  dict->maxkeys = dict->size * SCM_DICT_MAX_DEPTH;
  dict->entry = scm_must_alloc(dict->size * sizeof(SCM_DICT_ENTRY *));
  for (i = 0; i < dict->size; i++) dict->entry[i] = NULL;
  return(dict);
}

void scm_dict_dump(SCM_DICT *dict)
{
  SCM_DICT_ENTRY *l;
  int i;
  fprintf(stderr, "dict at %p: size=%d nkeys=%d maxkeys=%d\n",
	  dict, dict->size, dict->nkeys, dict->maxkeys);
  for (i = 0; i < dict->size; i++) {
    fprintf(stderr, "entry[%d]=%p:", i, dict->entry[i]);
    for (l = dict->entry[i]; l; l = l->next) {
      fprintf(stderr, "(%p %p) ", l->key, l->value);
    }
    fprintf(stderr, "\n");
  }
}

SCM_DICT_ENTRY *scm_dict_search(SCM_DICT *dict, void *key)
{
  SCM_DICT_ENTRY *de;
  int hcode;

  hcode = (*dict->hash)(key) % dict->size;
#ifdef DEBUG_DICT
  scm_dict_dump(dict);
  fprintf(stderr, "dict-search: %s (%d)\n", key, hcode);
#endif
  de = dict->entry[hcode];
  while(de != NULL) {
    if ( (*dict->cmp)(key, de->key) == 0)  break;
    de = de->next;
  }
  return(de);
}

SCM_DICT_ENTRY *scm_dict_set(SCM_DICT *dict, void *key, void *value)
{
  SCM_DICT_ENTRY *de;
  int hcode;
#ifdef DEBUG_DICT
  scm_dict_dump(dict);
#endif

  hcode = (*dict->hash)(key) % dict->size;
#ifdef DEBUG_DICT
  fprintf(stderr, "dict-set: key=%s hcode=%d value=%p\n",
	  key, hcode, value);
#endif
  de = dict->entry[hcode];
  while(de != NULL) {
    if ( (*dict->cmp)(key, de->key) == 0) {
      (*dict->setkey)(de, value);
      return(de);
    }
    de = de->next;
  }
  de = (*dict->newkey) (key, value);
  de->next = dict->entry[hcode];
  dict->entry[hcode] = de;
  dict->nkeys++;
  if (dict->nkeys >= dict->maxkeys) {
    SCM_DICT_ENTRY **ref, *l;
    int i, nr, hc;
#ifdef DEBUG_DICT
    fprintf(stderr, "dict_set: must grow (nkeys=%d maxkeys=%d)\n",
	    dict->nkeys, dict->maxkeys);
#endif
    ref = scm_must_alloc(dict->nkeys * sizeof(SCM_DICT_ENTRY*));
    nr = 0;
    for (i = 0; i < dict->size; i++) {
      for (l = dict->entry[i]; l; l = l->next) {
	ref[nr++] = l;
      }
    }
    dict->size *= 4;
    dict->maxkeys = dict->size * SCM_DICT_MAX_DEPTH;
    dict->entry = scm_must_realloc(dict->entry, dict->size*sizeof(SCM_DICT_ENTRY*));
    for (i = 0; i < dict->size; i++)
      dict->entry[i] = NULL;
    for (i = 0; i < nr; i++) {	/* rehash now */
      hc = (*dict->hash)(ref[i]->key) % dict->size;
      ref[i]->next = dict->entry[hc];
      dict->entry[hc] = ref[i];
    }
    scm_free(ref);
  }
  return(de);
}


/****************************************************************
 * Simple profiling
 ****************************************************************/
#ifdef VM3_PROFILE

#ifdef VM3_PROFILE_HIRES

#define READ_TSC(time) \
__asm__ __volatile__ (		\
	".align 8\n"		\
	"pushal\n"		\
	"rdtsc\n"		\
	"movl %%eax,(%0)\n"	\
	"movl %%edx,4(%0)\n"	\
	"popal\n"		\
	: /* no output */ : "S"(&time) : "eax", "edx", "memory")

/* estimated period for a cycle */

static int 		profile_inited=FALSE;
static long		profile_cost = 0;
static double		profile_cycle_period = 1 / 420e6;
static CLOCK_T		profile_vm_clock;

#endif /* VM3_PROFILE_HIRES */


static void profile_start(SCM_VM3_PRIM *p);
static void profile_stop(SCM_VM3_PRIM *p);

static void profile_init(SCM_VM3_PRIM *p)
{
#ifdef VM3_PROFILE_HIRES
  CLOCK_T t0, t1, ts;
  long s1,s2,s3,s4,s5;
  double sec;
  int i;
  struct tms t;

  if (!profile_inited) {

    printf("Init instruction profiler\n");
    for (i = 0; i < 100; i++) {
      p->spent = 0; 	profile_start(p);  profile_stop(p);	s1 = p->spent;
      p->spent = 0; 	profile_start(p);  profile_stop(p);	s2 = p->spent;
      p->spent = 0; 	profile_start(p);  profile_stop(p);	s3 = p->spent;
      p->spent = 0; 	profile_start(p);  profile_stop(p);	s4 = p->spent;
      p->spent = 0; 	profile_start(p);  profile_stop(p);	s5 = p->spent;
      if (s1 == s2 && s1 == s3 && s1 == s4 && s1 == s5)	break;
    }
    printf("s1=%d s2=%d s3=%d s4=%d s5=%d\n", s1,s2,s3,s4,s5);
    if (s1 == s2 && s1 == s3 && s1 == s4 && s1 == s5) {
      printf("  Stable measure cost: %d cycles\n", s1);
    } else {
      int min, max, v[5], nv, i;
      max = min = s1;
      if (s2 > max) max = s2;  if (s2 < min) min = s2;
      if (s3 > max) max = s3;  if (s3 < min) min = s3;
      if (s4 > max) max = s4;  if (s4 < min) min = s4;
      if (s5 > max) max = s5;  if (s5 < min) min = s5;
      nv = 0;
      if (s1 != max && s1 != min)  v[nv++] = s1;
      if (s2 != max && s2 != min)  v[nv++] = s2;
      if (s3 != max && s3 != min)  v[nv++] = s3;
      if (s4 != max && s4 != min)  v[nv++] = s4;
      if (s5 != max && s5 != min)  v[nv++] = s5;
      if (nv == 0) {
	v[nv++] = s1;	v[nv++] = s2;	v[nv++] = s3;
	v[nv++] = s4;	v[nv++] = s5;
      }
      s1 = 0;
      for (i = 0; i < nv; i++) {  s1 += v[i]; }
      s1 /= nv;
      printf("  Unstable measure cost: guess it's %d cycles "
	     "(choosen between", s1);
      for (i = 0; i < nv; i++) printf(" %d", v[i]);
      printf(")\n");
    }
    profile_cost = s1;
    if (profile_cycle_period == 0.0) {
      sec = 5;
      READ_TSC(t0);
      sleep(sec);
      READ_TSC(t1);
      ts = (t1 - t0);
      printf("%Ld cycles for %g seconds, %.3f Mhz, cycle is %g sec\n",
	     ts, sec, ts / (sec * 1e6), sec / ts);
      profile_inited = TRUE;
      profile_cycle_period = sec / ts;
    } else {
      printf("  CPU clock beating at %.3f Mhz\n",
	     1 / (profile_cycle_period * 1e6));
    }
  }
#endif /* VM3_PROFILE_HIRES */
  
  while(p->name) {
    p->count = 0;
    p->spent = 0;
    p++;
  }
}

static void profile_dump(SCM_VM3_PRIM *p)
{
  double usec;
#ifdef VM3_PROFILE_HIRES
  CLOCK_T tot = 0;
#endif
  
  fprintf(stderr, "Profiling result: resolution %g Mhz\n", (double)CLOCK_FQ / 1e6);

  while(p->name) {

    if (p->count) {
#ifdef VM3_PROFILE_HIRES
      fprintf(stderr, "%-16s: %-8d %10Ld cycles "
	      "(%-3d cycles, %.3f nsec /call)\n",
	      p->name, p->count,
	      p->spent,
	      (int)(p->spent / p->count),
	      p->spent * 1e9 / (CLOCK_FQ * p->count));
      tot += p->spent;
#else /*VM3_PROFILE_HIRES*/
      usec = (double)p->spent * 1000000 / CLOCK_FQ;
      fprintf(stderr, "%-16s: %-8d %.3f usec (%.3f nsec/call)\n",
	      p->name, p->count, usec, usec * 1000 / p->count);
#endif /*VM3_PROFILE_HIRES*/
    }
    p++;
  }
#ifdef VM3_PROFILE_HIRES
  fprintf(stderr, "VM exec   time: %10Ld cycles, %g sec\n",
	  profile_vm_clock, profile_vm_clock / CLOCK_FQ);

  fprintf(stderr, "VM opcode time: %10Ld cycles, %g sec\n",
	  tot, tot / CLOCK_FQ);

  fprintf(stderr, "VM overhead:    %10Ld cycles, %g sec\n",
	  profile_vm_clock - tot, (profile_vm_clock - tot) / CLOCK_FQ);
#endif
  
}

static void profile_start(SCM_VM3_PRIM *p)
{
#ifdef VM3_PROFILE_HIRES
  p->count++;
  READ_TSC(p->start_clock);
#else /*VM3_PROFILE_HIRES*/
  struct tms t;
  times(&t);
  p->count++;
  p->start_clock = t.tms_utime;
#endif /*VM3_PROFILE_HIRES*/
}

static void profile_stop(SCM_VM3_PRIM *p)
{
#ifdef VM3_PROFILE_HIRES
  CLOCK_T now;
  READ_TSC(now);
  p->spent += (now - p->start_clock) - profile_cost;
#else /*VM3_PROFILE_HIRES*/
  struct tms t;
  times(&t);
  p->spent += t.tms_utime - p->start_clock;
#endif /*VM3_PROFILE_HIRES*/
}

#endif /*VM3_PROFILE*/

/****************************************************************
 * Debugging / Trace tools
 ****************************************************************/
void scm_vm3_dump_env(SCM_ENVFRAME *e)
{
  int i, ns;
  while(e) {
    ns = SCM_INUM(e->nslots);
    printf("env %p: nslots=%d\n", e, ns);
    if (ns > 10) ns = 10;
    for (i = 0; i < ns; i++) {
      printf("  slot[%d] : ", i); scm_cprint(e->slot[i]);
    }
    if (ns != SCM_INUM(e->nslots))
      printf("  ...\n");

    printf("\n");
    e = e->next;
  }
}

void scm_vm3_dump_cont(CONTEXT *c)
{
  while(c) {
    printf("context %p: next=%p env=%p ip=%p\n",
	   c, c->next, c->env, c->ip);
    c = c->next;
  }
}

void scm_vm3_dump_stack(SOBJ *sp)
{
  SOBJ *p;
  p = scm_vm3_vmd.stack_base;
  printf("Stack dump:\n");
  while(p <= sp) {
    printf("\t%08lx: %08lx\t", (long)p, (long)*p);
    if ((void*)*p >= (void*)scm_vm3_vmd.stack_base &&
	(void*)*p <  (void*)scm_vm3_vmd.stack_limit) {
      printf("<stack ref>\n");
    } else {
      SCM_VM3_PRIM *te = scm_prim_table;
      void *addr = *p;
      while (te->name && te->start != addr)  te++;
      if (te->name) {
	printf("#<prim %s>\n", te->name);
      } else {
	scm_cprint(addr);
      }
    }
    p++;
  }
}

/****************************************************************
 * vm3 search 
 ****************************************************************/

SCM_VM3_PRIM *scm_vm3_search_name(SCM_VM3_PRIM *t, char *name)
{
  /*  fprintf(stderr, "scm_vm3_search_name: searching %s\n", name); */
  while(t->name != NULL) {
    /* fprintf(stderr, "scm_vm3_search_name: testing %s\n", t->name); */
    if (strcmp(t->name, name) == 0)	return(t);
    t++;
  }
  return(NULL);
}

SCM_VM3_PRIM *scm_vm3_search_addr(SCM_VM3_PRIM *t, void *addr)
{
  /*  fprintf(stderr, "scm_vm3_search_addr: searching %p\n", addr); */
  while(t->name != NULL) {
    /* fprintf(stderr, "scm_vm3_search_addr: testing %p\n", t->start); */
    if (t->start == addr)	return(t);
    t++;
  }
  return(NULL);
}


/****************************************************************
 * vm3 callbacks
 ****************************************************************/

void scm_vm3_engine_init(SCM_VMD *vm, SCM_VM3_PRIM *t)
{
  SCM_VM3_PRIM *te;
  int i;

  fprintf(stderr, "scm_vm3_engine_init:\n");

  scm_prim_table = t;
  
  fprintf(stderr, "scm_vm3_engine_init: stack init\n");

  vm->stack_size = 4096;
  vm->stack_base = scm_must_alloc(sizeof(SOBJ)*vm->stack_size);
  vm->stack_limit = vm->stack_base + vm->stack_size;

  fprintf(stderr, "scm_vm3_engine_init: regs init\n");
  vm->reg.sp = vm->stack_base;
  vm->reg.ip = NULL;
  vm->reg.env = NULL;
  vm->reg.cont= NULL;

  vm->catch_list = NULL;
  vm->thrown_tag = NULL;
  vm->thrown_msg = NULL;


  fprintf(stderr, "scm_vm3_engine_init: setup of call table\n");
  for (i = 0; i < SOBJ_T_MAX; i++) {
    scm_call_table[i] = t[SCM_OP_NOP].start; /* default to nop */
  }

  scm_call_table[SOBJ_T_PRIM] 	= t[SCM_OP_GEN_CALLPRIM].start;
  scm_call_table[SOBJ_T_CPRIM] 	= t[SCM_OP_GEN_CALLCPRIM].start;
  scm_call_table[SOBJ_T_CODE] 	= t[SCM_OP_GEN_CALLCODE].start;
  scm_call_table[SOBJ_T_MCODE] 	= t[SCM_OP_GEN_CALLMCODE].start;
  scm_call_table[SOBJ_T_PROC] 	= t[SCM_OP_GEN_CALLPROC].start;
  scm_call_table[SOBJ_T_CLOSURE]= t[SCM_OP_GEN_CALLCLOS].start;
  scm_call_table[SOBJ_T_MACRO] 	= t[SCM_OP_GEN_CALLMACRO].start;

  /* check alignement of instruction in t */
  for (te = t; te->name; te++) {
    if (((long)te->start % 2) != 0) {
      fprintf(stderr, "scm_vm3_engine_init: align of %s (%p) is not odd\n",
	      te->name, te->start);
    }
  }
  
  /* build a dictionnary with key=(char*)name and value=(SCM_VM3_PRIM*)
   * for later use in assembler.
   */
  scm_prim_name_dict = scm_dict_new();
  for (te = t; te->name; te++) {
    scm_dict_set(scm_prim_name_dict, te->name, te);
  }

  /* build dictionnary to map opcode address with prim table entry */
  scm_prim_addr_dict = scm_dict_new();
  scm_dict_set_addr_type(scm_prim_addr_dict);
  
  for (te = t; te->name; te++) {
    scm_dict_set(scm_prim_addr_dict, te->start, te);
  }

  scm_jc_install_prim_compiler(t);
}


/*** straight forward assembler */
/* The only special stuff about this assembler is that it reconizes
 * this special constructs:
 *   (%lab <lab>)
 *   (%bra|%brf|%brt <lab>)
 *   (%asm <array>)
 *
 * Labels are resolved using one pass logic.
 * %asm recursively assembles a piece of code.
 */
#define SCM_LABREF 	struct _SCM_LABREF
#define SCM_LABNODE	struct _SCM_LABNODE

SCM_LABREF {
  int resolved;
  SOBJ lab;
  union {
    int		ofs;		/* if resolved */
    SCM_LABNODE *node;		/* if ! resolved */
  } ref;
};

SCM_LABNODE {
  SCM_LABNODE 	*next;
  int		ofs;
};

static void labnode_new(SCM_LABREF *lr, int ofs)
{
  SCM_LABNODE *node;
  node = scm_must_alloc(sizeof(SCM_LABNODE));
  node->next = lr->ref.node;    lr->ref.node = node;
  node->ofs = ofs;
}

static SCM_LABREF *label_search(SCM_LABREF *lr, int count, SOBJ lab)
{
  while(--count >= 0) {
    if (lr->lab == lab) return(lr);
    lr++;
  }
  return(NULL);
}

static void label_resolve(SCM_LABREF *lr, int ofs, SOBJ code)
{
  SCM_LABNODE *ln, *node;
  if (lr->resolved) {
    /* should already have been used */
    fprintf(stderr, "asm:error: label already resolved\n");
    exit(1);
  }
  ln = lr->ref.node;
  while(ln != NULL) {		/* scan chain of labels to resolve */
    SCM_AREF(code, ln->ofs) = SCM_MKINUM(ofs - ln->ofs);
    node = ln;  ln = ln->next;
    scm_free(node);
  }
  lr->resolved = TRUE;
  lr->ref.ofs = ofs;
}

static SCM_LABREF *label_add_def(SCM_LABREF **lr, int *count, int *alloc,
				 int flag, SOBJ lab)
{
  SCM_LABREF *lp;
  int n;

  n = *count;
  
  if ((n+1) >= *alloc) {
    *alloc += 128;
    *lr = scm_must_realloc(*lr, *alloc * sizeof(SCM_LABREF));
  }
  lp = *lr + n;
  lp->resolved = flag;
  lp->lab = lab;
  lp->ref.node = NULL;
  *count = n + 1;
  return(lp);
}

static void labref_zap(SCM_LABREF *lr, int count)
{
  int i;
  SCM_LABNODE *ln, *node;
  for (i = 0; i < count; i++) {
    if (!lr[i].resolved) {
      ln = lr[i].ref.node;
      while (ln) {
	node = ln; ln = ln->next; scm_free(node);
      }
    }
  }
  scm_free(lr);
}

/* This assembler takes a source code array and build a VM code array.
 */
static SOBJ scm_vm3_engine_asm(SOBJ code, SCM_VM3_PRIM *t, SOBJ env, int inl)
{
  SOBJ inst, opc, list;
  SOBJ out;
  SCM_DICT_ENTRY *de;
  SCM_VM3_PRIM *te;
  SCM_LABREF *labref, *lr;
  int 	labref_alloc;
  int	labref_count;
  int 	i, size, cp, op;
  char *str;

  if (!SCM_ARRAYP(code))	SCM_ERR("bad code", code);

  labref_alloc = 128;
  labref_count = 0;
  labref = scm_must_alloc(labref_alloc * sizeof(SCM_LABREF));

  out = scm_mkarray(0, scm_false);
  for (cp = 0; cp < SCM_ASIZE(code); cp++) {
    inst = SCM_AREF(code, cp);
    if (!SCM_PAIRP(inst))	SCM_ERR("bad instruction", inst);

    opc = SCM_CAR(inst);
    str = SCM_ATOM_NAME(opc);

    /* handle special cases */
    if (streq(str, "%lab")) {	/* label definition */
      if ((lr = label_search(labref, labref_count, SCM_CADR(inst))) != NULL) {
	/* resolve location where this label was used */
	label_resolve(lr, SCM_ASIZE(out), out);
      } else {
	/* new label, never used before */
	lr = label_add_def(&labref, &labref_count, &labref_alloc,
			   TRUE, SCM_CADR(inst));
	lr->ref.ofs = SCM_ASIZE(out);
      }
      continue;
    }

    /* assemble some code, generate a code object and push a ref to it */
    if (streq(str, "%proc")) {	/* something to assemble */
      SOBJ list;
      SOBJ newcode, newenv, nargs, optargs;
      SOBJ proc;
      
      list= SCM_CDR(inst);
      if (list == NULL) 	SCM_ERR("%proc: syntax", inst);

      if (!SCM_PAIRP(list))	SCM_ERR("%proc: missing code param", NULL);
      newcode = SCM_CAR(list);	list = SCM_CDR(list);
      if (!SCM_PAIRP(list))	SCM_ERR("%proc: missing env param", NULL);
      newenv = SCM_CAR(list);	list = SCM_CDR(list);
      if (!SCM_PAIRP(list))	SCM_ERR("%proc: missing nargs param", NULL);
      nargs  = SCM_CAR(list);	list = SCM_CDR(list);
      if (!SCM_PAIRP(list))	SCM_ERR("%proc: missing optarg param", NULL);
      optargs= SCM_CAR(list);	list = SCM_CDR(list);
	

      proc  = scm_make_proc(scm_vm3_engine_asm(newcode, t, newenv, inl),
			    newenv, nargs, optargs);

      size = SCM_ASIZE(out);
      out = scm_vector_resize(out, SCM_MKINUM(size + 2));
      SCM_AREF(out, size++) = t[SCM_OP_PUSHLIT].start;
      SCM_AREF(out, size++) = proc;
      continue;
    }

    if ((de = scm_dict_search(scm_prim_name_dict, str)) == NULL) {
      labref_zap(labref, labref_count);
      SCM_ERR("instr not found", opc);
    }
    te = de->value;
    op = te - t;

    /* Special wrapper around setsym, pushsym, callcprim[0-10]* */
    /* %setsym, %pushsym and %callcprim argument may be a quoted
     * symbol (atom) or a symbol reference. If it's an atom, we have
     * to search for the real symbol and create it if it does not
     * exist.
     *
     * This code replaces the ref to atom with ref to symbols in the
     * instruction list */
    if (op == SCM_OP_SETSYM || op == SCM_OP_PUSHSYM) {
      SOBJ sym, obj;
      obj = SCM_CADR(inst);
      if (SCM_ATOMP(obj)) {	/* atom found */
	sym = scm_hash_search(scm_symbol_hash, obj);
	if (sym == scm_false) {
	  scm_hash_set(scm_symbol_hash, obj, scm_unbound);
	  sym = scm_hash_search(scm_symbol_hash, obj);
	}
	SCM_CADR(inst) = sym;
      }
    } else if (op == SCM_OP_SETEXT || op == SCM_OP_PUSHEXT) {
      SOBJ sym, obj;
      obj = SCM_CADR(inst);
      if (SCM_ATOMP(obj)) {	/* atom found */
	sym = scm_hash_search(scm_symbol_hash, obj);
	if (sym == scm_false) SCM_ERR("bad var", obj);
	SCM_CADR(inst) = SCM_SYM_VALUE(sym);
      }
    } else if (op >= SCM_OP_CALLCPRIM && op <= SCM_OP_CALLCPRIM10) {
      SOBJ sym, obj;
      obj = SCM_CADR(inst);
      if (SCM_ATOMP(obj)) {	/* atom found */
	sym = scm_hash_search(scm_symbol_hash, obj);
	if (sym == scm_false) SCM_ERR("bad cprim", obj);
	SCM_CADR(inst) = SCM_SYM_VALUE(sym);
      }
    } else if (op == SCM_OP_BRA || op == SCM_OP_BRF ||
	       op == SCM_OP_BRT ||
	       op == SCM_OP_BR_AND || op == SCM_OP_BR_OR ||
	       op == SCM_OP_CATCH
	       ) { /* label consumer */

      size = SCM_ASIZE(out);
      out = scm_vector_resize(out, SCM_MKINUM(size + te->follow +1));
      SCM_AREF(out, size++) = (void*)te->start;
      if ((lr = label_search(labref, labref_count, SCM_CADR(inst))) != NULL) {
	/* already defined */
	if (lr->resolved) {
	  SCM_AREF(out, size) = SCM_MKINUM(lr->ref.ofs - size);
	} else {
	  labnode_new(lr, size);
	}
      } else {
	/* new label, never seen before */
	lr = label_add_def(&labref, &labref_count, &labref_alloc,
			   FALSE, SCM_CADR(inst));
	labnode_new(lr, size);
	SCM_AREF(out, size  ) = NULL;
      }
      continue;
    }
    size = SCM_ASIZE(out);
    out = scm_vector_resize(out, SCM_MKINUM(size + te->follow +1));
    SCM_AREF(out, size++) = (void*)te->start;

    list = SCM_CDR(inst);
    for (i = 0; i < te->follow; i++) {
      if (!SCM_PAIRP(list)) SCM_ERR("not enough args for", opc);
      SCM_AREF(out, size++) = SCM_CAR(list);
      list = SCM_CDR(list);
    }
  }
  labref_zap(labref, labref_count); /* no leak please */
  return(out);
}

/********************************
 * Disassembler
 ********************************/
static SOBJ scm_vm3_engine_dis(SOBJ code, SCM_VM3_PRIM *t)
{
  int i, n, len;
  SCM_VM3_PRIM *te;
  SOBJ *pcode;

  if (SCM_ARRAYP(code)) {
    pcode = SCM_ARRAY(code);
    len   = SCM_ASIZE(code);
  } else if (SCM_CLOSUREP(code)) {
    pcode = SCM_PROC_CODE(SCM_CLOSURE_CODE(code))->code;
    len   = SCM_PROC_CODE(SCM_CLOSURE_CODE(code))->size;
  } else {
    pcode = NULL;
    len = 0;
    SCM_ERR("bad code", code);
  }
  
  i = 0;
  while(i < len) {
    te = scm_vm3_search_addr(t, pcode[i]);
    if (te == NULL) {
      fprintf(stderr, "scm_vm3_dis: no opcode matching %p\n", pcode[i]);
      SCM_ERR("bad opcode", NULL);
    }
    printf("%p: %s ", pcode + i, te->name);
    i++;
    n = te->follow;
    while(--n >= 0) {
      scm_cdisplay(pcode[i]); printf(" ");
      i++;
    }
    printf("\n");
  }
  return(scm_undefined);
}

/********************************
 * VM3 Env utils
 ********************************/
static SCM_ENVFRAME *env_clone(SCM_ENVFRAME *env)
{
  SCM_ENVFRAME *new;
  int nbytes;
  
  nbytes = offsetof(SCM_ENVFRAME, slot[SCM_INUM(env->nslots)]);
  new = scm_must_alloc(nbytes);
  memcpy(new, env, nbytes);
  fprintf(stderr, "VM3:env_clone: cloning env %p -> %p\n", env, new);
  return(new);
}

/* External function used inside the VM for opcode that may be copied.
 */
static SCM_ENVFRAME *scm_env_deep_copy(SCM_ENVFRAME *env,
				       void *sbase,
				       void *slimit)
{
  SCM_ENVFRAME *e, *new, *envchain, **ptr;
  if ((void*)env >= sbase && (void*)env <= slimit) {
#ifdef VM3_DEBUG
    fprintf(stderr, "VM3:scm_env_deep_copy: starting with env %p\n", env);
#endif
    envchain = NULL;
    ptr = &envchain;
    for (e = env; e; e = e->next) {
      new = env_clone(e);
      *ptr = new;
      ptr = &new->next;
    }
  } else {
    fprintf(stderr, "VM3:scm_env_deep_copy: env %p is not on stack\n", env);
    envchain = env;
  }
  return(envchain);
}

SOBJ scm_vm3_make_closure(SCM_VMD *vm, SCM_ENVFRAME *env, SOBJ *sp, SOBJ proc)
{
  SOBJ closure = scm_newcell(SOBJ_T_CLOSURE);

  SCM_CLOSURE_CODE(closure) = proc;
  SCM_CLOSURE_ENV(closure) = scm_env_deep_copy(env, vm->stack_base, sp);
  return(closure);
}



/* Loading reference to function in external variable force the
 * compiler to generate absolute addressing for calls.
 *
 * For example call like: TOS = (*scm_vm3_print2)(...) will be
 * compiled with absolute addressing and this code can be copied and
 * executed to other memory location */

SOBJ (*scm_vm3_print2)() = scm_print2;
SOBJ (*scm_vm3_write2)() = scm_write2;
SOBJ (*scm_vm3_newcell)() = scm_newcell;


/*** Initialize VM3 engine */
void scm_vm3(SCM_VMD *vm)
{
  register SOBJ		TOS	TOSREG;	/* Top of stack caching register */
  register SOBJ 	*sp	SPREG;	/* Stack pointer */
  register SOBJ 	*ip	IPREG;	/* Instruction pointer */
  register CONTEXT 	*cont	CREG;	/* Partial continuation */
  register SCM_ENVFRAME *env	EREG;	/* Env chain */

#ifdef VM3_PROFILE_HIRES
  CLOCK_T	tstart, tend;
#endif /* VM3_PROFILE_HIRES */

  static SCM_VM3_PRIM symbol[] = {
    /* name             arity/follow/term */
    DECL(nop,		0, 0, 0),
    DECL(end, 		0, 0, 1),

    DECL(load_env,	0, 0, 0),
    DECL(load_cont,	0, 0, 0),
    
    DECL(drop,		0, 0, 0),

    DECL(pushlit,	0, 1, 0),
    DECL(pushsym,	0, 1, 0),
    DECL(pushloc,	0, 1, 0),
    DECL(pushloc0,	0, 1, 0),
    
    DECL(pushext,	0, 1, 0),

    DECL(setsym,	1, 1, 0),
    DECL(setloc,	1, 1, 0),
    DECL(setloc0,	1, 1, 0),

    DECL(setext,	1, 1, 0),

    DECL(pushenv,	0, 0, 0),
    DECL(popenv,	0, 0, 0),

    DECL(bra,		0, 1, 0),
    DECL(brf,		1, 1, 0),
    DECL(brt,		1, 1, 0),

    DECL(br_and,	1, 1, 0),
    DECL(br_or,		1, 1, 0),

    DECL(prepcall,	0, 0, 0),
    DECL(makeclos,	1, 0, 0),
    DECL(callclos,	0, 0, 0),
    DECL(return,	0, 0, 0),

    DECL(letstart,	0, 0, 0),
    DECL(letbody,	0, 1, 0),
    DECL(letend,	0, 1, 0),
    DECL(letstar,	0, 1, 0),
    DECL(letrec,	0, 1, 0),

    DECL(gen_callprim,	-1, 0, 0),
    DECL(gen_callcprim,	-1, 0, 0),
    DECL(gen_callcode,	-1, 0, 0),
    DECL(gen_callmcode,	-1, 0, 0),
    DECL(gen_callproc,	-1, 0, 0),
    DECL(gen_callclos,	-1, 0, 0),
    DECL(gen_callmacro,	-1, 0, 0),

    DECL(call,		1, 0, 0),
    DECL(jump,		1, 1, 0),

    DECL(callcprim,	-1, 2, 0),
    DECL(callcprim0,	0, 1, 0),
    DECL(callcprim1,	1, 1, 0),
    DECL(callcprim2,	2, 1, 0),
    DECL(callcprim3,	3, 1, 0),
    DECL(callcprim4,	4, 1, 0),
    DECL(callcprim5,	5, 1, 0),
    DECL(callcprim6,	6, 1, 0),
    DECL(callcprim7,	7, 1, 0),
    DECL(callcprim8,	8, 1, 0),
    DECL(callcprim9,	9, 1, 0),
    DECL(callcprim10,	10, 1, 0),

    /* Catch block */
    DECL(catch,		2, 1, 0),
    DECL(uncatch,	0, 0, 0),

    DECL(vm3_exit,	0,0,1),

#include "prim3.x"

    {NULL}
  };

  if (vm->code != SCM_VM3_EXECUTE && vm->code != SCM_VM3_MCODE_EXECUTE) {
    switch(vm->code) {
    case SCM_VM3_INIT:
      scm_vm3_engine_init(vm, symbol);
      break;
    case SCM_VM3_ASSEMBLE:
      vm->ret.obj = scm_vm3_engine_asm(vm->arg.addr, symbol, NULL, 0);
      break;
    case SCM_VM3_NASSEMBLE:
      vm->ret.obj = scm_vm3_engine_nasm(vm->arg.addr, symbol);
      break;
    case SCM_VM3_DISASSEMBLE:
      vm->ret.obj = scm_vm3_engine_dis(vm->arg.addr, symbol);
      break;
    }
    return;
  }

#ifdef VM3_PROFILE_HIRES
  READ_TSC(tstart);
#endif /* VM3_PROFILE_HIRES */
  
  /* load initial register state */
  ip 	= vm->reg.ip;
  cont 	= (CONTEXT*)vm->reg.cont;
  env  	= (SCM_ENVFRAME*)vm->reg.env;
  sclear();

  if (vm->code == SCM_VM3_MCODE_EXECUTE) {
    spush(vm->arg.obj);
    goto mcode_execute;
  }
  
  START_VMCODE
  INSTR(nop) {	END(nop); }

  INSTR(end) {
    spop(vm->ret.obj);
    sresync();
#ifdef VM3_PROFILE_HIRES
    READ_TSC(tend);
    profile_vm_clock = tend - tstart;
#endif /* VM3_PROFILE_HIRES */
    PROFILE_DUMP();
    return;
    END(end);
  }

  /* Pseudo instructions for JIT compiler to load vm registers that
   * have not been allocated in CPU registers */

  /* The following instruction ensure a relatively portable way to
   * load the jit VM_TMP register with the value of whatever VM
   * register
   *
   * I need this trick because I don't know the address (relative to
   * FP) on the stack where GCC will put the non vm registers.  */
  
  INSTR(load_env) {
    LOAD_ENV();
    END(load_env);
  }

  INSTR(load_cont) {
    LOAD_CONT();
    END(load_cont);
  }

  /* Regular engine */
  
  INSTR(drop) {
    sdrop();
    END(drop);
  }

  /* Regular instruction */
  INSTR(pushlit) {
    spush(*ip++);
    END(pushlit);
  }

  INSTR(pushsym) {
    spush(SCM_SYM_VALUE((SOBJ)(*ip++)));
    //if (TOS == scm_unbound) SCM_ERR("symbol unbound", SCM_SYM_NAME(ip[-1]));
    END(pushsym);
  }

  INSTR(pushloc) {		/* -- value */
    /* nible is: LLII */
    long  nibble = SCM_INUM(*ip++);
    short depth = nibble >> 16;
    SCM_ENVFRAME *e = env;
    while(--depth >= 0) e = e->next;
    spush(e->slot[(nibble & 0xffff)]);
    END(pushloc);
  }

  /*** speedup */
  INSTR(pushloc0) {
    spush(env->slot[SCM_INUM(*ip++)]);
    END(pushloc0);
  }

  INSTR(pushext) {		/* -- value */
    SOBJ var = *ip++;
    supdate();
    spush( (*(SCM_VAR_AUX(var)->get))(var,NULL) );
    END(pushext);
  }

  INSTR(setsym) {		/* value -- value */
    SOBJ var = *ip++;
    SCM_SYM_VALUE(var) = TOS;
    END(setsym);
  }

  INSTR(setloc) {		/* value -- value */
    long  nibble = SCM_INUM(*ip++);
    short depth = nibble >> 16;
    SCM_ENVFRAME *e = env;
    while(--depth >= 0) e = e->next;
    e->slot[(nibble & 0xffff)] = TOS;
    END(setloc);
  }
  
  INSTR(setloc0) {		/* value -- value */
    env->slot[SCM_INUM(*ip++)] = TOS;
    END(setloc0);
  }

  INSTR(setext) {		/* value -- value */
    SOBJ var = *ip++;
    supdate();
    (*(SCM_VAR_AUX(var)->set))(var,NULL,TOS);
    TOS = *sp;
    END(setext);
  }


  /* 	Stack layout for pushenv
   *	Before			After
   *    			[ ??? ]	<-sp		.slot[0]
   *    		        [ ??? ]			.nslots
   *	             	     +--+--*  ] <-env		.next
   *	[ ... ]	<- sp	     |  [ ... ]		
   *	[ ... ]		     |  [ ... ]		
   *	[ ... ] <--env	     +->[ ... ]
   */
  INSTR(pushenv) {
    spush((void*)env);		/* create env->next */
    env = (void*)sp;		/* set new env ptr */
    spush(0);			/* push nslots */
    END(pushenv);
  }

  /*	Stack layout for popenv
   *	Before			After
   *    [ ... ] 
   *    [ v#1 ] <-sp
   *    [ v#0 ]
   *    [ l#1 ]
   *    [ l#0 ]
   *    [  2  ]			[ v#1 ]<-sp
   * +--+--*. ] <-env		[ v#0 ]
   * |  [ ... ]		       	[ ... ]
   * |  [ ... ]		       	[ ... ]
   * +->[ ... ]			[ ... ]<-env
   */
  INSTR(popenv) {
    SCM_ENVFRAME *e = env->next;
#ifdef MULTI_VALUE
    /* copy results from stack within [env->slot + env->nslot, sp] to env */
    SOBJ *s, *d, *l;

    d = (SOBJ*)env;
    s = env->slot + SCM_INUM(env->nslots);
    l = sp;
    while(s < l) { *d++ = *s++; }
    sp = d - 1; TOS = *sp;
    env = e;
#else /* SINGLE_VALUE */
    SOBJ r;
    r = TOS;    sp = (SOBJ*)env;     *sp = r;
    env = e;
#endif
    END(popenv);
  }

  /* Branches */
  INSTR(bra) {
    ip += SCM_INUM(*ip);
    END(bra);
  }

  INSTR(brf) {
    SOBJ flag;
    spop(flag);
    if (flag == scm_false) {
      ip += SCM_INUM(*ip);
    } else {
      ip++;
    }
    END(brf);
  }

  INSTR(brt) {
    SOBJ flag;
    spop(flag);
    if (flag != scm_false) {
      ip += SCM_INUM(*ip);
    } else {
      ip++;
    }
    END(brt);
  }

  INSTR(br_and) {
    if (TOS == scm_false) {
      ip += SCM_INUM(*ip);
    } else {
      sdrop();
      ip++;
    }
    END(br_and);
  }

  INSTR(br_or) {
    if (TOS != scm_false) {
      ip += SCM_INUM(*ip);
    } else {
      sdrop();
      ip++;
    }
    END(br_or);
  }
  

  /*	Stack layout for prepcall
   *	Before			After
   *	[ ... ]			[ ... ]<-sp
   *	[ ... ]		  +-----+--*. ]<-env
   *	[ ... ]<-sp	  |  +--+--*. ]<-cont
   *	~ ... ~		  |  |	~ ... ~
   *	[ ... ]<-env	  +--+->[ ... ]
   *	~ ... ~		     |	~ ... ~
   *	[ ... ]<-cont	     +->[ ... ]
   */
  INSTR(prepcall) {
    CONTEXT *c;
    supdate();
    c 	    = (CONTEXT*)(sp+1);
    c->next = cont;
    cont    = c;
    sp	    = ((void*)c) + (sizeof(CONTEXT) + offsetof(SCM_ENVFRAME, nslots));
    TOS	    = SCM_INUM0;
    END(prepcall);
  }

  INSTR(makeclos) {
    TOS = scm_vm3_make_closure(vm, env, sp, TOS);
    env = SCM_CLOSURE_ENV(TOS);
    END(makeclos);
  }
  
  INSTR(callclos) {
    SOBJ closure;
    SCM_Code *code;
    SOBJ *init_sp = sp;

    spop(closure);
    cont->ip    = ip;
    cont->env 	= env;

    env		= ((void*)cont) + sizeof(CONTEXT);
    env->next   = SCM_CLOSURE_ENV(closure);

    code 	= SCM_PROC_CODE(SCM_CLOSURE_CODE(closure));

    if (init_sp - env->slot < code->nargs - code->optargs)
      SCM_ERR("callclos: wrong number of arguments", closure);

    env->nslots = SCM_MKINUM(code->nargs);

    if (code->optargs) {	/* optionnal arguments */
      SOBJ list = NULL, *p = env->slot + (code->nargs-1);
      while(sp >= p) {
	list = scm_cons(*sp--, list);
      }
      TOS = list; sp++;
    }
    ip   	= code->code;
    END(callclos);
  }

  INSTR(return) {
    SOBJ val;
    val	= TOS;
    ip 	= cont->ip;
    env = cont->env;
    sp 	= (void*)cont;
    cont= cont->next;
    TOS = val;
    END(return);
  }

  INSTR(letstart) {
    supdate();
    sp += 2;			/* reserve space for SCM_ENVFRAME header */
    TOS = *sp;
    END(letstart);
  }
  
  /* letbody <nslots>
   *
   * The letbody instr has the number of slot in env as argument.
   * With nslots I can calculate the address of the env. Env is at
   * sp - nslots - 1
   *
   * Before
   * 	[ v#1 ]<-sp
   * 	[ v#0 ]
   * 	[ ??? ]   	(sp - nslots)
   * 	[ ??? ]         (sp - nslots - 1)
   *
   * After
   * 	[ v#1 ]<-sp
   * 	[ v#0 ]
   * 	[  2  ]
   * +--+--*  ]<-env
   * |	    
   * +-> previous env
   */
  INSTR(letbody) {		/* nslots */
    SOBJ ns;
    SCM_ENVFRAME *ep;
    ns 		= *ip++;
    ep 		= (SCM_ENVFRAME*)(sp - (SCM_INUM(ns) + 1));
    ep->next   	= env;		/* build envframe header */
    ep->nslots 	= ns;
    env 	= ep;
    END(letbody);
  }
  /* letend - pop the env and push the result
   *
   * Before
   *	[ res ] <-sp
   *    [ v#1 ]
   *	[ v#0 ]
   * 	[  2  ]
   * 	[ env ] <-env
   *
   * After
   *	[ res ]
   */
  INSTR(letend) {
    int n = SCM_INUM(*ip++);
    SOBJ result = TOS;
    sp  = sp - (n + ENV_HEADER_SLOTS);
    env = env->next;
    TOS	= result;
    END(letend);
  }

  /* letstar <nslots>
   */
  INSTR(letstar) {				/* just push env frame */
    SOBJ ns, *p, *l;

    spush((SOBJ)env);
    env = (SCM_ENVFRAME*)sp;
    ns  = *ip++;
    spush(ns);
    for (p = env->slot, l = env->slot + SCM_INUM(ns); p < l; p++)
      *p = scm_unbound;

    END(letstar);
  }

  /* letrec <nslots>
   */
  INSTR(letrec) {
    SOBJ ns = *ip++;
    SOBJ *l;
    spush((SOBJ)env);	env = (SCM_ENVFRAME*)sp;
    spush(ns);
    l = env->slot + (SCM_INUM(ns) - 1);
    while(sp < l) {  spush(scm_unbound); }

    END(letrec);
  }
  
  /* Generic call functions, to be used by %call
   * Stack before:
   * [ cont envframe arg1 arg2 .. argn func ]
   * Stack after:
   * [ result ]
   */
  
  INSTR(gen_callprim) {
    SOBJ prim;
    SCM_VM3_PRIM *te;
    spop(prim);
    te = (SCM_VM3_PRIM*)SCM_PRIM(prim);
    if (te->arity < 0) {	/* prim with variable number of args */
      /* Algo:
       * pop the cont 
       * move the env to the location where cont was
       * jump to primitive */
      SOBJ *src, *dst;
      src = (SOBJ*)env;
      dst = (SOBJ*)cont;
      cont = cont->next;
      env  = (SCM_ENVFRAME*)dst;
      while(src < sp) {  *dst++ = *src++; }
      sp = dst-1;
      TOS = *sp;
      goto *(te->start);

    } else {
      /* Algo:
       * pop the cont and the env
       * move the arguments where the cont was
       * Jump to primitive
       */
      SOBJ *src, *dst;
      src = env->slot;
      dst = (SOBJ*)cont;
      cont = cont->next;
      env  = env->next;
      while(src < sp) {  *dst++ = *src++; }
      sp = dst-1;
      TOS = *sp;
      goto *(te->start);
    }
    END(gen_callprim);
  }
  INSTR(gen_callcprim) {
    SOBJ func, r, *a;
    spop(func);
    a = env->slot;
    switch(SCM_CPRIM_NARGS(func)) {
    case 0: r=(*SCM_CPRIM_FUNC(func))(); break;
    case 1: r=(*SCM_CPRIM_FUNC(func))(a[0]); break;
    case 2: r=(*SCM_CPRIM_FUNC(func))(a[0],a[1]); break;
    case 3: r=(*SCM_CPRIM_FUNC(func))(a[0],a[1],a[2]); break;
    case 4: r=(*SCM_CPRIM_FUNC(func))(a[0],a[1],a[2],a[3]); break;
    case 5: r=(*SCM_CPRIM_FUNC(func))(a[0],a[1],a[2],a[3],a[4]); break;
    case 6: r=(*SCM_CPRIM_FUNC(func))(a[0],a[1],a[2],a[3],a[4],a[5]); break;
    case 7: r=(*SCM_CPRIM_FUNC(func))(a[0],a[1],a[2],a[3],a[4],a[5],a[6]); break;
    case 8: r=(*SCM_CPRIM_FUNC(func))(a[0],a[1],a[2],a[3],a[4],a[5],a[6],a[7]); break;
    case 9: r=(*SCM_CPRIM_FUNC(func))(a[0],a[1],a[2],a[3],a[4],a[5],a[6],a[7],a[8]); break;
    case 10:r=(*SCM_CPRIM_FUNC(func))(a[0],a[1],a[2],a[3],a[4],a[5],a[6],a[7],a[8],a[9]); break;
    default:r=(*SCM_CPRIM_FUNC(func))(sp - env->slot, env->slot);
    }
    sp   = (SOBJ*)cont;
    cont = cont->next;
    env  = env->next;
    TOS  = r;
    *sp  = TOS;
    END(gen_callcprim);
  }

  INSTR(gen_callcode) {
    SCM_ERR("gen_callcode not implemented", TOS);
    END(gen_callcode);
  }
  
  INSTR(gen_callmcode) {
    SOBJ func;
    SCM_MCode *mc;

  mcode_execute:
    spop(func);
    mc = SCM_MCODE(func);
    if (cont) cont->ip = ip;
    ip = mc->lit;
    goto *(mc->code);
    SCM_ERR("gen_callmcode not implemented", TOS);
    END(gen_callmcode);
  }
  
  INSTR(gen_callproc) {
    SCM_ERR("gen_callproc not implemented", TOS);
    END(gen_callproc);
  }

  INSTR(gen_callclos) {
    SOBJ closure;
    SCM_Code *code;
    SOBJ *init_sp = sp;
    spop(closure);
    cont->ip    = ip;
    cont->env 	= env;

    env		= ((void*)cont) + sizeof(CONTEXT);
    env->next   = SCM_CLOSURE_ENV(closure);

    code 	= SCM_PROC_CODE(SCM_CLOSURE_CODE(closure));

    if (init_sp - env->slot < code->nargs - code->optargs)
      SCM_ERR("callclos: wrong number of arguments", closure);

    env->nslots = SCM_MKINUM(code->nargs);

    if (code->optargs) {	/* optionnal arguments */
      SOBJ list = NULL, *p = env->slot + (code->nargs-1);
      while(sp >= p) {
	list = scm_cons(*sp--, list);
      }
      TOS = list; sp++;
    }
    ip   	= code->code;
    END(gen_callclos);
  }
  INSTR(gen_callmacro) {
    SCM_ERR("gen_callmacro not implemented", TOS);
    END(gen_callmacro);
  }

  /* general call of whatever
   *
   * Stack layout before:
   * [ cont envframe arg1 arg2 .. argn func ]
   * 
   * Stack layout after:
   * [ result ]
   */
  INSTR(call) {
    SOBJ func = TOS;
    if (func == NULL || SCM_INUMP(func)) SCM_ERR("illegal func", func);
#ifdef VM3_DEBUG
    fprintf(stderr, "VM3:call:func->type=%d\n", func->type);
    TOS = func;
#endif
    goto *(scm_call_table[func->type]);
    END(call);
  }

  /*** General Jump to whatever */
  INSTR(jump) {
    SOBJ *s, *d, *l;
    SCM_Code *code;
    int nargs = SCM_INUM(*ip++);
    SOBJ func = TOS;

    if (!SCM_CLOSUREP(func))
      SCM_ERR("VM3:jump:illegal func type", func);

    code =SCM_PROC_CODE(SCM_CLOSURE_CODE(func));

    /*KKK TAIL*/

    /* check number of args */
    if (nargs != code->nargs)
      SCM_ERR("VM3:jump:bad number of args", func);

    ip  = code->code;

    /* copy current arguments to env */
    s = sp - nargs;   l = sp;  d = env->slot;
    while(s < l) {  *d++ = *s++; }
    sp  = d - 1;
    TOS = *sp;
    /* SCM_ERR("VM3:jump: not implemented", TOS); */
    END(jump);
  }
 
  /* Stack layout when callprim:
   *
   * Before
   *
   * [ envframe arg1 arg2 .. argn ]
   *   ^                           ^
   *   env                         sp
   *
   * After
   *
   * [ result ]
   *   ^
   *   sp
   */
  INSTR(callcprim) {
    SOBJ func, ret;
    int  nargs;
    func = *ip++;
    nargs = SCM_INUM(*ip++);
    supdate();
    sp -= (nargs-1);
    ret = (*SCM_CPRIM_FUNC(func))(nargs, sp);
    TOS = ret;
    *sp = TOS;
    END(callcprim);
  }
  /* stack layout before callcprim?
   * [ arg0 ... argn-1 ]
   * after:
   * [ result ]
   */
  INSTR(callcprim0) {
    SOBJ func = *ip++, ret;
    supdate();
    ret = (*SCM_CPRIM_FUNC(func))();
    sp++;  TOS = ret;
    END(callcprim0);
  }
  INSTR(callcprim1) {
    SOBJ func = *ip++, ret;
    supdate();
    ret = (*SCM_CPRIM_FUNC(func))(sp[0]);
    TOS = ret;
    END(callcprim1);
  }
  INSTR(callcprim2) {
    SOBJ func = *ip++, ret;
    supdate();
    sp -= 1;
    ret = (*SCM_CPRIM_FUNC(func))(sp[0],sp[1]);
    TOS = ret;
    END(callcprim2);
  }
  INSTR(callcprim3) {
    SOBJ func = *ip++, ret;
    supdate();
    sp -= 2;
    ret = (*SCM_CPRIM_FUNC(func))(sp[0],sp[1],sp[2]);
    TOS = ret;
    END(callcprim3);
  }
  INSTR(callcprim4) {
    SOBJ func = *ip++, ret;
    supdate();
    sp -= 3;
    ret = (*SCM_CPRIM_FUNC(func))(sp[0],sp[1],sp[2],sp[3]);
    TOS = ret;
    END(callcprim4);
  }
  INSTR(callcprim5) {
    SOBJ func = *ip++, ret;
    supdate();
    sp -= 4;
    ret = (*SCM_CPRIM_FUNC(func))(sp[0],sp[1],sp[2],sp[3],sp[4]);
    TOS = ret;
    END(callcprim5);
  }
  INSTR(callcprim6) {
    SOBJ func = *ip++, ret;
    supdate();
    sp -= 5;
    ret = (*SCM_CPRIM_FUNC(func))(sp[0],sp[1],sp[2],sp[3],sp[4],sp[5]);
    TOS = ret;
    END(callcprim6);
  }
  INSTR(callcprim7) {
    SOBJ func = *ip++, ret;
    supdate();
    sp -= 6;
    ret = (*SCM_CPRIM_FUNC(func))(sp[0],sp[1],sp[2],sp[3],sp[4],sp[5],sp[6]);
    TOS = ret;
    END(callcprim7);
  }
  INSTR(callcprim8) {
    SOBJ func = *ip++, ret;
    supdate();
    sp -= 7;
    ret = (*SCM_CPRIM_FUNC(func))(sp[0],sp[1],sp[2],sp[3],sp[4],sp[5],sp[6],sp[7]);
    TOS = ret;
    END(callcprim8);
  }
  INSTR(callcprim9) {
    SOBJ func = *ip++, ret;
    supdate();
    sp -= 8;
    ret = (*SCM_CPRIM_FUNC(func))(sp[0],sp[1],sp[2],sp[3],sp[4],sp[5],sp[6],sp[7],sp[8]);
    TOS = ret;
    END(callcprim9);
  }
  INSTR(callcprim10) {
    SOBJ func = *ip++, ret;
    supdate();
    sp -= 9;
    ret = (*SCM_CPRIM_FUNC(func))(sp[0],sp[1],sp[2],sp[3],sp[4],sp[5],sp[6],sp[7],sp[8],sp[9]);
    TOS = ret;
    END(callcprim10);
  }

  /* catch / uncatch : error trapper */
  INSTR(catch) {
    SOBJ catch;
    SCM_CatchContext *c;
    SOBJ handler, tag;

    spop(handler);
    spop(tag);
    catch = scm_mkcatch();
    c = SCM_CATCH_CONTEXT(catch);
    vm->catch_list = scm_cons(catch, vm->catch_list);
    c->tag 	= tag;
    c->handler  = handler;
    c->vm.sp   	= sp;
    c->vm.ip   	= ip + SCM_INUM(*ip++);
    c->vm.cont 	= (void*)cont;
    c->vm.env  	= env;
    if (setjmp(c->env) != 0) {	/* got a catch */
      sp = c->vm.sp;
      ip = c->vm.ip;
      cont = (void*)c->vm.cont;
      env = c->vm.env;
      TOS = *sp;
      if (c->handler) {
	CONTEXT *ctxt;
	/* prepcall */
	ctxt    = (CONTEXT*)(sp+1);  ctxt->next = cont;
	cont = ctxt;
	sp   = ((void*)ctxt)+(sizeof(CONTEXT)+offsetof(SCM_ENVFRAME,nslots));
	TOS  = SCM_INUM0;
	spush(vm->thrown_tag);
	spush(vm->thrown_msg);
	spush(c->handler);
	goto l_call;
      }
      SCM_ERR("catch: no handler for", vm->thrown_tag);
    }
    END(catch);
  }

  INSTR(uncatch) {
    if (vm->catch_list) {
      vm->catch_list = SCM_CDR(vm->catch_list);
    }
    END(uncatch);
  }


  /* KKK */
#include "prim3.i"

  END_VMCODE;
 a_vm3_exit: ALIGN();
 l_vm3_exit:
 e_vm3_exit:
}

SOBJ scm_vm3_init()
{
  scm_vm3_vmd.code = SCM_VM3_INIT;
  scm_vm3(&scm_vm3_vmd);
  return(NULL);
}

SOBJ scm_vm3_asm(SOBJ code)
{
  SCM_VMD vm;

  vm.code = SCM_VM3_ASSEMBLE;  vm.arg.addr = code;
  scm_vm3(&vm);
  return(vm.ret.obj);
}

SOBJ scm_vm3_nasm(SOBJ code)
{
  SCM_VMD vm;

  vm.code = SCM_VM3_NASSEMBLE;  vm.arg.addr = code;
  scm_vm3(&vm);
  return(vm.ret.obj);
}

SOBJ scm_vm3_dis(SOBJ code)
{
  SCM_VMD vm;

  vm.code = SCM_VM3_DISASSEMBLE;  vm.arg.addr = code;
  scm_vm3(&vm);
  return(vm.ret.obj);
}

SOBJ scm_vm3_exec(SOBJ code)
{
  if (SCM_ARRAYP(code)) {
    scm_vm3_vmd.code = SCM_VM3_EXECUTE;
    scm_vm3_vmd.reg.ip = SCM_ARRAY(code);
    scm_vm3(&scm_vm3_vmd);
    return(scm_vm3_vmd.ret.obj);
  }
  if (SCM_CLOSUREP(code)) {
    SCM_Code *c = SCM_PROC_CODE(SCM_CLOSURE_CODE(code));
    scm_vm3_vmd.code = SCM_VM3_EXECUTE;
    scm_vm3_vmd.reg.ip = c->code;
    scm_vm3_vmd.reg.env = SCM_CLOSURE_ENV(code);
    scm_vm3(&scm_vm3_vmd);
    return(scm_vm3_vmd.ret.obj);
  }
  if (SCM_MCODEP(code)) {
    scm_vm3_vmd.code = SCM_VM3_MCODE_EXECUTE;
    scm_vm3_vmd.arg.obj = code;
    scm_vm3(&scm_vm3_vmd);
    return(scm_vm3_vmd.ret.obj);
  }
  SCM_ERR("bad code", code);
  return(NULL);
}

SOBJ scm_vm3_throw(SOBJ tag, SOBJ msg)
{
  SOBJ node, list, tagl;
  SCM_VMD *vm = &scm_vm3_vmd;
  
#ifdef DEBUG
  scm_puts("*** throw: tag="); scm_cdisplay(tag);
  scm_puts(" msg=");  scm_cprint(msg);
#endif

  /*-- prepare to pass tag and msg to handler */
  vm->thrown_tag = tag;
  vm->thrown_msg = msg;

  if (tag == scm_false) {		/* throw to enclosing exception level */
	if (vm->catch_list == NULL) { /* top level exception */
	  longjmp(scm_errjmp, SCM_ERR_ABORT);
	} else {
	  node = SCM_CAR(vm->catch_list);
	  vm->catch_list = SCM_CDR(vm->catch_list);
	  longjmp(SCM_CATCH_CONTEXT_ENV(node), SCM_ERR_ABORT);
	}
  }
  if (tag == scm_true) {		/* throw to top level */
	vm->catch_list = NULL;
	longjmp(scm_errjmp, SCM_ERR_ABORT);
  }
  
  if (!SCM_ATOMP(tag)) {
	scm_puts("*** ERROR: tag "); scm_cdisplay(tag);
	scm_puts(" is not an atom\n");
	longjmp(scm_errjmp, SCM_ERR_ABORT);
  }
  for (list = vm->catch_list; list; list = SCM_CDR(list)) {
	node = SCM_CAR(list);
	tagl = SCM_CATCH_CONTEXT_TAG(node);

	/* note: #t as taglist match all tags */
	if (tagl == scm_true ||
		(SCM_PAIRP(tagl) && scm_memq(tag, tagl) != scm_false)) {
	  vm->catch_list = SCM_CDR(list); /* consume catchlist to next catch */
	  vm->thrown_tag = tag;	  
	  vm->thrown_msg = msg;
	  longjmp(SCM_CATCH_CONTEXT_ENV(node), SCM_ERR_ABORT);
	}
  }
  /* no tag match: goto toplevel */
#ifdef DEBUG
  scm_puts("*** error: tag="); scm_cdisplay(tag);
  scm_puts(" message="); 
#endif
  scm_cprint(msg);
  longjmp(scm_errjmp, SCM_ERR_ABORT);
}


/*E* (make-proc VECTOR ENV NARG OPTARG?) => PROC */
/*D* Create a new procedure. The VECTOR contains the code of the body
  of the procedure, the ENV is the full compile-time environment
  chain, the NARGS is the number of arguments (an optionnal argument
  counts as one argument) and OPTARG? is #t when last argument is
  optionnal. */
SOBJ scm_make_proc(SOBJ v, SOBJ env, SOBJ nargs, SOBJ optarg)
{
  SOBJ proc;
  SCM_Code *code;
  long size;
  
  if (!SCM_ARRAYP(v))		SCM_ERR("bad vector", v);
  if (env && !SCM_PAIRP(env))	SCM_ERR("bad env chain", env);
  if (!SCM_INUMP(nargs)) 	SCM_ERR("bad nargs", nargs);

  size 		= SCM_ASIZE(v);
  code 		= scm_must_alloc(offsetof(SCM_Code, code[size]));
  code->envlist = NULL;
  code->nargs 	= SCM_INUM(nargs);
  code->optargs	= (optarg != scm_false);
  code->nlocals = 0;
  code->size 	= size;

  memcpy(code->code, SCM_ARRAY(v), size * sizeof(SOBJ*));

  proc = scm_newcell(SOBJ_T_PROC);

  SCM_PROC_CODE(proc) = code;
  SCM_PROC_ENV(proc) = NULL;
  return(proc);
}

SOBJ scm_mcode_dis(SOBJ start_addr, SOBJ length)
{
  if (!SCM_NUMBERP(start_addr)) SCM_ERR("bad number", start_addr);
  if (!SCM_NUMBERP(length))	SCM_ERR("bad length", length);
  scm_disassemble_region((void*)scm_number2long(start_addr),
			 scm_number2long(length));
  return(scm_undefined);
}

/*E* (vm3-boot FILE) => BOOLEAN */
/*D* Boot from a file. File content is expected to be a bunch of
  assembly arrays. If a 'run symbol is defined and this symbol is a
  procedure, then this procedure is automatically executed. */

SOBJ scm_vm3_boot(SOBJ file)
{
  SOBJ got;
  PORT *in;
  SOBJ sym;
  char *p, *q, *name;

  if (!SCM_STRINGP(file))	SCM_ERR("bad filename", file);

  in = port_open_input_file(SCM_STR_VALUE(file));
  if (in == NULL) 		SCM_ERR("cannot open bootfile", file);
  
  while ( (got = scm_read_port(in)) != scm_eof ) {
    if (!SCM_ARRAYP(got)) 	SCM_ERR("bad file contents", got);
    scm_vm3_exec(scm_vm3_asm(got));
  }
  port_close(in);

  /* extract filename from possible path */
  if ((p = strrchr(SCM_STR_VALUE(file), '/')) == NULL)
    p = SCM_STR_VALUE(file);

  if ((q = strrchr(p, '.')) == NULL)
    q = p + strlen(p);

  /* alloc and append "-run" */
  name = scm_must_alloc(q - p + 8);
  strncpy(name, p, q - p);
  strcpy(name + (q - p), "-run");
  
  sym = scm_hash_search(scm_symbol_hash, scm_mkatom(name));
  scm_free(name);

  if (!SCM_SYMBOLP(sym))  return(scm_false);
  if (!SCM_CLOSUREP(SCM_SYM_VALUE(sym)))  return(scm_false);
  scm_vm3_exec(SCM_SYM_VALUE(sym));
  return(scm_true);
}

void scm_init_vm3()
{
  scm_vm3_init();
  scm_add_cprim("make-procedure",	scm_make_proc,	4);
  scm_add_cprim("vm3-asm",	scm_vm3_asm,	1);
  scm_add_cprim("vm3-nasm",	scm_vm3_nasm,	1);
  scm_add_cprim("vm3-dis",	scm_vm3_dis,	1);
  scm_add_cprim("vm3-exec",	scm_vm3_exec,	1);
  scm_add_cprim("vm3-throw",	scm_vm3_throw,	2);
  scm_add_cprim("mcode-dis",	scm_mcode_dis,	2);
  scm_add_cprim("vm3-boot",	scm_vm3_boot,	1);
}

